/****************************************************************************/
/*

DESCRIPTION

	Implements a CharLayer Transport Interface using Freescale Topaz byte channel implementation.
	Reading / Writing a character is currently non blocking.


COPYRIGHT

	Copyright 2008-2010 Freescale Semiconductor, Inc.

*/
/****************************************************************************/

#include "byte_chan_wrapper.h"

#include <stdint.h>
#include <devtree.h>
#include <events.h>

#include "usr_put.h"
#include "trk_agent.h"
#include "targimpl_ppc.h"
#include "targimpl.h"

#include "PrivateData.h"

static DSError InitializeByteChan( LayerParams *p );
static DSError TerminateByteChan( LayerParams *p );
static DSError ReadByteChan( LayerParams *params, char *c );
static DSError WriteByteChan( LayerParams *p, char c );
static DSError FlushByteChan( LayerParams *p );
static DSError IrqByteChan( LayerParams *p );

/******************************************************************************/
/* global variables                                                           */
/******************************************************************************/

CharLayerApi gByteChanApi =
{
	InitializeByteChan,				/* init */
	TerminateByteChan,				/* terminate */
	ReadByteChan,					/* read */
	WriteByteChan,					/* write */
	FlushByteChan,					/* flush */
	IrqByteChan						/* irq */
};

/******************************************************************************/
/* Topaz hypervisor implementation                                            */
/******************************************************************************/

static DSError InitializeByteChan( LayerParams *params )
{
	dt_node_t* dbg_stub_tree = NULL;
	gcpu_t *gcpu = get_gcpu();

	if (unlikely(!gcpu))
	{
		return kInternalError;
	}

	dbg_stub_tree = gcpu->dbgstub_cfg;

	if (unlikely(!dbg_stub_tree))
	{
		return kInternalError;
	}

	if (!dbg_stub_tree->bch)
	{
		init_byte_channel(dbg_stub_tree);
	}
	/* note: the byte-channel handle is returned in node->bch */
	if (unlikely(!dbg_stub_tree->bch))
	{
		TRACE("Could not intialize the byte-channel for this core");
		return kInternalError;
	}

	/* This core is debugged using HyperTRK Stub */

	/* No callback on a TX, since we're polling.
	 * Register RX callback to be the TransportIrq handler
	 * from the Portable transport layer in order to put a
	 * new GEV_TRK in the hypervisor vent queue when a new char
	 * is received on a byte channel for the first time. It will
	 * call all the transport Irq handler for each available layer
	 */
	ByteChannelHandler* chanHandler = (ByteChannelHandler*)params;
	if (chanHandler != NULL)
	{
		chanHandler->m_currentPos = 0;
		chanHandler->m_noReceived = 0;
		chanHandler->m_sndBufferStart = chanHandler->m_sndBuffer;
		chanHandler->m_sndBufferEnd   = chanHandler->m_sndBuffer;
		chanHandler->m_byteChanHandle = dbg_stub_tree->bch;
	}

	__puts("Byte channel successfully registered\r\n");
	return kNoError;
}


static DSError TerminateByteChan( LayerParams *params )
{
	__puts("Closing byte channel\r\n");
	ByteChannelHandler* chanHandler = (ByteChannelHandler*)params;
	if (chanHandler && chanHandler->m_byteChanHandle)
	{
		byte_chan_handle_t* byte_channel_handle = chanHandler->m_byteChanHandle;
		if (byte_channel_handle->rx)
		{
			byte_channel_handle->rx->data_avail     = NULL;
			smp_lwsync();
			byte_channel_handle->rx->consumer       = NULL;
		}
	}

	return kNoError;
}


static DSError ReadByteChan( LayerParams *params, char *c )
{
	ByteChannelHandler* chanHandler = (ByteChannelHandler*)params;

	if (chanHandler->m_currentPos < chanHandler->m_noReceived)
	{
		*c = chanHandler->m_recBuffer[chanHandler->m_currentPos++];
		return kNoError;
	}

	/* the received buffer is now empty */
	chanHandler->m_noReceived = byte_chan_receive( chanHandler->m_byteChanHandle,
		(uint8_t*)chanHandler->m_recBuffer, (size_t)CHANNEL_RCV_BUFFER );

	if (chanHandler->m_noReceived <= 0)
	{
		chanHandler->m_noReceived = 0;
		return kNotFoundError;
	}

	// reset the buffer
	chanHandler->m_currentPos = 0;
	*c = chanHandler->m_recBuffer[chanHandler->m_currentPos++];
	return kNoError;
}


static DSError WriteByteChan( LayerParams *params, char c )
{
	ssize_t byte_count = 0;
	ByteChannelHandler* chanHandler = (ByteChannelHandler*)params;

	// compute where our buffer will end
	ui8* bufferEnd  = chanHandler->m_sndBuffer + CHANNEL_SND_BUFFER;

	// we compute the next position of the start buffer address
	ui8* sndNextPos = ( (chanHandler->m_sndBufferStart + 1) <  bufferEnd ) ?
						chanHandler->m_sndBufferStart + 1 :
						chanHandler->m_sndBuffer;

	// place the byte in the buffer
	*chanHandler->m_sndBufferStart = c;

	// we will use the one byte open solution
	if ( sndNextPos == chanHandler->m_sndBufferEnd )
	{
		// the buffer is full. Try to send as many characters as we can till the end.
		// At least one must be sent to free a byte position
		do
		{
			byte_count = byte_chan_send( chanHandler->m_byteChanHandle,
										 (uint8_t*)chanHandler->m_sndBufferEnd,
										 (size_t)(bufferEnd - chanHandler->m_sndBufferEnd) );
		} while (byte_count <= 0);

		chanHandler->m_sndBufferEnd += byte_count;
		// we reset the end pointer if needed
		chanHandler->m_sndBufferEnd  = (chanHandler->m_sndBufferEnd >= bufferEnd) ?
										chanHandler->m_sndBuffer :
										chanHandler->m_sndBufferEnd;
	}

	// we must reset the new buffer start position
	chanHandler->m_sndBufferStart  = sndNextPos;

	return kNoError;
}


static DSError FlushByteChan( LayerParams *params )
{
	ByteChannelHandler* chanHandler = (ByteChannelHandler*)params;
	ssize_t byte_count = 0;
	ui8* bufferEnd = chanHandler->m_sndBuffer + CHANNEL_SND_BUFFER;

	// send the data from m_sndBufferEnd to m_sndBufferStart
	size_t toSend = (chanHandler->m_sndBufferEnd <= chanHandler->m_sndBufferStart) ?
			chanHandler->m_sndBufferStart - chanHandler->m_sndBufferEnd :
			bufferEnd - chanHandler->m_sndBufferEnd;

	while (toSend > 0)
	{
		byte_count = byte_chan_send( chanHandler->m_byteChanHandle,
									 chanHandler->m_sndBufferEnd, toSend );
		if (byte_count < 0)
		{
			byte_count = 0;
		}

		chanHandler->m_sndBufferEnd += byte_count;
		toSend -= byte_count;

		if (toSend == 0)
		{
			// we reset the end pointer if needed
			chanHandler->m_sndBufferEnd   = (chanHandler->m_sndBufferEnd >= bufferEnd) ?
											chanHandler->m_sndBuffer :
											chanHandler->m_sndBufferEnd;
			// we compute an additional size to send if necessary
			toSend = (chanHandler->m_sndBufferEnd < chanHandler->m_sndBufferStart) ?
					chanHandler->m_sndBufferStart - chanHandler->m_sndBufferEnd :
					0;
		}
	}

	return kNoError;
}


static DSError IrqByteChan( LayerParams *params )
{
	/* when a new byte is available on the byte channel TransportIrq() will finally
	 * call this function to check if new data has been received
	 */
	bool receivedData = ((TRKState_PPC*)PRIVATE_DATA_HYPER(TRK_STATE))->inputActivated == true;
	if (receivedData)
	{
		__puts("Received New Data in Byte Channel! \r\n");
		// return a value > 0 to show that a new data has been received
		return 1;
	}
	return 0;
}


void TransportIrqHelper(queue_t *q, int blocking)
{
	/* data available is received on the first core for a muxer byte channel
	 * We set a new TRK event in hypervisor guest event queue, after we set a pending
	 * transport irq to the core the data is available for.
	 */
	if (unlikely(!q))
	{
		return;
	}
	gcpu_t* consumerCpu = (gcpu_t*)q->consumer;
	if (unlikely(consumerCpu == NULL))
		return;

	int trkEventId = PRIVATE_DATA_HYPER_CPU(GEVENT_ID, consumerCpu);
	if (trkEventId > 0)
	{
		((TRKState_PPC*)PRIVATE_DATA_HYPER_CPU(TRK_STATE, consumerCpu))->inputActivated = true;

		/* Record the event type and setevent GEV_TRK on the the desired virtual CPU. */
		setgevent(consumerCpu, trkEventId);
	}
	/* TODO: What if there were some other event in progress?  It'll be
	 * lost. The TRK event handler should check whether the input queue
	 * is empty
	 */
}

bool EnableByteChannelEvents( void )
{
	byte_chan_handle_t* byte_channel_handle = PRIVATE_DATA_HYPER(BYTE_CHANNEL);
	byte_channel_handle->rx->consumer   = get_gcpu();
	smp_lwsync();
	byte_channel_handle->rx->data_avail = TransportIrqHelper;
	smp_lwsync();
	return true;
}

bool DisableByteChannelEvents( void )
{
	PRIVATE_DATA_HYPER(BYTE_CHANNEL)->rx->data_avail = NULL;
	return true;
}

bool EnableByteChannelReceive( void )
{
	__puts("Enabling byte channel receive\r\n");

	/* Deregister callback first not to place other TRK events in hypervisor
	 * internal event queue after we enable byte channel support
	 * (critical interrupts) and we are still in the agent
	 */
	PRIVATE_DATA_HYPER(BYTE_CHANNEL)->rx->data_avail = NULL;
	smp_lwsync();
	/* enable interrupts, so UART keeps working */
	/* FIXME: we shouldn't enable EE as this will cause a scheduale to occur */
	enable_int();

	return true;
}

bool DisableByteChannelReceive( void )
{
	if (PRIVATE_DATA_HYPER(BYTE_CHANNEL) == NULL)
	{
		/* no data to be received in the queue */
		((TRKState_PPC*)PRIVATE_DATA_HYPER(TRK_STATE))->inputActivated = false;
		return true;
	}

	if ( !XTargetStopped() || queue_empty(PRIVATE_DATA_HYPER(BYTE_CHANNEL)->rx) ||
			PRIVATE_DATA_HYPER(IN_RESET) )
	{
		__puts("No more data in the queue. Disabling byte chan receive.\r\n");
		/* We will be exiting the agent soon so we will need to place another
		 * TRK event in Topaz internal event queue in order to be called
		 * again if there are any data received on the byte channel
		 */
		PRIVATE_DATA_HYPER(BYTE_CHANNEL)->rx->data_avail = TransportIrqHelper;
		smp_lwsync();
		/* no data to be received in the queue */
		((TRKState_PPC*)PRIVATE_DATA_HYPER(TRK_STATE))->inputActivated = false;
		return true;
	}
	return false;
}


