/****************************************************************************/
/*

DESCRIPTION

	Implements the guest access API to use the existing Topaz internal API to
	access memory and registers. Provides basic access to guest resources.

COPYRIGHT

	Copyright 2008-2010 Freescale Semiconductor, Inc.

*/
/****************************************************************************/

#include "hypervisor_guest_access.h"

#include <byte_chan.h>
#include <guestmemio.h>
#include <greg.h>
#include <paging.h>

#include "HyperRegTable.h"
#include "trk_agent.h"
#include "PrivateData.h"

#include "dserror.h"


/****************************************************************************/
/*
 *	InitializeTarget
 *
 *	Performs initialization of target-program state that is performed at
 *	startup.
 *
 */
/****************************************************************************/
DSError InitializeTarget (void)
{
	return( kNoError );
}

/***************************************************************************/
/*
 *	GetGuestTrapFrame
 *
 *	Returns the current topaz structure that can be used to access guest register values
 *
 */
/****************************************************************************/
trapframe_t* GetGuestTrapFrame()
{
	return PRIVATE_DATA_HYPER(GUEST_FRAME);
}

/***************************************************************************/
/*
 *	ConvertTrkRegisterId
 *
 *  Gets the information required to read a guest register using topaz API
 *  from a trk register set identifier and a set offset
 *
 */
/****************************************************************************/
bool ConvertTrkRegisterId(DSMessageRegisterOptions regSet, ui32 trkSetId, HyperRegisterInfo** toHyperRegInfo)
{
	ui32 tableIndex;

	if (toHyperRegInfo == NULL)
		return false;

	switch (regSet) {
		case kDSRegistersDefault:
		{
			tableIndex = DEFAULT_REGISTERS_HYPER_INDEX + trkSetId;
			break;
		}
		case kDSRegistersFP:
		{
			tableIndex = FP_REGISTERS_HYPER_INDEX + trkSetId;
			break;
		}
		case kDSRegistersExtended1:
		{
			tableIndex = EXTENDED1_REGISTERS_HYPER_INDEX + trkSetId;
			break;
		}
		case kDSRegistersExtended2:
		{
			tableIndex = EXTENDED2_REGISTERS_HYPER_INDEX + trkSetId;
			break;
		}

		default:
			return false;
	}

	if (tableIndex < kTrkRegisterTableNum)
	{
		*toHyperRegInfo = &trk_register_table[tableIndex];
		return true;
	}
	TRACE("Received an invalid trk register set id %d .", (int)trkSetId);
	return false;
}

/***************************************************************************/
/*
 *	read_trk_mmr
 *
 *  Reads a mmr register that has been specified by it's id
 *  In our case it will read the CCSRBARH and CCSRBARL.
 *
 */
/****************************************************************************/
static int read_trk_mmr(int mmr, register_t *val)
{
	size_t ccsrSize = 0;
	ui64 ccsrbar = get_ccsr_phys_addr(&ccsrSize);
	
	switch(mmr)
	{
		case 0:
			*val = (register_t)(ccsrbar >> 32);
			return 0;
		case 1:
			*val = (register_t)ccsrbar;
			return 0;
		default:
			return 1;
	}
}

/***************************************************************************/
/*
 *	read_trk_debug
 *
 *  Reads a debug register that has been specified by it's core id
 *
 */
/****************************************************************************/
static int read_trk_debug(int dbgId, register_t *val)
{
	switch (dbgId)
	{
		case SPR_DBSR:
			*val = mfspr(SPR_DBSR);
			break;
		case SPR_DBCR0:
			*val = mfspr(SPR_DBCR0);
			break;
		case SPR_DBCR1:
			*val = mfspr(SPR_DBCR1);
			break;
		case SPR_DBCR2:
			*val = mfspr(SPR_DBCR2);
			break;
		case SPR_IAC1:
			*val = mfspr(SPR_IAC1);
			break;
		case SPR_IAC2:
			*val = mfspr(SPR_IAC2);
			break;
		case SPR_DAC1:
			*val = mfspr(SPR_DAC1);
			break;
		case SPR_DAC2:
			*val = mfspr(SPR_DAC2);
			break;
		default:
			return 1;
	}
	return 0;
}

/***************************************************************************/
/*
 *	write_trk_debug
 *
 *  Writes a debug register that has been specified by it's core id
 *
 */
/****************************************************************************/
static int write_trk_debug(int dbgId, register_t val)
{
	switch (dbgId)
	{
		case SPR_DBSR:
			mtspr(SPR_DBSR, val);
			break;
		case SPR_DBCR0:
			mtspr(SPR_DBCR0, val);
			break;
		case SPR_DBCR1:
			mtspr(SPR_DBCR1, val);
			break;
		case SPR_DBCR2:
			mtspr(SPR_DBCR2, val);
			break;
		case SPR_IAC1:
			mtspr(SPR_IAC1, val);
			break;
		case SPR_IAC2:
			mtspr(SPR_IAC2, val);
			break;
		case SPR_DAC1:
			mtspr(SPR_DAC1, val);
			break;
		case SPR_DAC2:
			mtspr(SPR_DAC2, val);
			break;
		default:
			return 1;
	}
	return 0;
}

/***************************************************************************/
/*
 *	read_register
 *
 *  Reads a guest register using the current trap_frame in a transparent way
 *
 */
/****************************************************************************/
bool read_register(trapframe_t *trap_frame, HyperRegisterInfo* regInfo, byte* reg_value)
{
	int c;
	if ((regInfo == NULL) || (reg_value == NULL))
		return false;

	switch (regInfo->group)
	{
		case kGroupGPR:
			c = read_ggpr(trap_frame, regInfo->groupId, (register_t*)reg_value);
			break;
		case kGroupPC:
			c = read_gpc(trap_frame, (register_t*)reg_value);
			break;
		case kGroupMSR:
			c = read_gmsr(trap_frame, (register_t*)reg_value);
			break;
		case kGroupCR:
			c = read_gcr(trap_frame, (register_t*)reg_value);
			break;
		case kGroupSPR:
			c = read_gspr(trap_frame, regInfo->groupId, (register_t*)reg_value);
			break;
		case kGroupFPR:
			c = read_gfpr(trap_frame, regInfo->groupId, (uint64_t*)reg_value);
			break;
		case kGroupPMR:
			c = read_pmr(trap_frame, regInfo->groupId, (register_t*)reg_value);
			break;
		case kGroupFPSCR:
			c = read_fpscr((uint64_t*)reg_value);
			break;
		case kGroupMMR:
			c = read_trk_mmr(regInfo->groupId, (register_t*)reg_value);
			break;
		case kGroupDebug:
			c = read_trk_debug(regInfo->groupId, (register_t*)reg_value);
			break;
		case kGroupUnknown:
		default:
			return false;
	}
	return (c == 0);
}

/***************************************************************************/
/*
 *	write_register
 *
 *  Writes a guest register using the current trap_frame in a transparent way
 *
 */
/****************************************************************************/
bool write_register(trapframe_t *trap_frame, HyperRegisterInfo* regInfo, byte* reg_value)
{
	int c;
	if ((regInfo == NULL) || (reg_value == NULL))
		return false;

	uint64_t value = 0xbadbadbad;

	if (regInfo->size == kRegister32)
		value = *((uint32_t*)reg_value);
	else if (regInfo->size == kRegister64)
		value = *((uint64_t*)reg_value);

	switch (regInfo->group)
	{
		case kGroupGPR:
			c = write_ggpr(trap_frame, regInfo->groupId, value);
			break;
		case kGroupPC:
			c = write_gpc(trap_frame, value);
			break;
		case kGroupMSR:
			c = write_gmsr(trap_frame, value, 1);
			break;
		case kGroupCR:
			c = write_gcr(trap_frame, value);
			break;
		case kGroupSPR:
			c = write_gspr(trap_frame, regInfo->groupId, value);
			break;
		case kGroupFPR:
			c = write_gfpr(trap_frame, regInfo->groupId, &value);
			break;
		case kGroupFPSCR:
			c = write_fpscr(value);
			break;
		case kGroupPMR:
			c = write_pmr(trap_frame, regInfo->groupId, value);
			break;
		case kGroupDebug:
			c = write_trk_debug(regInfo->groupId, value);
			break;
		case kGroupMMR:
		case kGroupUnknown:
		default:
			return false;
	}
	return (c == 0);
}

/***************************************************************************/
/*
 *	read_tlb
 *
 *  Reads a guest TLB
 *
 */
/****************************************************************************/
int read_tlb(tlb_entry_t *tlb, unsigned int *flag)
{
    return guest_tlb_read(tlb, flag);
}

/***************************************************************************/
/*
 *	get_CCSR_access_chunk
 *
 *  Checks if the suplied memory access touches CCSR space. Returns the
 *  start of the CCSR access in startCCSR and the size in sizeCCSR
 *
 */
/****************************************************************************/
static bool get_CCSR_access_chunk(ui64 startAddress, ui32 size, ui64* startCCSR, ui32* sizeCCSR)
{
	if ((startCCSR == NULL) || (sizeCCSR == NULL))
	{
		return false;
	}
	
	size_t ccsrSize = 0;
	ui64 ccsrbar = get_ccsr_phys_addr(&ccsrSize);
	
	ui64 endAddress = startAddress + size;
	if ( (endAddress < ccsrbar) || (startAddress >= (ccsrbar + ccsrSize)) )
	{
		/* not in CCSR space */
		*startCCSR = 0;
		*sizeCCSR  = 0;
		return false;
	}

	*startCCSR = (startAddress < ccsrbar)? ccsrbar : startAddress;
	*sizeCCSR  = (endAddress < (ccsrbar + ccsrSize)) ?
			(endAddress - *startCCSR) : ((ccsrbar + ccsrSize) - *startCCSR);

	return true;
}

/***************************************************************************/
/*
 *	access_ccsr_address
 *
 *  Reads or writes to a memory address from the IO CCSR space
 *
 */
/****************************************************************************/
static ui32 access_ccsr_address(ui32 startAddress, ui32 size, byte* buf, bool toWrite)
{
    switch (size)
    {
	case 1:
	    if (toWrite)
    	    {
        	uint8_t val = *((uint8_t*)buf);
		out8((uint8_t*)startAddress, val);
    	    }
    	    else
    	    {
		uint8_t value = in8((uint8_t*)startAddress);
	        memcpy(buf, &value, sizeof(uint8_t));
	    }
	    break;	

	case 2:
	    if (toWrite)
    	    {
		uint16_t val = *((uint16_t*)buf);
        	out16((uint16_t*)startAddress, val);
    	    }
    	    else
    	    {
		uint16_t value = in16((uint16_t*)startAddress);
	        memcpy(buf, &value, sizeof(uint16_t));
	    }	    
	    
	    break;

	case 4:
	    if (toWrite)
    	    {
		uint32_t val = *((uint32_t*)buf);
        	out32((uint32_t*)startAddress, val);
    	    }
    	    else
    	    {
		uint32_t value = in32((uint32_t*)startAddress);
	        memcpy(buf, &value, sizeof(uint32_t));
	    }	    
	    break;
	
	default:
	    return 0;
    }

    return size;
}

/***************************************************************************/
/*
 *  	access_ccsr_space
 *
 *  Reads or writes a memory block from the IO CCSR space. 
 *
 */
/****************************************************************************/
static ui32 access_ccsr_space(ui32 startAddress, ui32 size, byte* buf, bool toWrite)
{
    ui32 currentAddress = startAddress;
    ui32 endAddress = startAddress + size;
    ui32 leftSize = size;
    ui32 i;
    
    // what size should be accessed first for making afterwards an alligned access
    ui32 allignment = currentAddress % 4;
    ui32 sizeToAllign = 0;
    if (allignment > 0)
	sizeToAllign = (size < 4 - allignment) ? size : 4 - allignment;
    
    while (sizeToAllign > 0)
    {
        if (currentAddress % 2 == 0 && sizeToAllign > 1)
        {
    	    currentAddress += access_ccsr_address(currentAddress, 2, buf, toWrite);
	    buf += 2;
	    sizeToAllign -= 2;
	}
	else
	{
	    currentAddress += access_ccsr_address(currentAddress, 1, buf, toWrite);
	    buf++;
	    sizeToAllign--;
	}
    }	
		
    // if we reached the end, just return
    if (currentAddress == endAddress)
        return size;
    
    leftSize -= sizeToAllign;
    // what size should be accessed alligned at the end after in32/out32 calls from the current address if necessary
    ui32 sizeToAccessAtEnd = (leftSize < endAddress % 4) ? leftSize : endAddress % 4;
    // how many times should the in32/out32 be used
    ui32 fullAccessTimes = (leftSize < sizeToAccessAtEnd) ? 0 : (leftSize - sizeToAccessAtEnd) / 4;
    
    if (fullAccessTimes > 0)
    {
	for (i = 0; i < fullAccessTimes; i++)
        {
    	    currentAddress += access_ccsr_address(currentAddress, 4, buf, toWrite);
	    buf += 4;
	}
    }
    
    // access now the end part    
    while (sizeToAccessAtEnd > 0)
    {
	if (currentAddress % 2 == 0 && sizeToAccessAtEnd > 1)
	{
	    currentAddress += access_ccsr_address(currentAddress, 2, buf, toWrite);
	    buf += 2;
	    sizeToAccessAtEnd -= 2;
	}
	else
	{
	    currentAddress += access_ccsr_address(currentAddress, 1, buf, toWrite);
	    buf++;
    	    sizeToAccessAtEnd--;
        }
    }	
    
    return size;
}

/***************************************************************************/
/*
 *	read_memory
 *
 *  Read memory from guest target. The read can be either physical or virtual.
 *  For physical guest it is Topaz's API responsability to add a tlb mapping to
 *  cover the guest physical address required.
 *  For virtual address we will use the External Load/Store facility with the
 *  current guest PID available. The AS bit is set to guest data address space value.
 *  To read instructions it should be set to guest instruction address space value.
 *
 */
/****************************************************************************/
ui32 read_memory(MemorySpace space, ui64 startAddress, ui32 size, byte* buf)
{
	ui64 endAddress;
	
	switch (space)
	{
		case kPhysicalGuest:
		{
			gcpu_t* virtualCPU;
			guest_t* partition;
			ui64 startCCSR = 0;
			ui32 sizeCCSR = 0;
			ui32 virtualCCSR = 0;
			ui32 readSize = 0;
			virtualCPU = get_gcpu();
			partition = virtualCPU->guest;

			// check if we need to access the CCSR space
			if (get_CCSR_access_chunk(startAddress, size, &startCCSR, &sizeCCSR))
			{
				if (startAddress < startCCSR)
				{
					ui32 toReadMemSize = startCCSR - startAddress;
					if (toReadMemSize > 0)
					{
						readSize += copy_from_gphys(partition->gphys, buf, (phys_addr_t)startAddress, (size_t)toReadMemSize);
						if (readSize != toReadMemSize)
							return readSize;
						
						startAddress += readSize;
						buf += readSize;
						size -= readSize;
					}
				}
				
				// get the Hypervisor virtual address for acessing the CCSR space
				virtualCCSR = (unsigned long)map(startCCSR, sizeCCSR, TLB_MAS2_IO, TLB_MAS3_KERN);
				
				ui32 readCCSRspace = access_ccsr_space(virtualCCSR, sizeCCSR, buf, false);
				if (readCCSRspace != sizeCCSR)
					return readSize + readCCSRspace;

				readSize     += readCCSRspace;
				startAddress += readCCSRspace;
				buf          += readCCSRspace;
				size         -= readCCSRspace;
			}

			readSize += copy_from_gphys(partition->gphys, buf, (phys_addr_t)startAddress, (size_t)size);
			
			return readSize;
		}
		case kVirtualInstrGuest32:
		{
			trapframe_t *guestFrame = GetGuestTrapFrame();
			/* enable guest address space bit to access guest virtual
			 * using external load/store facility
			 */
			guestmem_set_insn(guestFrame);

			endAddress = startAddress + size;
			for (; startAddress < endAddress; startAddress++ )
			{
				guestmem_in8( (uint8_t*)((ui32)startAddress), (uint8_t*)buf );
				buf++;
			}
			break;
		}
		case kVirtualDataGuest32:
		{
			trapframe_t *guestFrame = GetGuestTrapFrame();
			/* enable guest address space bit to access guest virtual
			 * using external load/store facility
			 */
			guestmem_set_data(guestFrame);

			endAddress = startAddress + size;
			for (; startAddress < endAddress; startAddress++ )
			{
				guestmem_in8( (uint8_t*)((ui32)startAddress), (uint8_t*)buf );
				buf++;
			}
			break;
		}
		default:
			size = 0;
	}

	return size;
}

/***************************************************************************/
/*
 *	write_memory
 *
 *  Writes memory from guest target. The read can be either physical or virtual.
 *  For physical guest it is Topaz's API responsability to add a tlb mapping to
 *  cover the guest physical address required.
 *  For virtual address we will use the External Load/Store facility with the
 *  current guest PID available. The AS bit is set to guest data address space value.
 *  To read instructions it should be set to guest instruction address space value.
 *
 */
/****************************************************************************/
ui32 write_memory(MemorySpace space, ui64 startAddress, ui32 size, byte* buf)
{
	ui64 endAddress;

	switch (space)
	{
		case kPhysicalGuest:
		{
			gcpu_t* virtualCPU;
			guest_t* partition;
			ui64 startCCSR = 0;
			ui32 sizeCCSR = 0;
			ui32 writeSize = 0;
			ui32 virtualCCSR = 0;
			virtualCPU = get_gcpu();
			partition = virtualCPU->guest;

			if (get_CCSR_access_chunk(startAddress, size, &startCCSR, &sizeCCSR))
			{
				if (startAddress < startCCSR)
				{
					ui32 toWriteMemSize = startCCSR - startAddress;
					if (toWriteMemSize > 0)
					{
						writeSize = copy_from_gphys(partition->gphys, buf, (phys_addr_t)startAddress, (size_t)toWriteMemSize);
						if (writeSize != toWriteMemSize)
							return writeSize;
						
						startAddress += writeSize;
						buf += writeSize;
						size -= writeSize;
					}
				}
				// get the Hypervisor virtual address for acessing the CCSR space
				virtualCCSR = (unsigned long)map(startCCSR, sizeCCSR, TLB_MAS2_IO, TLB_MAS3_KERN);
				
				ui32 writeCCSRspace = access_ccsr_space(virtualCCSR, sizeCCSR, buf, true);
				if (writeCCSRspace != sizeCCSR)
					return writeSize + writeCCSRspace;

				writeSize    += writeCCSRspace;
				startAddress += writeCCSRspace;
				buf          += writeCCSRspace;
				size         -= writeCCSRspace;
			}
			
			writeSize = copy_to_gphys(partition->gphys, (phys_addr_t)startAddress, buf, (size_t)size, 1);
			
			return writeSize;
		}
		case kVirtualInstrGuest32:
		{
			trapframe_t *guestFrame = GetGuestTrapFrame();
			/* enable guest address space bit to access guest virtual
			 * using external load/store facility
			 */
			guestmem_set_insn(guestFrame);

			endAddress = startAddress + size;
			for (; startAddress < endAddress; startAddress++ )
			{
				guestmem_out8( (uint8_t*)((ui32)startAddress), *((uint8_t*)buf) );
				buf++;
			}
			break;
		}
		case kVirtualDataGuest32:
		{
			trapframe_t *guestFrame = GetGuestTrapFrame();
			/* enable guest address space bit to access guest virtual
			 * using external load/store facility
			 */
			guestmem_set_data(guestFrame);

			endAddress = startAddress + size;
			for (; startAddress < endAddress; startAddress++ )
			{
				guestmem_out8( (uint8_t *)((ui32)startAddress), *((uint8_t*)buf) );
				buf++;
			}
			break;
		}
		default:
			size = 0;
	}

	return size;
}

/*******************************************************************************/
/*
 *	ClearStepPrivateData
 *
 *	Clears the private data used for pre and post single stepping operations.
 */
/*******************************************************************************/
inline void ClearStepPrivateData(void)
{
    PRIVATE_DATA_HYPER(MSR_INSTR) = PRIVATE_DATA_HYPER(STEP_PC_ADDR) = PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS) = 0;
}

/****************************************************************************/
/*
 *	PreStepOperations
 *
 *	Operations performed before a single step.
 */
/****************************************************************************/
void PreStepOperations()
{
	trapframe_t *guestFrame = GetGuestTrapFrame();
	register_t pc, msr;
	uint32_t insn = 0;
	
	/* instruction that could affect MSR  */
	PRIVATE_DATA_HYPER(MSR_INSTR) = 0;
	
	/* current PC  */
	PRIVATE_DATA_HYPER(STEP_PC_ADDR) = 0;
	
	/* bits we might modify in the MSR */
	PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS) = 0;
	
	/* set up override bit mask */
	uint32_t msr_bits_overridden = MSR_CE | MSR_EE | MSR_ME;

	/* read MSR and PC */ 
	if (read_gmsr(guestFrame, &msr) != 0 || 
	    read_gpc(guestFrame, &pc) != 0)
	    return;

	/* read and save the instruction at PC location */
	guestmem_set_data(guestFrame);
	if (guestmem_in32((uint32_t*)pc, &insn) != 0)
	    return;
	
	PRIVATE_DATA_HYPER(MSR_INSTR) = insn;
		
	/* normalize instruction pattern for comparisons */ 
	insn &= E500MC_MSR_INSTR;
	
	/* MSR is going to be written */
	if (insn == E500MC_MTMSR || insn == E500MC_MFMSR || insn == E500MC_WRTEE)
	{
	    /* get register containing MSR mask */ 
	    uint32_t regIndex = (PRIVATE_DATA_HYPER(MSR_INSTR) >> 21) & 0x1F;
	
	    register_t regval;
	    if (read_ggpr(guestFrame, regIndex, &regval) != 0)
	    {
		ClearStepPrivateData();
		return;
	    }
	    
	    /* preserve modified bits in source register */
	    if (insn == E500MC_MTMSR)
	    {
		PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS) = regval & msr_bits_overridden;
	    }
	    else if (insn == E500MC_MFMSR)
	    {
		PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS) = msr_bits_overridden;
	    }
	    else
	    {
		PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS) = msr & (MSR_CE | (regval & MSR_EE) | MSR_ME);
	    }
	    
	    /* modify bits in source register */
	    regval &= ~msr_bits_overridden;
	    if (write_ggpr(guestFrame, regIndex, regval) != 0)
	    {
		ClearStepPrivateData();
		return;
	    }
	}
	else if (insn == E500MC_WRTEEI)
	{
	    if (PRIVATE_DATA_HYPER(MSR_INSTR) & MSR_EE) /* wrteei 1  */
	    {
		/* change <wrteei 1> instruction to <wrteei 0> and
	         * ignore any error as it might be a read_only area 
	         * but save the PC for trying to restore the instruction after the step
	         */
	        if (guestmem_out32((uint32_t*)pc, (uint32_t)E500MC_WRTEEI) == 0)
		    PRIVATE_DATA_HYPER(STEP_PC_ADDR) = pc;
	    
		/* save off bits we might modify in the MSR and
		 * overwrite MSR[EE] to 1 as <wrteei 1> should enable it
		 */
	        PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS) = (msr & msr_bits_overridden) | MSR_EE;
	    }
	    else
	    {
		/* save off bits we might modify in the MSR and
		 * overwrite MSR[EE] to 0 as <wrteei 0> should disable it
		 */
		PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS) = (msr & msr_bits_overridden) & ~MSR_EE;
	    }
	}
	else if (insn == E500MC_RFI || insn == E500MC_RFCI || insn == E500MC_RFDI ||
		 insn == E500MC_RFMCI || insn == E500MC_RFGI)
	{
	    ClearStepPrivateData();
	    return;
	}
	else
	{
	    /* save off bits we might modify in the MSR */
	    PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS) = msr & msr_bits_overridden;
	}
	
	/* clear bits in MSR we want to override */ 
	msr &= ~msr_bits_overridden;
	
	/* write the modified MSR */ 
	if (write_gmsr(guestFrame, msr, 1) != 0)
	{
	    ClearStepPrivateData();
	    return;
	}
	
	return;
}

/****************************************************************************/
/*
 *	PostStepOperations
 *
 *	Operations performed after a single step.
 */
/****************************************************************************/
void PostStepOperations()
{
	trapframe_t *guestFrame = GetGuestTrapFrame();
	register_t msr;
	uint32_t insn = 0;
	bool exception = false;
	
	if (!PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS))
	    return;

	/* read MSR */ 
	if (read_gmsr(guestFrame, &msr) != 0)
	{
	    ClearStepPrivateData();
	    return;
	}
	    
	if (PRIVATE_DATA_HYPER(MSR_INSTR)) /* saved instruction that affects MSR */ 
	{
	    /* normalize instruction pattern for comparisons */ 
	    insn = PRIVATE_DATA_HYPER(MSR_INSTR) & E500MC_MSR_INSTR;
	    
	    /* Instruction contains write to/from MSR, need to restore register value */
	    if (insn == E500MC_MTMSR || insn == E500MC_MFMSR || insn == E500MC_WRTEE)
	    {
		/* get register containing MSR mask */ 
		uint32_t regIndex = (PRIVATE_DATA_HYPER(MSR_INSTR) >> 21) & 0x1F;
	
		register_t regval;
		if (read_ggpr(guestFrame, regIndex, &regval) != 0)
		{
		    goto restore_msr;
		}
		    
		/* restore overridden bits */
		regval |= PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS);
		if (write_ggpr(guestFrame, regIndex, regval) != 0)
		{
		    goto restore_msr;
		}
	    }
	    
	    if (insn == E500MC_WRTEEI && (PRIVATE_DATA_HYPER(MSR_INSTR) & MSR_EE) &&
		PRIVATE_DATA_HYPER(STEP_PC_ADDR))
	    {
		/* change back the instruction to <wrteei 1> from <wrteei 0> */
	        if (guestmem_out32((uint32_t*)PRIVATE_DATA_HYPER(STEP_PC_ADDR), (uint32_t)(E500MC_WRTEEI | MSR_EE)) != 0)
		{
	    	    goto restore_msr;
		}
	    }
	}
	
	/* check highest-priority exceptions first */
	if (PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS) & MSR_ME)
	{
	    register_t pc, ivpr, ivor1;
	    
	    if (read_gpc(guestFrame, &pc) != 0 || 
		read_gspr(guestFrame, SPR_IVPR, &ivpr) != 0 || 
		read_gspr(guestFrame, SPR_IVOR1, &ivor1) != 0)
	    {
		goto restore_msr;
	    }
		
	    if (pc == ivpr + ivor1)
	    {		
		/* at start of machine check service routine, restore MCSRR1 bits */
		register_t mcsrr1;
		if (read_gspr(guestFrame, SPR_MCSRR1, &mcsrr1) != 0)
		{
		    goto restore_msr;
		}
		
		mcsrr1 |= PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS);
		
		if (write_gspr(guestFrame, SPR_MCSRR1, mcsrr1) != 0)
		{
		    goto restore_msr;
		}
		
		exception = true;
	    }
	}
	else /* see if we're in a non-machine-check exception */
	{
	    register_t dbsr;
	    register_t mask = DBCR0_ICMP | DBCR0_IRPT | DBCR0_CIRPT;
	    	    
	    /* read debug status register */
	    if (read_gspr(guestFrame, SPR_DBSR, &dbsr) != 0)
	    {
		goto restore_msr;
	    }
	    
	    if (dbsr & DBCR0_CIRPT)
	    {
		/* at start of critical exception service routine, restore CSSR1 bits */
		register_t csrr1;
		if (read_gspr(guestFrame, SPR_CSRR1, &csrr1) != 0)
		{
		    goto restore_msr;
		}
	    
		csrr1 |= PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS);
		
		if (write_gspr(guestFrame, SPR_CSRR1, csrr1) != 0)
		{
		    goto restore_msr;
		}
		
		exception = true;
	    }
	    
	    if (dbsr & DBCR0_IRPT)
	    {
		/* at start of normal exception service routine, restore SRR1 bits */	
	    	register_t srr1;
		if (read_gspr(guestFrame, SPR_SRR1, &srr1) != 0)
		{
		    goto restore_msr;
		}
	    
		srr1 |= PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS);
		
		if (write_gspr(guestFrame, SPR_SRR1, srr1) != 0)
		{
		    goto restore_msr;
		}
		    
		exception = true;
	    }
	    
	    /* Clear status register bits */
	    if (write_gspr(guestFrame, SPR_DBSR, mask) != 0)
	    {
		goto restore_msr;
	    }
	}
	
restore_msr:
	/* If the MSR was modified in PreStepOperations(), restore it by resetting modified bits 
	 * Don't restore it if there was an exception or the step instruction was an rfxi
	 */
	if (PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS) && !exception)
	{
	    msr |= PRIVATE_DATA_HYPER(STEP_MODIF_MSR_BITS);
	    
	    if (write_gmsr(guestFrame, msr, 1) != 0)
	    {
		ClearStepPrivateData();
		return;
	    }
	}
	
	ClearStepPrivateData();
		
	return;
}
