/****************************************************************************/
/*

DESCRIPTION

	Contains the definition of the callbacks used by the hypervisor to enter the agent.

COPYRIGHT

	Copyright 2008-2010 Freescale Semiconductor, Inc.

*/
/****************************************************************************/
#include "trk_prefix.h"

#include "trk_agent.h"
#include "byte_chan_wrapper.h"
#include "dserror.h"
#include "dstypes.h"
#include "nubinit.h"
#include "usr_put.h"
#include "nubevent.h"
#include "notify.h"
#include "targimpl.h"
#include "targimpl_ppc.h"

#include "PrivateData.h"

#include "transport.h"

#include <errors.h>
#include <events.h>
#include <debug-stub.h>
#include <greg.h>

static bool StubHandlesDebugEvent(trapframe_t *trap_frame);
#if TRK_TRANSPORT_INT_DRIVEN
void TargetSetInputPendingPtr( volatile unsigned char *inputPendingPtr );
#endif /* TRK_TRANSPORT_INT_DRIVEN */

#ifndef IS_SMP
extern ProcessorState_PPC 	gCPUState;
#else
#define gCPUState (*(ProcessorState_PPC*)PRIVATE_DATA_HYPER(GUEST_STATE))
#endif


static stub_ops_t attr_debug_stub trk_ops = {
        .compatible      = "trk-stub",
        .vcpu_init       = trk_agent_init,
        .vcpu_start      = trk_agent_vcpu_start,
        .vcpu_stop       = trk_agent_vcpu_stop,
        .debug_interrupt = trk_agent_process_trap,
};

void trk_agent_init(void)
{
	gcpu_t *virtualCpu = get_gcpu();
	TopazTrkVcpuContext *privateData = NULL;

	puts("Starting to initialize hyperTRK agent.");

	ZERO_MALLOC(privateData, TopazTrkVcpuContext);
	if (privateData == NULL)
	{
		TRACE("Error allocating HyperTRK stub private virtual cpu data!");
		return;
	}

	/* set the current virtual cpu debug stub private data */
	virtualCpu->dbgstub_cpu_data = privateData;

	/* check the current guest debug mode */
	if (virtualCpu->guest && (virtualCpu->guest->guest_debug_mode != 0))
	{
		TRACE("User set guest debug mode %d - Debug resources cannot be accesed by the agent", virtualCpu->guest->guest_debug_mode);
	}
	else
	{
	}

#ifdef IS_SMP
	/* initialize private data if any */
	InitializePrivateData();

	/* initialize guest event handler */
	if ( (PRIVATE_DATA_HYPER(GEVENT_ID) = register_gevent(trk_agent_event_handler)) == ERR_BUSY )
	{
		TRACE("Could not register trk guest event on this core.");
	}

	/* initialize the transport table for this core */
	InitializeTransportTable();
#endif
	/* first we initialize the byte channel and other stuff */
	DSError mainError = InitializeNub();

	if (mainError == kNoError)
	{
		puts("Succesfully initialized hyperTRK agent");
	}
	else
	{
		printf("Could not fully initialize hyperTRK agent. Error: 0x%08X", (unsigned int)mainError);
	}

	/* we are not reseting the first time we launch the hypervisor */
	PRIVATE_DATA_HYPER(IN_RESET) = 0;

	return;
}

void trk_agent_vcpu_start(trapframe_t* startFrame)
{
	if (unlikely (!startFrame))
	{
		__puts("Got an invalid stack frame. Returning.\r\n");
		return;
	}

	#if TRK_TRANSPORT_INT_DRIVEN
		gInputPending = false;
		TargetSetInputPendingPtr( &gInputPending );
	#endif /* TRK_TRANSPORT_INT_DRIVEN */

	/* enable guest MSR[DE] bit */
	register_t gmsr = 0x0;
	read_gmsr(startFrame, &gmsr);
	gmsr |= MSR_DE;
	write_gmsr(startFrame, gmsr, 0);

	/* enable guest DBCR0[IDM] and trap events to go to debug interrupt*/
	register_t dbcr0 = mfspr(SPR_DBCR0);
	dbcr0 |= DBCR0_IDM | DBCR0_TRAP;
	mtspr(SPR_DBCR0, dbcr0);

	EnableByteChannelEvents();

	if (!PRIVATE_DATA_HYPER(IN_RESET))
	{
		__puts("------STARTED THE PARTITION------\r\n");
		NubWelcome();
	}
	else
	{
		__puts("------RESTARTED THE PARTITION------\r\n");
		/* if the agent was reset to debug he will be in a logical
		 * guest stop mode and will await debugger commands
		 */
		PRIVATE_DATA_HYPER(IN_RESET) = 0; /* we finished our soft reset */
				
		/* set the current guest PC value */
		read_gpc(startFrame, &gCPUState.PC);
		
		if (PRIVATE_DATA(TRK_CONNECTED))
		{
		    gcpu_t *virtualCpu = get_gcpu();
		    DoNotifyCoreReleased(virtualCpu->cpu->coreid);
		    
		    /*
             * for the event handler to process host requests we need to have
             * a stopped state; after reset the host will continue the target
             * if necessary or will perform other requests
             */
		    TargetSetStopped(true);
		    trk_agent_event_handler(startFrame);
		}
	}
	
	return;
}

void trk_agent_vcpu_stop( void )
{
	DisableByteChannelEvents();
	PRIVATE_DATA_HYPER(IN_RESET) = true;
}

/*
 * We are in the debug interrupt CE=0, EE=0
 */
int trk_agent_process_trap(trapframe_t *trap_frame)
{
	NubEvent		event;

	/* read the current guest PC value */
	read_gpc(trap_frame, &gCPUState.PC);

	if (!StubHandlesDebugEvent(trap_frame))
	{
		__puts("Got a debug event that is not handled by hyperTRK at 0x");
		__puthex8(gCPUState.PC);
		__puts(". We are returning to hypervisor. \r\n");
		return 1;
	}

	/* clear all hw events */
	register_t dbsr = mfspr(SPR_DBSR);
	mtspr(SPR_DBSR, dbsr);

	ui16 debugType = 0;
	if (dbsr & DBSR_TRAP)
	{
		debugType |= PPC_TRAP_DEBUG_EXCEPTION;
	}
	if (dbsr & DBSR_ICMP)
	{
		debugType |= PPC_ICMP_DEBUG_EXCEPTION;
	}
	if (dbsr & DBSR_IAC1)
	{
		debugType |= PPC_IAC1_DEBUG_EXCEPTION;
	}
	if (dbsr & DBSR_IAC2)
	{
		debugType |= PPC_IAC2_DEBUG_EXCEPTION;
	}
	if (dbsr & DBSR_DAC1R)
	{
		debugType |= PPC_DAC1R_DEBUG_EXCEPTION;
	}
	if (dbsr & DBSR_DAC1W)
	{
		debugType |= PPC_DAC1W_DEBUG_EXCEPTION;
	}
	if (dbsr & DBSR_DAC2R)
	{
		debugType |= PPC_DAC2R_DEBUG_EXCEPTION;
	}
	if (dbsr & DBSR_DAC2W)
	{
		debugType |= PPC_DAC2W_DEBUG_EXCEPTION;
	}
	/* set the current exception type for performance reasons */
	gCPUState.exceptionID = debugType;

	__puts("Got a debug event at 0x");
	__puthex8(gCPUState.PC);
	__puts(". hyperTRK is handling the event. \r\n");

	/* add the interrupt event to the queue */
	if (dbsr & DBSR_ICMP)
	    ConstructEvent(&event, kExceptionEvent);
	else
	    ConstructEvent(&event, kBreakpointEvent);
	
	PostEvent(&event);

	gcpu_t *virtualCpu = get_gcpu();
	int trkEventId = PRIVATE_DATA_HYPER_CPU(GEVENT_ID, virtualCpu);
	if (virtualCpu && (trkEventId > 0))
	{
		setgevent(virtualCpu, trkEventId);
	}

	return 0;
}

void trk_agent_event_handler(trapframe_t *trap_frame)
{
	if (unlikely (!trap_frame))
	{
		__puts("Got an invalid stack frame. Returning.\r\n");
		return;
	}

	/* we do not want to receive any messages while resetting */
	if (PRIVATE_DATA_HYPER(IN_RESET))
	{
		return;
	}
	/* save current trap frame */
	PRIVATE_DATA_HYPER(GUEST_FRAME) = trap_frame;

	while ( 1 )
	{
		EnableByteChannelReceive();

		/* When a new event needs to be processed before
		 * returning to guest we jump in trk main loop
		 */
		NubMainLoop();

		/* returning to guest */
		if (DisableByteChannelReceive())
			return;
	}

}


typedef struct {
	bool active;
	char extension[0];
} ICMPdebugStatus;
static bool StubHandlesDebugEvent(trapframe_t *trap_frame)
{
	/* if it was a trap debug event, IAC or DAC or ICMP
	 */
	/*
	if ( (mfspr(SPR_ESR) == ESR_PTR) ||
			((ICMPdebugStatus*)PRIVATE_DATA_HYPER(STEP_STATUS))->active )
		return true;
	return false;
	*/
	return true;
}
