/****************************************************************************/
/*

DESCRIPTION

    	Target-specific implementation for a PowerPC guest running under a hypervisor

COPYRIGHT

	Copyright 2008-2010 Freescale Semiconductor, Inc.

*/
/****************************************************************************/

/* TRK includes */
#include "trk_prefix.h"

#include "usr_put.h"		// must be included before msgcmd.h
#include "dstypes.h"
#include "dserror.h"
#include "nubassrt.h"
#include "nubevent.h"
#include "notify.h"
#include "target.h"
#include "msg.h"
#include "nubinit.h"
#include "targimpl_ppc.h"
#include "targimpl.h"
#include "support.h"

/* Hyper TRK includes */
#include "hypervisor_guest_access.h" /* TOREMOVE - TRACE debug purposes */
#include "trk_agent.h"
#include "PrivateData.h"

typedef struct {
	bool active;
	DSMessageStepOptions type;
	unsigned count;
	ui32 rangeStart;
	ui32 rangeEnd;
} PpcStepStatus;

static bool PpcTargetCheckStep(NubEventType type);

DSError DoUnsupported(MessageBuffer* b);

/****************************************************************************/
/*
 *	Target program's processor state (exception state)
 */
/****************************************************************************/
#ifndef IS_SMP
ProcessorState_PPC 	gCPUState;
#else
#define gCPUState (*(ProcessorState_PPC*)PRIVATE_DATA_HYPER(GUEST_STATE))
#endif

/****************************************************************************/
/*
 *	TRK processor state (prior to entering user code)
 */
/****************************************************************************/
#ifndef IS_SMP
TRKState_PPC 			gTRKState;
#else
#define gTRKState (*(TRKState_PPC*)PRIVATE_DATA_HYPER(TRK_STATE))
#endif

#ifndef IS_SMP
static PpcStepStatus gStepStatus =
{
	false,
	kDSStepIntoCount,
	0,
	0,
	0
};
#else
#define gStepStatus (*(PpcStepStatus*)PRIVATE_DATA_HYPER(STEP_STATUS))
#endif

#ifndef IS_SMP
tlb_entry_t			gTLBEntry;
#else
#define gTLBEntry (*(tlb_entry_t*)PRIVATE_DATA_HYPER(TLB_ENTRY))
#endif

#ifdef IS_SMP
/****************************************************************************/
/*
 *	InitializePrivateData
 *
 *	this should be called to initialize per core data.
 */
/****************************************************************************/
void InitializePrivateData( void )
{
	PRIVATE_DATA_HYPER(TRK_STATE)   = malloc(sizeof(TRKState_PPC));
	PRIVATE_DATA_HYPER(GUEST_STATE) = malloc(sizeof(ProcessorState_PPC));
	PpcStepStatus *stepStatus       = malloc(sizeof(PpcStepStatus));
	stepStatus->type = kDSStepIntoCount;
	PRIVATE_DATA_HYPER(STEP_STATUS) = stepStatus;
	PRIVATE_DATA_HYPER(TLB_ENTRY)   = malloc(sizeof(tlb_entry_t));
	TargetSetStopped( false );
}
#endif


/****************************************************************************/
/*
 *	ValidMemory32
 *
 *	this should to a memory test without crashing
 */
/****************************************************************************/
DSError ValidMemory32(const void* addr, size_t length, ValidMemoryOptions readWriteable)
{
	// TODO hyper
	return kNoError;
}

/*************************release_secondary_cores***************************************************/
/*
 *	HyperTargetAccessMemory
 *
 *	Read/write memory, testing for accessability
 */
/****************************************************************************/
DSError HyperTargetAccessMemory(void* Data,
						void* virtualAddress,
						size_t* memorySize,
						MemoryAccessOptions accessOptions,
						bool read )
{
	ui64 startAddress;
	MemorySpace space;

	startAddress = (ui32)virtualAddress;

	if (accessOptions == kPhysicalSpace)
	    space = kPhysicalGuest;
	else if (accessOptions == kVirtualDataSpace)
    	    space = kVirtualDataGuest32;
	else
    	    space = kVirtualInstrGuest32;

	if (read)
		read_memory(space, startAddress, (ui32)*memorySize, (byte*)Data);
	else
		write_memory(space, startAddress, (ui32)*memorySize, (byte*)Data);

	return kNoError;
}


/****************************************************************************/
/*
 *	HyperTargetAccessExtendedMemory
 *
 *	Read/write memory, testing for accessability
 */
/****************************************************************************/
DSError HyperTargetAccessExtendedMemory(void* Data,
						void* addressHigh,
						void* addressLow,
						size_t* memorySize,
						MemoryAccessOptions accessOptions,
						bool read )
{
	ui64 startAddress = (ui32)addressHigh;
	MemorySpace space;
	
	startAddress = (startAddress << 32) | (ui32)addressLow;	
	
	if (accessOptions == kPhysicalSpace)
	    space = kPhysicalGuest;
	else
	{
	    /* for virtual access are supported only 32 bit addresses  */
	    if ((ui32)addressHigh > 0)
		return  kParameterError;
	    else
	    {
		if (accessOptions == kVirtualDataSpace)
    	    	    space = kVirtualDataGuest32;
		else
    		    space = kVirtualInstrGuest32;	    
	    }
	}

	if (read)
		read_memory(space, startAddress, (ui32)*memorySize, (byte*)Data);
	else
		write_memory(space, startAddress, (ui32)*memorySize, (byte*)Data);

	return kNoError;
}


/****************************************************************************/
/*
 *	HyperTargetReadInstruction
 *
 *	Read instruction from memory, determining instruction size.  The
 *  storage pointed to by readData must be big enough to store the
 *  largest possible instruction.
 *  Assume when this function is called that the address provided
 *  is in Virtual Guest Space. We will use External PID Load/Store
 *  facility for memory access
 */
/****************************************************************************/
DSError HyperTargetReadInstruction(InstructionType* readData,
								void* virtualAddress )
{
	DSError err = kNoError;
	size_t length = sizeof( InstructionType );

	/*
	 * Determine instruction size and store in destination.
	 * Instructions on PPC are 32 bits.
	 */

	err = XTargetAccessMemory( readData, virtualAddress, &length,
			kVirtualInstructionSpace, true );

	/*
	 * Verify that an entire instruction was read.
	 */
	if (err == kNoError && length != sizeof( InstructionType ))
		err = kInvalidMemory;

	return( err );
}


/****************************************************************************/
/*
 *	HyperTargetWriteInstruction
 *
 *	Writes instruction to memory.
 *	Assume when this function is called that the address provided
 *  is in Virtual Guest Space. We will use External PID Load/Store
 *  facility for memory access
 */
/****************************************************************************/
DSError HyperTargetWriteInstruction(InstructionType *data, void* virtualAddress)
{
	DSError err = kNoError;
	size_t length = sizeof( InstructionType );

	/*
	** Write instruction to memory.
	*/
	err = XTargetAccessMemory( (void *)data, virtualAddress, &length,
			kVirtualInstructionSpace, false );

	/*
	** Verify that an entire instruction was written.
	*/
	if (err == kNoError && length != sizeof( InstructionType ))
		err = kInvalidMemory;

	return( err );
}


/****************************************************************************/
/*
 *	HyperTargetAccessDefault
 *
 *	R/W a range in the default register block.  W/R the registers in the
 *  range directly to/from the given message buffer.  Returns the size of the
 *  requested range in bytes.
 */
/****************************************************************************/
DSError	HyperTargetAccessDefault(unsigned int firstRegister, unsigned int lastRegister,
						  MessageBuffer *b, size_t* registerStorageSize,
						  bool read )
{
	DSError err;
	trapframe_t* guestFrame;
	HyperRegisterInfo* regInfo;
	ui32 buffer[2];

	if (firstRegister < TRK_DEFAULT_MIN_REGISTER ||
		lastRegister > TRK_DEFAULT_MAX_REGISTER)
		return( kInvalidRegister );

	/* Optain the current trap frame */
	guestFrame = GetGuestTrapFrame();
	err = kNoError;

	/* For each register get the register information */
	for (; firstRegister <= lastRegister; firstRegister++)
	{
		if ( ConvertTrkRegisterId(kDSRegistersDefault, firstRegister, &regInfo) )
		{
			if (read == true)
			{
				read_register(guestFrame, regInfo, (byte*)buffer);
				(regInfo->size == kRegister64) ?
					MessageAddBlock_ui64(b, (ui64*)buffer, 1) :
					MessageAddBlock_ui32(b, buffer, 1);
			}
			else
			{
				(regInfo->size == kRegister64) ?
					MessageGetBlock_ui64(b, (ui64*)buffer, 1) :
					MessageGetBlock_ui32(b, buffer, 1);
				write_register(guestFrame, regInfo, (byte*)buffer);
			}
		}
		else
		{
			__puts("Got invalid register id");
			err = kInvalidRegister;
			break;
		}
	}

	return( err );
}

/****************************************************************************/
/*
 *	HyperTargetAccessExtended1
 *
 *	R/W a range in the extended1 register block.  W/R the registers in the
 *  range directly to/from the given message buffer.  Returns the size of the
 *  requested range in bytes.
 */
/****************************************************************************/
DSError	HyperTargetAccessExtended1(unsigned int firstRegister, unsigned int lastRegister,
								MessageBuffer *b, size_t* registerStorageSize,
								bool read )
{
	DSError err = kNoError;
	trapframe_t* guestFrame;
	HyperRegisterInfo* regInfo;
	ui32 buffer[2];

	if (firstRegister < DS_EXTENDED1_MIN_REGISTER ||
		lastRegister > DS_EXTENDED1_MAX_REGISTER)
		return( kInvalidRegister );

	/* Optain the current trap frame */
	guestFrame = GetGuestTrapFrame();

	/* For each register get the register information */
	for (; firstRegister <= lastRegister; firstRegister++)
	{
		if ( ConvertTrkRegisterId(kDSRegistersExtended1, firstRegister, &regInfo) )
		{
			if (read == true)
			{
				read_register(guestFrame, regInfo, (byte*)buffer);
				(regInfo->size == kRegister64) ?
					MessageAddBlock_ui64(b, (ui64*)buffer, 1) :
					MessageAddBlock_ui32(b, buffer, 1);
			}
			else
			{
				(regInfo->size == kRegister64) ?
					MessageGetBlock_ui64(b, (ui64*)buffer, 1) :
					MessageGetBlock_ui32(b, buffer, 1);
				write_register(guestFrame, regInfo, (byte*)buffer);
			}
		}
		else
		{
			__puts("Got invalid register id");
			err = kInvalidRegister;
			break;
		}
	}

	return( err );
}

/****************************************************************************/
/*
 *	HyperTargetAccessFP
 *
 *	R/W a range in the floating point register block.  W/R the registers in the
 *  range directly to/from the given message buffer.  Returns the size of the
 *  requested range in bytes.
 */
/****************************************************************************/
DSError	HyperTargetAccessFP(unsigned int firstRegister, unsigned int lastRegister,
								MessageBuffer *b, size_t* registerStorageSize,
								bool read )
{
	DSError err = kNoError;
	trapframe_t* guestFrame;
	HyperRegisterInfo* regInfo;
	ui32 buffer[2];

	if (firstRegister < TRK_FLOAT_MIN_REGISTER ||
		lastRegister > TRK_FLOAT_MAX_REGISTER)
		return( kInvalidRegister );


	/* Optain the current trap frame */
	guestFrame = GetGuestTrapFrame();

	/* For each register get the register information */
	for (; firstRegister <= lastRegister; firstRegister++)
	{
		if ( ConvertTrkRegisterId(kDSRegistersFP, firstRegister, &regInfo) )
		{
			if (read == true)
			{
				read_register(guestFrame, regInfo, (byte*)buffer);
				(regInfo->size == kRegister64) ?
					MessageAddBlock_ui64(b, (ui64*)buffer, 1) :
					MessageAddBlock_ui32(b, buffer, 1);
			}
			else
			{
				(regInfo->size == kRegister64) ?
					MessageGetBlock_ui64(b, (ui64*)buffer, 1) :
					MessageGetBlock_ui32(b, buffer, 1);
				write_register(guestFrame, regInfo, (byte*)buffer);
			}
		}
		else
		{
			__puts("Got invalid register id");
			err = kInvalidRegister;
			break;
		}
	}

	return( err );
}

/****************************************************************************/
/*
 *	HyperTargetAccessExtended2
 *
 *	R/W a range in the floating point register block.  W/R the registers in the
 *  range directly to/from the given message buffer.  Returns the size of the
 *  requested range in bytes.
 */
/****************************************************************************/
DSError	HyperTargetAccessExtended2(unsigned int firstRegister, unsigned int lastRegister,
								MessageBuffer *b, size_t* registerStorageSize,
								bool read )
{
	DSError err = kNoError;
	trapframe_t* guestFrame;
	HyperRegisterInfo* regInfo;
	ui32 buffer[2];

	if (firstRegister < DS_EXTENDED2_MIN_REGISTER ||
		lastRegister > DS_EXTENDED2_MAX_REGISTER)
		return( kInvalidRegister );

	/* Optain the current trap frame */
	guestFrame = GetGuestTrapFrame();

	/* For each register get the register information */
	for (; firstRegister <= lastRegister; firstRegister++)
	{
		if ( ConvertTrkRegisterId(kDSRegistersExtended2, firstRegister, &regInfo) )
		{
			if (read == true)
			{
				read_register(guestFrame, regInfo, (byte*)buffer);
				(regInfo->size == kRegister64) ?
					MessageAddBlock_ui64(b, (ui64*)buffer, 1) :
					MessageAddBlock_ui32(b, buffer, 1);
			}
			else
			{
				(regInfo->size == kRegister64) ?
					MessageGetBlock_ui64(b, (ui64*)buffer, 1) :
					MessageGetBlock_ui32(b, buffer, 1);
				write_register(guestFrame, regInfo, (byte*)buffer);
			}
		}
		else
		{
			__puts("Got invalid register id");
			err = kInvalidRegister;
			break;
		}
	}

	return( err );
}

/****************************************************************************/
/*
 *	HyperTargetAccessTLBs
 *
 *	Reads the TLBs registers set. This function will read the whole TLB0 or TLB1 set.
 */
/****************************************************************************/
DSError	HyperTargetAccessTLBs(unsigned int firstRegister, unsigned int lastRegister, MessageBuffer *b, bool read, unsigned int tlb)
{
	DSError err = kNoError;
	unsigned int flags;
	tlb_entry_t tlbEntry;
	ui32 buffer[MAS_NR];
	int i;
		
	if ((tlb == 0 && (firstRegister < DS_TLB0_MIN_REGISTER || lastRegister > DS_TLB0_MAX_REGISTER)) ||
	    (tlb == 1 && (firstRegister < DS_TLB1_MIN_REGISTER || lastRegister > DS_TLB1_MAX_REGISTER)))
		return( kInvalidRegister );

	if (firstRegister == 0)
	    memset(&tlbEntry, 0, sizeof(tlbEntry));
	else
	    tlbEntry = gTLBEntry;

	if (tlb == 1)
	    tlbEntry.mas0 |= MAS0_TLB1_SEL;

	/* Read the requested TLBs */
	if (firstRegister == 0)
	    flags = TLB_READ_FIRST;
	else
	    flags = 0;
	
	do
	{
	    if (read_tlb(&tlbEntry, &flags) == ERR_NOTFOUND)
	    {
		buffer[0] = MAS0_TLB1_SEL ? tlb == 1 : 0;
	        buffer[1] = buffer[2] = buffer[3] = buffer[4] = buffer[5] = 0;
		
		for (i = 0; i < MAS_NR ; i++)
		    MessageAddBlock_ui32(b, &buffer[i], 1);
		
		break;
	    }
	    else
	    {
		if (!(tlbEntry.mas1 & MAS1_VALID))
		    continue;
				
	        buffer[0] = tlbEntry.mas0;
	        buffer[1] = tlbEntry.mas1;
	        buffer[2] = tlbEntry.mas2;
	        buffer[3] = tlbEntry.mas3;
	        buffer[4] = tlbEntry.mas7;
		buffer[5] = tlbEntry.mas8;
	    
		for (i = 0; i < MAS_NR ; i++)
		    MessageAddBlock_ui32(b, &buffer[i], 1);
	    }
	    
	    firstRegister++;
	}
	while (firstRegister <= lastRegister);
	
	gTLBEntry = tlbEntry;

	return( err );
}

/****************************************************************************/
/*
 *	TargetVersions
 *
 *	Fill in version struct for this target.
 */
/****************************************************************************/
DSError TargetVersions( DSVersions *versions )
{
	versions->kernelMajor = DS_KERNEL_MAJOR_VERSION;
	versions->kernelMinor = DS_KERNEL_MINOR_VERSION;
	versions->protocolMajor = DS_PROTOCOL_MAJOR_VERSION;
	versions->protocolMinor = DS_PROTOCOL_MINOR_VERSION;

	return kNoError;
}

/****************************************************************************/
/*
 *	TargetSupportMask
 *
 *	Fill in support mask for this target.
 */
/****************************************************************************/
DSError TargetSupportMask( DSSupportMask *mask )
{
	(*mask)[0] = DS_SUPPORT_MASK_00_07;
	(*mask)[1] = DS_SUPPORT_MASK_08_0F;
	(*mask)[2] = DS_SUPPORT_MASK_10_17;
	(*mask)[3] = DS_SUPPORT_MASK_18_1F;
	(*mask)[4] = DS_SUPPORT_MASK_20_27;
	(*mask)[5] = DS_SUPPORT_MASK_28_2F;
	(*mask)[6] = DS_SUPPORT_MASK_30_37;
	(*mask)[7] = DS_SUPPORT_MASK_38_3F;
	(*mask)[8] = DS_SUPPORT_MASK_40_47;
	(*mask)[9] = DS_SUPPORT_MASK_48_4F;
	(*mask)[10] = DS_SUPPORT_MASK_50_57;
	(*mask)[11] = DS_SUPPORT_MASK_58_5F;
	(*mask)[12] = DS_SUPPORT_MASK_60_67;
	(*mask)[13] = DS_SUPPORT_MASK_68_6F;
	(*mask)[14] = DS_SUPPORT_MASK_70_77;
	(*mask)[15] = DS_SUPPORT_MASK_78_7F;
	(*mask)[16] = DS_SUPPORT_MASK_80_87;
	(*mask)[17] = DS_SUPPORT_MASK_88_8F;
	(*mask)[18] = DS_SUPPORT_MASK_90_97;
	(*mask)[19] = DS_SUPPORT_MASK_98_9F;
	(*mask)[20] = DS_SUPPORT_MASK_A0_A7;
	(*mask)[21] = DS_SUPPORT_MASK_A8_AF;
	(*mask)[22] = DS_SUPPORT_MASK_B0_B7;
	(*mask)[23] = DS_SUPPORT_MASK_B8_BF;
	(*mask)[24] = DS_SUPPORT_MASK_C0_C7;
	(*mask)[25] = DS_SUPPORT_MASK_C8_CF;
	(*mask)[26] = DS_SUPPORT_MASK_D0_D7;
	(*mask)[27] = DS_SUPPORT_MASK_D8_DF;
	(*mask)[28] = DS_SUPPORT_MASK_E0_E7;
	(*mask)[29] = DS_SUPPORT_MASK_E8_EF;
	(*mask)[30] = DS_SUPPORT_MASK_F0_F7;
	(*mask)[31] = DS_SUPPORT_MASK_F8_FF;

	return kNoError;
}


/****************************************************************************/
/*
 *	TargetCPUType
 *
 *	Fill in cpu type struct for this target.
 */
/****************************************************************************/
DSError TargetCPUType( DSCPUType *cpuType )
{
	cpuType->cpuMajor  = DS_CPU_MAJOR_TYPE;
	cpuType->cpuMinor  = DS_CPU_MINOR_TYPE;
	cpuType->bigEndian = gBigEndian;
	cpuType->defaultTypeSize   = 0;
	cpuType->fpTypeSize        = 0;
	cpuType->extended1TypeSize = 0;
	cpuType->extended2TypeSize = 0;

	return kNoError;
}


/****************************************************************************/
/*
 *	TargetCheckException
 *
 *  Determine whether an exception has occurred within the DS.  Clears the
 *  state so that successive calls will report no exceptions until another
 *  one occurs.
 *
 *  This function can be used to reset the exception status.
 *
 *  Returns:
 *		kNoError: No exception has occurred
 *      kCWDSException: Exception has occurred.  In this case, a successive
 *			call to TargetAddExceptionInfo will place the information pertaining
 *		 	to the exception in a message buffer.
 */
/****************************************************************************/
DSError TargetCheckException()
{
	DSError result = kNoError;

	return( result );
}


/****************************************************************************/
/*
 *	TargetInterrupt
 *
 *	Interrupt/breakpoint occurred.  Event was posted by the exception handler.
 *	Service the event, sending notification to the host debugger
 */
/****************************************************************************/
DSError TargetInterrupt(NubEvent* event)
{
	DSError result = kNoError;

	switch (event->fType) {
		case kBreakpointEvent:
		case kExceptionEvent:

			/*
			** Don't notify the debugger of the individual trace
			** breakpoints that occur during stepping - just the
			** last one.
			**
			** Just check the step status.  If it is active, don't
			** notify the debugger.  If it is not, then either we
			** are not stepping at all or we just finished, so
			** send out the notification.
			*/

			if (!PpcTargetCheckStep(event->fType))
			{
				/*
				** Notify the debugger that the target process is
				** stopped.  Also, set the target stop status
				** to indicate that it is stopped.
				*/

				TargetSetStopped( true );
				result = DoNotifyStopped( kDSNotifyStopped );
			}
			break;

		default:
			ASSERT(0);			/* not a valid exception */
			break;
	}
	return result;
}

/****************************************************************************/
/*
 *	HyperTargetAddStopInfo
 *
 *  Building NotifyStopped message to send to the host.  Add target-specific
 *  block to the message buffer.
 */
/****************************************************************************/
DSError HyperTargetAddStopInfo( MessageBuffer* buffer )
{
	ui32 instruction;
	DSError result = kNoError;

	// Append the PC.
	result = MessageAdd_ui32( buffer, gCPUState.PC );

	// Get the instruction at the PC and append that.
	if (result == kNoError)
		result = HyperTargetReadInstruction( (void *)&instruction,
									(void *)gCPUState.PC );

	if (result == kNoError)
		result = MessageAdd_ui32( buffer, instruction );

	// Append the exception ID.
	if (result == kNoError)
		result = MessageAdd_ui16( buffer, (ui16)gCPUState.exceptionID );

	return( result );
}

/****************************************************************************/
/*
 *	HyperTargetAddExceptionInfo
 *
 *  Building NotifyException message to send to the host.  Add target-specific
 *  block to the message buffer.  The added information pertains to the
 *  last exception which occurred in TRK.
 */
/****************************************************************************/
DSError HyperTargetAddExceptionInfo( MessageBuffer* buffer )
{
	DSError result = kNoError;

	return( result );
}

/****************************************************************************/
/*
 *	TargetEnableTrace
 *
 *	Enables/disables the ICMP (Instructrion Complete) bit in the MSR of the
 *	target program.  When enabled, it will cause a debug exception after
 *	one instruction is executed.
 */
/****************************************************************************/
DSError TargetEnableTrace( bool enable )
{
	trapframe_t* guestFrame;
	register_t dbcr0 = 0x0;
	guestFrame = GetGuestTrapFrame();

	dbcr0 = mfspr(SPR_DBCR0);
	if (enable)
	{
		/* add ICMP to dbcr0 */
		dbcr0 |= DBCR0_ICMP;
	}
	else
	{
		/* clear ICMP in DBCR0 */
		dbcr0 &= ~DBCR0_ICMP;
	}
	mtspr(SPR_DBCR0, dbcr0);

	__puts("Modified DBCRO:");
	__puthex8(dbcr0);
	__puts("\r\n");

	return kNoError;
}

/****************************************************************************/
/*
 *	HyperTargetTraceDone
 *
 *	Determines whether trace stop conditions are met.
 */
/****************************************************************************/
bool HyperTargetStepDone( void )
{
	bool result = true;

	/*
	** If the current exception is not a trace exception, then
	** there was some other problem with execution and the
	** trace should be aborted.
	**
	** If it was a trace exception, then check the stop conditions.
	*/

	if (gStepStatus.active &&
		((ui16)gCPUState.exceptionID) == PPC_ICMP_DEBUG_EXCEPTION)
	{
		switch (gStepStatus.type)
		{

			/*
			** For single step, the trace is not done as long as the count
			** is larger than 0.
			*/

			case kDSStepIntoCount:
				__puts( "Remaining step count: " );
				__puthex2( (ui8)gStepStatus.count );
				__puts( "\r\n" );
				if (gStepStatus.count > 0)
					result = false;
				break;

			/*
			** For stepping out of a memory range, the trace is not done
			** as long as the PC is in the memory range, where the
			** range includes the start and end points.
			*/

			case kDSStepIntoRange:
				if (gCPUState.PC >= gStepStatus.rangeStart &&
					gCPUState.PC <= gStepStatus.rangeEnd)
					result = false;
				break;

			default:
				break;
		}
	}

	return( result );
}


/****************************************************************************/
/*
 *	HyperTargetDoStep
 *
 *	General-purpose step handler.  Sets up the instruction trace and
 *  returns control to the target program.  Used to initiate the step
 *	and to continue the stepping as long as needed.
 */
/****************************************************************************/
DSError HyperTargetDoStep( void )
{
	DSError err = kNoError;

	PreStepOperations();

	gStepStatus.active = true;

	/*
	** Once the trace is enabled, just continue.  The target program
	** will break after each instruction.
	*/

	TargetEnableTrace( true );

	if (gStepStatus.type == kDSStepIntoCount ||
		gStepStatus.type == kDSStepOverCount)
		gStepStatus.count--;

	/*
	** To continue, just clear the stopped flag and return.  The
	** main loop will begin execution once it runs out of messages
	** to process.
	*/

	TargetSetStopped( false );

	// FIXME: for now we just put a shutdown in the queue
	// we should use the suspend/run core API once it will be available
	NubEvent event;
	ConstructEvent( &event, kShutdownEvent );
	/*
	** Post the event to be handled in the main loop. This will
	** cause the mainloop to exit, before processing other events,
	** since the event queue should be empty as this function is called
	** in Idle(), after the target stopped flag was set to false
	*/
	PostEvent(&event);

	return err;
}


/****************************************************************************/
/*
 *	PpcTargetCheckStep
 *
 *	General-purpose step handler - checks status and continues the
 *	instruction trace if the step is not complete.  Should be called
 *	whenever a breakpoint is hit.
 *
 *	Returns true if the step is still active, false otherwise.
 */
/****************************************************************************/
bool PpcTargetCheckStep(NubEventType type)
{
	/*
	** We hit a breakpoint.  If a step is in-process, continue it
	** or stop it.
	*/

	if (gStepStatus.active)
	{
	
		PostStepOperations();
		
		/*
		** The trace must be cleaned up regardless of the error.
		*/

		TargetEnableTrace( false );

		/*
		** If we're finished, aborting or a breakpoint was hit, de-activate the step.
		** Otherwise, continue with the next trace.
		*/

		if (HyperTargetStepDone() || type == kBreakpointEvent)
			gStepStatus.active = false;
		else
			HyperTargetDoStep();
	}

	return( gStepStatus.active );
}


/****************************************************************************/
/*
 *	HyperTargetSingleStep
 *
 *	Single step through <count> instructions.
 */
/****************************************************************************/
DSError HyperTargetSingleStep( unsigned count, bool stepOver )
{
	DSError err = kNoError;

	/*
	** Just set up the global data and call the general-purpose routine.
	*/

	if (stepOver)
	{
		err = kUnsupportedError;
	}
	else
	{
		gStepStatus.type = kDSStepIntoCount;
		gStepStatus.count = count;

		err = HyperTargetDoStep();
	}

	return( err );
}


/****************************************************************************/
/*
 *	HyperTargetStepOutOfRange
 *
 *	Step until PC is out of specified memory range.
 */
/****************************************************************************/
DSError HyperTargetStepOutOfRange( ui32 start, ui32 end, bool stepOver )
{
	DSError err = kNoError;

	/*
	** Just set up the global data and call the general-purpose routine.
	*/

	if (stepOver)
	{
		err = kUnsupportedError;
	}
	else
	{
		gStepStatus.type = kDSStepIntoRange;
		gStepStatus.rangeStart = start;
		gStepStatus.rangeEnd = end;

		err = HyperTargetDoStep();
	}

	return( err );
}



/****************************************************************************/
/*
 *	HyperTargetGetPC
 *
 *	Return the target PC.
 */
/****************************************************************************/
PCType HyperTargetGetPC()
{
	return( (PCType)gCPUState.PC );
}


/****************************************************************************/
/*
 *	TargetSupportRequest
 *
 *	Service support request from the target program and resume target program
 *	execution.  Currently supports only file read/write requests.
 *
 *	Expects the following input parameters in the target program registers:
 *		r3: command (kDSWriteFile or kDSReadFile)
 *		r4: file_handle (One of the values from DSFileHandle enum)
 *		r5: length * (ptr to # of bytes to read/write)
 *		r6: buffer * (ptr to buffer containing the data to read/write)
 *
 *	Returns the following result values in the target program registers:
 *		r3: result (one of the values from DSIOResult enum)
 */
/****************************************************************************/
#define SUPP_COMMAND_REG		3
#define SUPP_HANDLE_REG			4
#define SUPP_IN_LEN_REG			5
#define	SUPP_BUFFER_REG			6

#define	SUPP_RESULT_REG			3

DSError TargetSupportRequest(NubEvent *event)
{;
	return( kUnsupportedError );
}

/****************************************************************************/
/*
 *	TargetFlushCache
 *
 *	Flushes the cache corresponding to the specified memory range using
 *	the given options.
 *
 */
/****************************************************************************/
DSError TargetFlushCache( ui8 options,
						  void *start, void *end )
{
	// TODO hyper
	if (start < end)
	{
		return( kNoError );
	}
	else
	{
		return( kInvalidMemory );
	}
}

/****************************************************************************/
/*
 *	HyperTargetStopped
 *
 *	Returns true if the target process is stopped, false if it is
 *	temporarily suspended.
 */
/****************************************************************************/
bool HyperTargetStopped( void )
{
	return (bool)gTRKState.stopped;
}

/****************************************************************************/
/*
 *	TargetSetStopped
 *
 *	Sets the stopped status of the target process.  Set to true if
 *	the target process is stopped, false if it is temporarily
 *	suspended.
 */
/****************************************************************************/
void TargetSetStopped( bool stop_state )
{
	gTRKState.stopped = stop_state;
}

/****************************************************************************/
/*
 *	HyperTargetStop
 *
 *	Stops the target.
 */
/****************************************************************************/
DSError HyperTargetStop( void )
{
	/*
	** Just set the stopped state to true.
	** This will prevent the TRK from returning control to the
	** target program.
	*/

	TargetSetStopped( true );

	return( kNoError );
}

#if TRK_TRANSPORT_INT_DRIVEN
/****************************************************************************/
/*
 *	TargetSetInputPendingPtr
 *
 *	Sets the ptr to the input pending status flag for the interrupt-driven
 *	transport.
 */
/****************************************************************************/
void TargetSetInputPendingPtr( volatile unsigned char *inputPendingPtr )
{
	gTRKState.inputPendingPtr = inputPendingPtr;
	gTRKState.inputActivated = (bool)(*inputPendingPtr) == true;
}
#endif

/****************************************************************************/
/*
 *	HyperTargetBlockFileIo
 *
 *	Block/unblock file IO requests from the target.
 *
 */
/****************************************************************************/
DSError HyperTargetBlockFileIo( bool block )
{
	TRACE();
	return( kUnsupportedError );
}

/****************************************************************************/
/*
	HyperTargetNotifyFileInput

	The host has notified CodeWarriorTRK that it has new input for one of
	the target program's input files.

*/
/****************************************************************************/
DSError HyperTargetNotifyFileInput( ui32 handle )
{
	TRACE();
	return( kUnsupportedError );
}

/****************************************************************************/
/*
 *	CleanupTarget
 *
 *	Performs cleanup of target-program state to prepare it for a new
 *	debug session.  Removes any residual state from the previous session,
 *	such as running processes and breakpoints.
 *
 */
/****************************************************************************/
DSError CleanupTarget()
{
	return( kNoError ); /* nothing to do */
}

/****************************************************************************/
/*
 *	TargetConnect
 *
 *	Target-specific actions performed on connect.
 *
 */
/****************************************************************************/
DSError TargetConnect()
{
	return( kNoError );
}

/****************************************************************************/
/*
 *	TargetDisconnect
 *
 *	Target-specific actions performed on disconnect.
 *
 */
/****************************************************************************/
DSError TargetDisconnect()
{
	/*
	** Just set the stopped state to false.
	** This will return the control from TRK to the
	** target program.
	*/

	TargetSetStopped( false );

	return( kNoError );
}

