/****************************************************************************/
/*

DESCRIPTION

	Target-specific single-instruction trace breakpoint implementation for
	PPC.
	
COPYRIGHT

	Copyright 2000-2002, 2004, 2006, 2008, 2009 Freescale Semiconductor, Inc.
	
*/
/****************************************************************************/

/*
** TRK includes
*/
#include "trk_prefix.h"

#include "usr_put.h"		// must be included before msgcmd.h
#include "target.h"
#include "step.h"
#include "break.h"
#include "ppc_instr.h"
#include "targimpl.h"
#include "nubassrt.h"
#include "targimpl_ppc.h"

/* we cannot have tracing facility without the target agent implementing breakpoint support */
#if TRK_SUPPORT_BREAKPOINTS

/****************************************************************************/
/*
 *	TargetSetupTrace
 *
 *	Sets up single-instruction trace using breakpoints.
 *	If status indicates an active step command, the appropriate
 *	step-into/step-over mode is used.  Otherwise, step-into is assumed.
 */
/****************************************************************************/
DSError TargetSetupTrace( StepStatus *status,
							ui32 processId, ui32 threadId )
{
	DSError err = kNoError;
	InstructionType instruction;
	Breakpoint *bp;
	BreakpointId bpId;
	void *address;
	bool branch;
	bool conditional;
	bool link;
	bool bp_next;
	bool bp_dest;
	bool stepOver;
	void *destination;
	int next_offset;
	int instruction_size;
	
	int breakpoint_index = 0;

	__puts( "TargetSetupTraceBreakpoints:" );
	
	/*
	** If a host-generated step command is active, use it. 
	** Otherwise, assume step-into.
	*/

	stepOver = (status->active) ? status->stepOver : false;

	/*
	** PPC can do single-instruction trace without the use of breakpoints.
	** If this is a step-into, then use that.
	** Note: Defining TRK_FORCE_BREAKPOINT_TRACE will force
	** TRK to always use breakpoints for tracing instead.
	*/
#ifndef TRK_FORCE_BREAKPOINT_TRACE
	if (!stepOver)
	{
		return( XTargetSetupTraceNoBreakpoints( status
	#if DS_PROTOCOL == DS_PROTOCOL_RTOS
											, processId, threadId
	#endif
											) );
	}
#endif

	/*
	** Get the instruction at the current PC.
	*/

	address = (void *)XTargetGetPC(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
								processId, threadId
#endif
									);

	err = XTargetAccessInstruction( (void *)&instruction, address, -1,
								kUserMemory, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
								, processId, threadId
#endif
								 );

	/*
	** Determine the size of the instruction.  For PPC it is always
	** 4 bytes.
	*/
	instruction_size = 4;

	if (err == kNoError)
	{
	
		/*
		** Determine if the instruction is a branch.  If so, extract various
		** branch properties needed to set trace breakpoints.
		*/

		branch = DecodeBranch( (void *)&instruction, address,
							   &conditional, &destination,
							   &link, processId, threadId );

		__puts( "  br: " );
		__puthex2( branch );
		__puts( "  cond: " );
		__puthex2( conditional );
		__puts( "  link: " );
		__puthex2( link );
		__puts( "  dest: " );
		__puthex8( (ui32)destination );
		__puts( "\r\n" );

		/*
		** PPC can do single-instruction trace without the use of breakpoints.
		** If this is not a linking branch, we can use that.
		** Note: Defining TRK_FORCE_BREAKPOINT_TRACE in target.h will force
		** TRK to always use breakpoints for tracing instead.
		*/
#ifndef TRK_FORCE_BREAKPOINT_TRACE
		if (!branch || !link)
		{
			return( XTargetSetupTraceNoBreakpoints( status
	#if DS_PROTOCOL == DS_PROTOCOL_RTOS
												, processId, threadId
	#endif
												) );
		}
#endif

		/*
		** Set up the trace breakpoints.
		*/

		bp_next = false;
		bp_dest = false;
		
		if (branch)
		{
			if (conditional || (stepOver && link))
			{
				/*
				** For conditional branches set a breakpoint
				** after the instruction, in case it does not
				** take the branch.  Also, when stepping over
				** linking branches, set a breakpoint after
				** the instruction.
				**
				*/
		
				next_offset = instruction_size;
				bp_next = true;

			}

			if (err == kNoError &&
				(!stepOver || !link) && 
				address != destination)
			{
				/*
				** When not stepping over a linking branch, for all
				** branches set a breakpoint at the destination
				** address.
				**
				** Note: If the destination is the same as the
				** instruction address, then we can't set a break
				** point there.  If we do, then we can't execute
				** the instruction.  A branch like this is most likely
				** a nop, but it could also be an infinite loop.
				*/
		
				bp_dest = true;
			}
							
		}
		else
		{
			/*
			** For non-branch instructions, simply set a breakpoint at
			** the next instruction.
			*/

			next_offset = instruction_size;
			bp_next = true;	
		}
		
		/*
		** If required, set a breakpoint at the next instruction.
		*/
		if (err == kNoError && bp_next)
		{
			err = GetFreeBreakpoint( &bpId, &bp );
			if (err == kNoError)
			{
			
				err = CheckBreakpointExistence( (unsigned int)address + next_offset, 1,
									            false, DS_MSG_BREAK_CONDITION_EXECUTE, 0, 0,
									            processId, threadId );

				if ( err == kBreakpointConflict ) // erase shared breakpoint
				{
					ReleaseBreakpoint( bpId );
					err = kNoError;
					return( err );
				}

				err = SetBreakpoint( bp, (unsigned int)address + next_offset, 1,
									 false, DS_MSG_BREAK_CONDITION_EXECUTE, 0, 0,
									 processId, threadId );
				if (err == kNoError)
					status->traceBreakpoints[ breakpoint_index++ ] = bpId;
				else
					ReleaseBreakpoint( bpId );
			}
		}

		/*
		** If required, set a breakpoint at the branch destination.
		*/
		if (err == kNoError && bp_dest)
		{
			err = GetFreeBreakpoint( &bpId, &bp );
			if (err == kNoError)
			{
				err = CheckBreakpointExistence( (unsigned int)destination, 1,
									            false, DS_MSG_BREAK_CONDITION_EXECUTE, 0, 0,
									            processId, threadId );

				if ( err == kBreakpointConflict ) // erase shared breakpoint
				{
					ReleaseBreakpoint( bpId );
					err = kNoError;
					return( err );
				}
				
				err = SetBreakpoint( bp, (unsigned int)destination, 1,
									 false, DS_MSG_BREAK_CONDITION_EXECUTE, 0, 0,
									 processId, threadId );
				if (err == kNoError)
					status->traceBreakpoints[ breakpoint_index++ ] = bpId;
				else
					ReleaseBreakpoint( bpId );
			}
		}

	}

	return( err );
}

/****************************************************************************/
/*
 *	TargetCleanupTrace
 *
 *	Cleans up any previously setup single-instruction breakpoints.
 */
/****************************************************************************/
DSError TargetCleanupTrace( StepStatus *status,
							ui32 processId, ui32 threadId )
{
	DSError err = kNoError;
	DSError temp_err;
	Breakpoint *bp;
	
	signed int index;

	/*
	** Cleanup any non-breakpoint trace that is set up.
	*/
	err = XTargetCleanupTraceNoBreakpoints( status
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
											, processId, threadId
#endif
											);
	
	/*
	** It is important to clear the breakpoints in the opposite order
	** in which they were set.  If two breakpoints refer to the same
	** address, then the first one will have the original instruction
	** stored.  The second one will have stored the breakpoint instr.
	** from the first one.  If we restore in the opposite order, then
	** the original instruction will always be the last instruction
	** restored to that address.
	*/
	
	for (index = kTraceBreakpointCount - 1;
		 index >= 0;
		 index--)
	{
		/*
		** If there is an error, record it but continue the cleanup.
		*/
	
		if (status->traceBreakpoints[ index ] != kInvalidBreakpointId)
		{
		
			bp = GetBreakpoint( status->traceBreakpoints[ index ] );
			ASSERT( bp != NULL );
			temp_err = ClearBreakpoint( bp );
			ReleaseBreakpoint( status->traceBreakpoints[ index ] );
			if (temp_err != kNoError)
			{
				__puts( "TargetCleanupTrace error: " );
				__puthex8( temp_err );
				__puts( "  bp: " );
				__puthex8( index );
				__puts( "\r\n" );
			
				err = temp_err;
			}
		}
	}
	
	return( err );
}

/****************************************************************************/
/*
 *	DecodeBranch
 *
 *	Determines whether the given instruction is a branch-type
 *	instruction.  If so, determines whether or not it is a
 *	conditional branch, the destination address, and whether
 *	or not it is a linking branch.
 *
 *  If the caller does not wish to know any set of the output
 *	argument values, it may pass NULL for those arguments.
 *
 *  Args:
 *		instruction => ptr to the instruction (input argument)
 *		address => address of the instruction (input argument)
 *
 * 		conditional_ptr => pointer to storage for conditional flag.
 *			May be NULL. (output argument)
 *		destination_ptr => pointer to storage for destination address
 *			May be NULL. (output argument)
 *		link_ptr => pointer to storage for link flag.  May be
 *			NULL. (output argument)
 *		processId => ID of the process in which this instruction
 *			will execute
 *		threadId => ID of the thread in which this instruction will
 *			execute
 *
 *	Returns true if the instruction is a branch-type and false
 *	otherwise.
 */
/****************************************************************************/
bool DecodeBranch( InstructionType *instruction, void *address,
				   bool *conditional_ptr, void **destination_ptr,
				   bool *link_ptr, ui32 processId, ui32 threadId )
{
	DSError err;
	bool branch = false;
	bool conditional = false;
	bool link = false;
	void *destination = NULL;
	unsigned int destreg;
	DefaultType temp;

	__puts( "\r\nDecodeInstruction: " );
	__puthex8( (ui32)*instruction );
	__puts( "\r\n" );
	switch( PPC_OPCODE( *instruction ) )
	{
		case OP_BX:
			branch = true;
			destination = (void *)PPC_BX_DEST( *instruction, address );
			break;

		case OP_BCX:
			branch = true;
			conditional = true;
			destination = (void *)PPC_BCX_DEST( *instruction, address );
			break;

		case OP_BCXX:
			
			/*
			** Need additional testing to determine the actual instruction.  
			** Don't assume anything until it is identified.
			*/
			
			if (IS_BCCTRX( *instruction ))
			{
				branch = true;
				conditional = true;
				destreg = TRK_DEFAULT_CTR;
			}
			else if (IS_BCCLRX( *instruction ))
			{
				branch = true;
				
				if (!IS_PPC_BLR( *instruction ))
				    conditional = true;
				
				destreg = TRK_DEFAULT_LR;
			}
			
			if (branch)
			{
				/*
				** Destination is <destreg> register contents with lower 2 bits masked.
				*/
				err = XTargetAccessDefaultSingle( destreg, &temp, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
												, processId, threadId
#endif
												);
				ASSERT( err == kNoError );
				destination = (void *)((ui32)temp & 0xfffffffc);
			}
							
			break;
	}

	if (branch)
		link = (PPC_B_LK( *instruction ) != 0);

	if (conditional_ptr != NULL)
		*conditional_ptr = conditional;
		
	if (destination_ptr != NULL)
		*destination_ptr = destination;
		
	if (link_ptr != NULL)
		*link_ptr = link;
		
	return( branch );
}

#endif /* TRK_SUPPORT_BREAKPOINTS */
