/******************************************************************************************/
/*

DESCRIPTION

	Main loop for TRK.
	
	Repeatedly handle host- and target-generated events.

COPYRIGHT

	Copyright 1999-2004, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"

#include "usr_put.h"
#include "nubevent.h"
#include "mainloop.h"
#include "nubassrt.h"
#include "dispatch.h"
#include "targimpl.h"
#include "transport.h"
#include "input.h"
#include "msghndlr.h"
#include "support.h"

#ifndef IS_SMP
extern bool gRequestCached;
extern MessageBuffer *gPreviousAck;
extern volatile unsigned char gInputPending;
#else
#define gPreviousAck PRIVATE_DATA(PREVIOUS_ACKNOWLEDGE)
#define gRequestCached PRIVATE_DATA(REQUEST_CACHED)
#define gInputPending PRIVATE_DATA(INPUT_PENDING)
#endif

/******************************************************************************/
/*
	HandleRequestEvent
*/
/******************************************************************************/
void HandleRequestEvent(NubEvent* event)
{
	MessageBuffer*	buffer;
#if TRK_MSG_SEQUENCE_IDS
	MessageBuffer *discardPending;
	
	
	static MessageBufferId discardPendingId = kInvalidMessageBufferId;
#endif

	buffer = GetBuffer(event->fMessageBufferId);
	ASSERT(buffer);

	DispatchMessage(buffer);
	
#if TRK_MSG_SEQUENCE_IDS
	/*
	** If this was a cached request, it has been handled now, so clear the flag.
	*/
	gRequestCached = false;

	/*
	** If the buffer was reused to form the reply, it will be saved in
	** a global previous-ACK location (gPreviousAckId).  Disassociate it from
	** the event so it won't be discarded with the event, but keep a
	** local backup of the ID (discardPendingId).  Once it is no longer
	** referenced from the gPreviousAck variable it can be discarded.
	*/
	if (discardPendingId != kInvalidMessageBufferId)
	{
		discardPending = GetBuffer( discardPendingId );
		if (gPreviousAck != discardPending)
		{
			ReleaseBuffer( discardPendingId );
			discardPendingId = kInvalidMessageBufferId;
		}
	}

	if (gPreviousAck == buffer)
	{
		discardPendingId = event->fMessageBufferId;
		event->fMessageBufferId = kInvalidMessageBufferId;
	}
#endif
}

/******************************************************************************/
/*
	HandleSupportEvent
*/
/******************************************************************************/

void HandleSupportEvent(NubEvent* event)
{
	TargetSupportRequest( event );
}

/******************************************************************************/
/*
	PostContinueEvent
*/
/******************************************************************************/

void PostContinueEvent(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
						ui32 processId, ui32 threadId
#endif
						)
{
	NubEvent event;

	ConstructEvent( &event, kContinueEvent );

#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	event.fProcessId = processId;
	event.fThreadId = threadId;
#endif /* #if DS_PROTOCOL == DS_PROTOCOL_RTOS */

	/*
	** Post the event to be handled in the main loop.
	*/
	PostEvent(&event);
}

/******************************************************************************/
/*
	HandleContinueEvent
*/
/******************************************************************************/

void HandleContinueEvent(NubEvent* event)
{
#if DS_PROTOCOL == DS_PROTOCOL_RTOS

	__console_dump();
	XTargetContinue( event->fProcessId, event->fThreadId );
	__console_dump();

#else /* #if DS_PROTOCOL == DS_PROTOCOL_RTOS */
	#if __MWERKS__
		#pragma unused( event )
	#endif /* __MWERKS__ */
	
	/*
	** Defer until Idle on bare-board.
	*/
	TargetSetStopped( false );
#endif /* #if DS_PROTOCOL == DS_PROTOCOL_RTOS */
}

/****************************************************************************/
/*
 *	NubMainLoop
 */
/****************************************************************************/
void NubMainLoop(void)
{
	NubEvent	event;
	bool		stopRequested = false;
	bool		isIdle = false;
	
	while (!stopRequested) {
		if (GetNextEvent(&event)) {
			/*
			** An event arrived, so get out of the idle state.
			*/
			isIdle = false;
			
			switch (event.fType) {
				case kNullEvent:
					break;

				case kRequestEvent:
					HandleRequestEvent(&event);
					break;
				
				case kShutdownEvent:
					stopRequested = true; 
					break;
					
				case kBreakpointEvent:
				case kExceptionEvent:
					TargetInterrupt(&event);	/* let target-specific code handle it */
					break;
				
				case kSupportEvent:
					HandleSupportEvent(&event);
					break;
					
				case kContinueEvent:
					HandleContinueEvent(&event);
					break;
					
				default:
					break;
			}
			
			DestructEvent(&event);		/* release resources used by the event */
		} else {

			/*
			** Arrival here means there were no pending events.
			*/
#if TRK_TRANSPORT_INT_DRIVEN
			if (!isIdle || gInputPending)
#else /* not TRK_TRANSPORT_INT_DRIVEN */
			if (!isIdle)
#endif /* TRK_TRANSPORT_INT_DRIVEN */
			{
				/*
				** The TRK is not idle yet.  Switch to the idle state,
				** but get input.  The input may spawn new events,
				** which would kick us back out of the idle state.
				*/
			
				isIdle = true;
				GetInput();
			}
			else
			{
				
				/*
				** The TRK is idle.  We only get to this point when
				** no events are pending and all input has been
				** processed.  Wait until something interesting
				** occurs.  Once it does, get out of the idle
				** state and continue with the loop.
				*/

				Idle();
				isIdle = false;
			
			}
		}
	}
}
