/******************************************************************************************/
/*
	
DESCRIPTION

	Message buffers
	
COPYRIGHT

	Copyright 1999-2003, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"

#include "usr_put.h"		// must be included before msgcmd.h
#include "nubassrt.h"
#include "dserror.h"
#include "mutex_TRK.h"
#include "msgbuf.h"
#include "nubinit.h"
#include "mem_TRK.h"
#ifdef IS_SMP
#include "smp.h"
#endif
#ifndef _NO_SYSTEM_INCLUDES
#include <stddef.h>
#endif


/****************************************************************************/
/*
	Local Types and Variables
*/
/****************************************************************************/

#define MAX_DATA_ELEMENT_LENGTH		16	// max 128-bit data elements

/*
** Add one extra byte to the message buffers in case we want to
** put a terminating '\0' on an incoming string value.
*/

typedef struct MessageBuffers {
	MessageBuffer	fBuffers[kMessageBufferCount + 1];
} MessageBuffers;



#ifndef IS_SMP
MessageBuffers gMsgBufs;
#else
#define gMsgBufs (*perCpuMsgBufs)
#endif

/****************************************************************************/
/*
	static function declarations
*/
/****************************************************************************/

static void SetBufferUsed(MessageBuffer* buffer, bool inUse);

/****************************************************************************/
/*
	SetBufferUsed
*/
/****************************************************************************/

static void SetBufferUsed(MessageBuffer* buffer, bool inUse)
{
	ASSERT(buffer);
	buffer->fInUse = inUse;
}

/****************************************************************************/
/*
	InitializeMessageBuffers
*/
/****************************************************************************/

DSError InitializeMessageBuffers(void)
{
	int i;
#ifdef IS_SMP
	MessageBuffers* perCpuMsgBufs = (MessageBuffers*)TRK_malloc(sizeof(MessageBuffers));
	PRIVATE_DATA(MESSAGE_BUFFERS) = perCpuMsgBufs;
#endif

	for (i = 0; i < kMessageBufferCount; i++) {
		InitializeMutex(&gMsgBufs.fBuffers[i].fMutex);
		AcquireMutex(&gMsgBufs.fBuffers[i].fMutex);
		SetBufferUsed(&gMsgBufs.fBuffers[i], false);
		ReleaseMutex(&gMsgBufs.fBuffers[i].fMutex);
	}

	return kNoError;
}

/****************************************************************************/
/*
	TerminateMessageBuffers
*/
/****************************************************************************/

DSError TerminateMessageBuffers(void)
{
#ifdef IS_SMP
	MessageBuffers* perCpuMsgBufs = PRIVATE_DATA(MESSAGE_BUFFERS);
	TRK_free(perCpuMsgBufs);
	PRIVATE_DATA(MESSAGE_BUFFERS) = NULL;
#endif
	return kNoError;
}

/****************************************************************************/
/*
	GetFreeBuffer
*/
/****************************************************************************/

DSError _GetFreeBuffer(MessageBufferId* resultBufferID,
							MessageBuffer** resultBuffer)
{
	DSError result = kMessageBuffersFull;		/* assume failure */
	MessageBufferId i;
	MessageBuffer* b;
	
	*resultBuffer = NULL;
	
	/*
		linear search through the buffers to find one that is not in use
	*/
	
	for (i = 0; i < kMessageBufferCount; i++) {
		b = GetBuffer(i);
		AcquireMutex(&b->fMutex);
		
		if (b->fInUse == false) {
		
			/* initialize buffer */
			ResetBuffer(b, kDontZero);
			SetBufferUsed(b, true);
			
			*resultBuffer = b;
			*resultBufferID = i;
			result = kNoError;
			
			i = kMessageBufferCount;		/* terminate iteration */
		}
		
		ReleaseMutex(&b->fMutex);
	}	

	return result;
}

/****************************************************************************/
/*
	GetBuffer
	
	Get a pointer to a single buffer structure
*/
/****************************************************************************/

MessageBuffer* GetBuffer(MessageBufferId bufferID)
{
#ifdef IS_SMP
	MessageBuffers* perCpuMsgBufs = (MessageBuffers*)PRIVATE_DATA(MESSAGE_BUFFERS);
#endif
	MessageBuffer* result = NULL;
	
	if (bufferID >= 0 && bufferID < kMessageBufferCount) {
		result = &gMsgBufs.fBuffers[bufferID];
	}
	
	return result;
}

/****************************************************************************/
/*
	ReleaseBuffer
	
	Mark the buffer as unused
*/
/****************************************************************************/

void _ReleaseBuffer(MessageBufferId bufferID)
{
#ifdef IS_SMP
	MessageBuffers* perCpuMsgBufs = PRIVATE_DATA(MESSAGE_BUFFERS);
#endif
	MessageBuffer* b;
	
	if (bufferID == kInvalidMessageBufferId)
		return;
	
	if (bufferID >= 0 && bufferID < kMessageBufferCount) {


		b = &gMsgBufs.fBuffers[bufferID];

		AcquireMutex(&b->fMutex);
		SetBufferUsed(b, false);
		ReleaseMutex(&b->fMutex);		

	}
}

/****************************************************************************/
/*
	ResetBuffer
*/
/****************************************************************************/

void ResetBuffer(MessageBuffer* buffer, ResetBufferOption option)
{
	ASSERT(buffer);

	buffer->fLength = 0;					/* reset logical length */
	buffer->fPosition = 0;					/* reset current stream position */

	if (option == kZero) {
		TRK_memset(buffer->fData, 0, kMessageBufferSize);
	}
}

/****************************************************************************/
/*
	SetBufferPosition
*/
/****************************************************************************/

DSError SetBufferPosition(MessageBuffer* buffer, unsigned int position,
							bool truncate)
{
	DSError err = kNoError;

	ASSERT(buffer);
	
	if (position > kMessageBufferSize) {

		err = kMessageBufferOverflow;

	} else {

		buffer->fPosition = position;
		if (truncate || position > buffer->fLength) {
			buffer->fLength = position;
		}
	}
			
	return err;
}


/****************************************************************************/
/*
	GetBufferRemaining
*/
/****************************************************************************/

unsigned int GetBufferRemaining(MessageBuffer* buffer)
{
	return( kMessageBufferSize - buffer->fPosition );
}

/****************************************************************************/
/*
 *	AppendBuffer
 */
/****************************************************************************/

DSError AppendBuffer(MessageBuffer* buffer, const void* data, size_t length )
{
	DSError			err = kNoError;
	unsigned int	remaining;

	ASSERT(buffer);							/* buffer must be valid */
	ASSERT(data);							/* data pointer must be valid */
	ASSERT(buffer->fInUse);					/* buffer must be allocated */

	if (length == 0)
		return err;

	/*
	 *	Check the remaining space in buffer
	 */
	remaining = kMessageBufferSize - buffer->fPosition;
	
	if (remaining < length) {				/* limit the copy amount */
		err = kMessageBufferOverflow;
		length = remaining;
	}

	/*
	** Optimized for the 1-byte case.  This helps the input loop
	** deal with high-speed communications.
	*/
	
	if (length == 1)
		buffer->fData[buffer->fPosition] = *(unsigned char *)data;
	else
		TRK_memcpy(buffer->fData + buffer->fPosition, data, length);
	
	buffer->fPosition += length;
	buffer->fLength = buffer->fPosition;
	
	ASSERT(buffer->fPosition >= 0 && buffer->fPosition <= kMessageBufferSize);
		// Warning : unsigned comparison with 0 always true: '>="
	
	return err;
}

/****************************************************************************/
/*
 *	ReadBuffer
 */
/****************************************************************************/
DSError ReadBuffer(MessageBuffer* buffer, void* data, size_t length )
{
	DSError			err = kNoError;
	unsigned int	remaining;

	ASSERT(buffer);							/* buffer must be valid */
	ASSERT(data);							/* data pointer must be valid */
	ASSERT(buffer->fInUse);					/* buffer must be allocated */
	
	if (length == 0)
		return err;

	/*
	 *	Check the remaining space in buffer
	 */
	remaining = buffer->fLength - buffer->fPosition;
	
	if (length > remaining) {				/* limit the copy amount */
		err = kMessageBufferUnderflow;
		length = remaining;
	}
	
	TRK_memcpy(data, buffer->fData + buffer->fPosition, length);
	
	buffer->fPosition += length;
	
	ASSERT(buffer->fPosition >= 0 && buffer->fPosition <= kMessageBufferSize);
		// Warning : unsigned comparison with 0 always true: '>="
	
	return err;
}

/****************************************************************************/
/*
 *	ByteSwap
 */
/****************************************************************************/
void ByteSwap( ui8 *dest, const ui8 *src, size_t length )
{
	int front;
	int back;
	
	for (front = 0, back = length - 1; front < length; front++, back--)
		dest[ front ] = src[ back ];
}

/****************************************************************************/
/*
 *	AppendBuffer1_ui16
 */
/****************************************************************************/

DSError AppendBuffer1_ui16(MessageBuffer* buffer, const ui16 data)
{
	return( AppendBuffer_uix( buffer, (const void *)&data, sizeof( data ), 1 ) );
}
	
/****************************************************************************/
/*
 *	AppendBuffer1_ui32
 */
/****************************************************************************/

DSError AppendBuffer1_ui32(MessageBuffer* buffer, const ui32 data)
{
	return( AppendBuffer_uix( buffer, (const void *)&data, sizeof( data ), 1 ) );
}


/****************************************************************************/
/*
 *	AppendBuffer1_ui64
 */
/****************************************************************************/

DSError AppendBuffer1_ui64(MessageBuffer* buffer, const ui64 data)
{
	return( AppendBuffer_uix( buffer, (const void *)&data, sizeof( data ), 1 ) );
}


/****************************************************************************/
/*
 *	AppendBuffer1_ui80
 */
/****************************************************************************/

DSError AppendBuffer1_ui80(MessageBuffer* buffer, const ui80 data)
{
	return( AppendBuffer_uix( buffer, (const void *)&data, sizeof( data ), 1 ) );
}


/****************************************************************************/
/*
 *	AppendBuffer1_ui128
 */
/****************************************************************************/

DSError AppendBuffer1_ui128(MessageBuffer* buffer, const ui128 data)
{
	return( AppendBuffer_uix( buffer, (const void *)&data, sizeof( data ), 1 ) );
}


/****************************************************************************/
/*
 *	AppendBuffer_uix
 */
/****************************************************************************/
DSError AppendBuffer_uix( MessageBuffer* buffer, const void *data, 
							size_t length, int count )
{
	DSError err;
	int offset;
	const void *bigEndianData;
	ui8 swapBuffer[ MAX_DATA_ELEMENT_LENGTH ];
	
	ASSERT( length <= MAX_DATA_ELEMENT_LENGTH );
	
	/*
	** Translate data here based on target-endianness
	*/

	for (offset = 0, err = kNoError;
		 err == kNoError && offset < (count * length);
		 offset += length)
	{
		if (gBigEndian)
		{
			bigEndianData = (const ui8 *)data + offset;
		}
		else
		{
			bigEndianData = (const void *)swapBuffer;
			ByteSwap( swapBuffer, (const ui8 *)data + offset, length );
		}
		err = AppendBuffer( buffer, bigEndianData, length );
	}
		
	return( err );
}



/****************************************************************************/
/*
 *	AppendBuffer_ui8
 */
/****************************************************************************/
DSError AppendBuffer_ui8( MessageBuffer* buffer, const ui8 *data, int count )
{
	return( AppendBuffer( buffer, (const void *)data, count ) );
}

	
/****************************************************************************/
/*
 *	AppendBuffer_ui16
 */
/****************************************************************************/
DSError AppendBuffer_ui16( MessageBuffer* buffer, const ui16 *data, int count )
{
	return( AppendBuffer_uix( buffer, (const void *)data, sizeof( *data ),
			count ) );
}

	
/****************************************************************************/
/*
 *	AppendBuffer_ui32
 */
/****************************************************************************/
DSError AppendBuffer_ui32( MessageBuffer* buffer, const ui32 *data, int count )
{
	return( AppendBuffer_uix( buffer, (const void *)data, sizeof( *data ),
			count ) );
}

/****************************************************************************/
/*
 *	AppendBuffer_ui64
 */
/****************************************************************************/
DSError AppendBuffer_ui64( MessageBuffer* buffer, const ui64 *data, int count )
{
	return( AppendBuffer_uix( buffer, (const void *)data, sizeof( *data ),
			count ) );
}

/****************************************************************************/
/*
 *	AppendBuffer_ui80
 */
/****************************************************************************/
DSError AppendBuffer_ui80( MessageBuffer* buffer, const ui80 *data, int count )
{
	return( AppendBuffer_uix( buffer, (const void *)data, sizeof( *data ),
			count ) );
}

/****************************************************************************/
/*
 *	AppendBuffer_ui128
 */
/****************************************************************************/
DSError AppendBuffer_ui128( MessageBuffer* buffer, const ui128 *data, int count )
{
	return( AppendBuffer_uix( buffer, (const void *)data, sizeof( *data ),
			count ) );
}
	
/****************************************************************************/
/*
 *	ReadBuffer1_ui8
 */
/****************************************************************************/

DSError ReadBuffer1_ui8( MessageBuffer* buffer, ui8 *data )
{
	return( ReadBuffer( buffer, (void *)data, 1 ) );
}

/****************************************************************************/
/*
 *	ReadBuffer1_ui16
 */
/****************************************************************************/
DSError ReadBuffer1_ui16( MessageBuffer* buffer, ui16 *data )
{
	return( ReadBuffer_uix( buffer, (void *)data, sizeof( *data ), 1 ) );
}


/****************************************************************************/
/*
 *	ReadBuffer1_ui32
 */
/****************************************************************************/

DSError ReadBuffer1_ui32( MessageBuffer* buffer, ui32 *data )
{
	return( ReadBuffer_uix( buffer, (void *)data, sizeof( *data ), 1 ) );
}

/****************************************************************************/
/*
 *	ReadBuffer1_ui64
 */
/****************************************************************************/

DSError ReadBuffer1_ui64( MessageBuffer* buffer, ui64 *data )
{
	return( ReadBuffer_uix( buffer, (void *)data, sizeof( *data ), 1 ) );
}

/****************************************************************************/
/*
 *	ReadBuffer1_ui80
 */
/****************************************************************************/

DSError ReadBuffer1_ui80( MessageBuffer* buffer, ui80 *data )
{
	return( ReadBuffer_uix( buffer, (void *)data, sizeof( *data ), 1 ) );
}

/****************************************************************************/
/*
 *	ReadBuffer1_ui128
 */
/****************************************************************************/

DSError ReadBuffer1_ui128( MessageBuffer* buffer, ui128 *data )
{
	return( ReadBuffer_uix( buffer, (void *)data, sizeof( *data ), 1 ) );
}

/****************************************************************************/
/*
 *	ReadBuffer_uix
 */
/****************************************************************************/
DSError ReadBuffer_uix( MessageBuffer* buffer, void *data, 
						size_t length, int count )
{
	DSError err;
	int offset;
	ui8 swapBuffer[ MAX_DATA_ELEMENT_LENGTH ];
	
	ASSERT( length <= MAX_DATA_ELEMENT_LENGTH );

	for (offset = 0, err = kNoError;
		 err == kNoError && offset < (count * length);
		 offset += length)
	{
		if (gBigEndian)
		{
			err = ReadBuffer( buffer, (ui8 *)data + offset, length );
		}
		else
		{
			err = ReadBuffer( buffer, swapBuffer, length );
			if (err == kNoError)
				ByteSwap( (ui8 *)data + offset, swapBuffer, length );
		}
	}

	return( err );
}

/****************************************************************************/
/*
 *	ReadBuffer_ui8
 */
/****************************************************************************/
DSError ReadBuffer_ui8( MessageBuffer* buffer, ui8 *data, int count )
{
	return( ReadBuffer( buffer, (void *)data, count ) );
}

/****************************************************************************/
/*
 *	ReadBuffer_ui16
 */
/****************************************************************************/
DSError ReadBuffer_ui16( MessageBuffer* buffer, ui16 *data, int count )
{
	return( ReadBuffer_uix( buffer, (void *)data, sizeof( *data ),
			count ) );
}

/****************************************************************************/
/*
 *	ReadBuffer_ui32
 */
/****************************************************************************/
DSError ReadBuffer_ui32( MessageBuffer* buffer, ui32 *data, int count )
{
	return( ReadBuffer_uix( buffer, (void *)data, sizeof( *data ),
			count ) );
}

/****************************************************************************/
/*
 *	ReadBuffer_ui64
 */
/****************************************************************************/
DSError ReadBuffer_ui64( MessageBuffer* buffer, ui64 *data, int count )
{
	return( ReadBuffer_uix( buffer, (void *)data, sizeof( *data ),
			count ) );
}

/****************************************************************************/
/*
 *	ReadBuffer_ui80
 */
/****************************************************************************/
DSError ReadBuffer_ui80( MessageBuffer* buffer, ui80 *data, int count )
{
	return( ReadBuffer_uix( buffer, (void *)data, sizeof( *data ),
			count ) );
}

/****************************************************************************/
/*
 *	ReadBuffer_ui128
 */
/****************************************************************************/
DSError ReadBuffer_ui128( MessageBuffer* buffer, ui128 *data, int count )
{
	return( ReadBuffer_uix( buffer, (void *)data, sizeof( *data ),
			count ) );
}

	
