/******************************************************************************************/
/*

DESCRIPTION
	Message buffers for TRK.
	
	Message buffers are allocated out of a pool of buffers, preallocated
	at initialization.  MessageBuffers are structs that contain a
	synchronization primitive for concurrent access, position and size
	information, as well as the memory for the buffer itself.
	
	Usage
	
		Call InitializeMessageBuffers once to initialize the buffers. When a
		new buffer is desired, call "GetFreeBuffer" and a buffer ID plus a
		pointer to the MessageBuffer struct is returned.  Call "ResetBuffer"
		before using the buffer.  When using the MessageBuffer, acquire and
		release the mutex before and after using or modifying it.  Once the
		buffer is no longer being used, call ReleaseBuffer.
		To reset the buffer, clearing its contents, call ResetBuffer.  To
		add data to the buffer, advancing the position and size information,
		call AppendBuffer.
	
	Initialization
	
		DSError InitializeMessageBuffers(void);
		
			Call once at the beginning of the program.
	
	Getting and Releasing
	
		DSError GetFreeBuffer(MessageBufferId* returnedBufferID,
									MessageBuffer** buffer);
		MessageBuffer* GetBuffer(MessageBufferId bufferID);
		
		
			GetFreeBuffer will acquire a buffer if it is available.
			Returned values:
			kNoError	the returnedBufferID and buffer pointers refer
						to the allocated buffer
						
			kMessageBuffersFull
						no more buffers are available.
			
			GetBuffer gets a pre-allocated MessageBuffer pointer from the
			buffer's ID.  If the buffer was not allocated or the ID is out
			of range, GetBuffer retuns NULL.
			
			
		void ReleaseBuffer(MessageBufferId id);
			
			When a buffer is no longer needed, it can be returned to
			the storage pool.
		
			
	Operations on MessageBuffers
		
		DSError SetBufferPosition(MessageBuffer* buffer, size_t position,
									bool truncate);
			
			Set the buffer to the given "position". Advances "length" if
			greater than position.  Forces "length" to equal position if
			truncate is true.  If the buffer length is
			exceeded, kMessageBufferOverflow is returned. 
			Mutex must be acquired before entering.
			
		DSError AppendBuffer(MessageBuffer* buffer, const void* data,
								size_t length );
		
			Append "length" bytes, starting at "data".  Buffer is filled until
			full. If the buffer length is exceeded, kMessageBufferOverflow is
			returned. Mutex must be acquired before entering.
		DSError ReadBuffer(MessageBuffer* buffer, void* data,
								size_t length );
		
			Read "length" bytes, into "data".  Buffer is read until
			exhausted. If the request exceeds available data,
			kMessageBufferUnderflow is returned.
			Mutex must be acquired before entering.
			
COPYRIGHT

	Copyright 1999-2002, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#ifndef _MSGBUF_H_
#define _MSGBUF_H_

/****************************************************************************/
/*
 *	Includes
 */
/****************************************************************************/
#ifndef _NO_SYSTEM_INCLUDES
#include <stddef.h>
#endif
#include "trk_prefix.h"
#include "usr_put.h"		// must be included before msgcmd.h
#include "dstypes.h"
#include "dserror.h"
#include "mutex_TRK.h"
#include "target.h"
#include "msgcmd.h"

/*
** Set this to 1 to get debug messages showing the get/release calls
** for the message buffers.  This helps to track down leaks.
*/
#define DEBUG_MSGBUF_USAGE		(0)

#if DEBUG_MSGBUF_USAGE && (defined( DEBUGIO_SERIAL ) || defined( DEBUGIO_RAM ) || defined( DEBUGIO_CUSTOM ))
	#define GetFreeBuffer( id, b )		( __puts( "Calling GetFreeBuffer from " ), __puts( __FILE__ ),   __putchar( ':' ),   __puthex8( __LINE__ ),  __puts( "\r\n" ), _GetFreeBuffer( id, b ) )
	#define ReleaseBuffer( id )			( __puts( "Calling ReleaseBuffer from " ),   __puts( __FILE__ ), __putchar( ':' ), __puthex8( __LINE__ ),   __puts( "\r\n" ), _ReleaseBuffer( id ) )
#else /* #if DEBUG_MSGBUF_USAGE */
	#define GetFreeBuffer( id, b )		_GetFreeBuffer( id, b )
	#define ReleaseBuffer( id )			_ReleaseBuffer( id )
#endif /* #if DEBUG_MSGBUF_USAGE */

/****************************************************************************/
/*
 *	Errors
 */
/****************************************************************************/
enum {								/* DSErroror; see dserror.h */
	kMessageBuffersFull = 0x0300,
	kMessageBufferOverflow,			/* AppendBuffer failed */
	kMessageBufferUnderflow			/* ReadBuffer failed */
};

/****************************************************************************/
/*
 *	Types
 */
/****************************************************************************/

/*
 *	MessageBufferId
 */
typedef int MessageBufferId;
enum {
	kInvalidMessageBufferId = -1
};

/*
** Total number of buffers.  A basic, bare-board implementation
** needs three.
**
** If sequence IDs are used, an additional buffer is required
** to store the previous reply.
**
** Add one buffer when using the debug RAM console dump feature.
**
** If a particular target requires more, it should define
** TRK_EXTRA_BUFFER_COUNT in target.h.
*/
#ifndef TRK_EXTRA_BUFFER_COUNT
	#define TRK_EXTRA_BUFFER_COUNT 0
#endif

#if defined(DEBUGIO_RAM) && defined(DB_RAM_CONSOLE_DUMP)
	#define DB_RAM_BUFFERS					1
#else
	#define DB_RAM_BUFFERS					0
#endif	

#if TRK_MSG_SEQUENCE_IDS
	#define TRK_MSG_SEQUENCE_ID_BUFFERS		1
#else
	#define TRK_MSG_SEQUENCE_ID_BUFFERS		0
#endif

#define kMessageBufferCount (3 + DB_RAM_BUFFERS + TRK_MSG_SEQUENCE_ID_BUFFERS + TRK_EXTRA_BUFFER_COUNT)

/*
 *	struct MessageFields
 */
#define kMessageFieldCount (DS_MAXMESSAGEFIELDCOUNT)
typedef ui32 MessageFieldType;
typedef struct MessageFields {
	MessageFieldType fData[kMessageFieldCount];
} MessageFields;

/*
 *	struct MessageBuffer
 */
#define kMessageBufferSize (DS_MAXMESSAGESIZE)	/* 2K + overhead, usually */
typedef struct MessageBuffer {
	DSMutex			fMutex;					/* synchronization */
	bool			fInUse;					/* true if buffer is not free */
	unsigned int	fLength;				/* logical end of buffer */
	unsigned int	fPosition;				/* current index in buffer */
	unsigned char	fData[kMessageBufferSize];
	MessageFields	fFields;				/* Values of message fields */
} MessageBuffer;

/*
 *	Options for the ResetBuffer function
 */
typedef enum ResetBufferOption {
	kZero,		/* zero out the buffer */
	kDontZero
} ResetBufferOption;

/****************************************************************************/
/*
 *	Public Function Prototypes
 */
/****************************************************************************/
__extern_c
	/*
	 *	MessageBuffer management
	 */
	DSError InitializeMessageBuffers(void);
	DSError TerminateMessageBuffers(void);
	MessageBuffer* GetBuffer(MessageBufferId bufferID);
	DSError _GetFreeBuffer(MessageBufferId* returnedBufferID, MessageBuffer** buffer);
	void _ReleaseBuffer(MessageBufferId id);
	/*
	 * operations on MessageBuffers
	 */
	void ResetBuffer(MessageBuffer* buffer, ResetBufferOption option);
	DSError SetBufferPosition(MessageBuffer* buffer, unsigned int bufferPosition,
								bool truncate);
	unsigned int GetBufferRemaining(MessageBuffer* buffer);
	DSError AppendBuffer(MessageBuffer* buffer, const void* data, size_t length);
	DSError ReadBuffer(MessageBuffer* buffer, void* data, size_t length);
	
	void ByteSwap( ui8 *dest, const ui8 *src, size_t length );
	/*
	** This ui8 one is a special case and can be short-circuited.
	** This optimization helps the input loop deal with high-
	** speed communications.
	*/
#ifdef _ENTERPRISE_C_
	static DSError AppendBuffer1_ui8(MessageBuffer* buffer, const ui8 data)
#else
	static inline DSError AppendBuffer1_ui8(MessageBuffer* buffer, const ui8 data)
#endif
	{
#ifdef _ENTERPRISE_C_
	#pragma inline
#endif
		if (buffer->fPosition >= kMessageBufferSize)
			return( kMessageBufferOverflow );
			
		buffer->fData[buffer->fPosition++] = data;
		buffer->fLength++;
		return( kNoError );
	}
	DSError AppendBuffer1_ui16(MessageBuffer* buffer, const ui16 data);
	DSError AppendBuffer1_ui32(MessageBuffer* buffer, const ui32 data);
	DSError AppendBuffer1_ui64(MessageBuffer* buffer, const ui64 data);
	DSError AppendBuffer1_ui80(MessageBuffer* buffer, const ui80 data);
	DSError AppendBuffer1_ui128(MessageBuffer* buffer, const ui128 data);
	
	DSError AppendBuffer_uix( MessageBuffer* buffer, const void *data, 
							size_t length, int count );
	DSError AppendBuffer_ui8(MessageBuffer* buffer, const ui8 *data, int count );
	DSError AppendBuffer_ui16(MessageBuffer* buffer, const ui16 *data, int count );
	DSError AppendBuffer_ui32(MessageBuffer* buffer, const ui32 *data, int count );
	DSError AppendBuffer_ui64(MessageBuffer* buffer, const ui64 *data, int count );
	DSError AppendBuffer_ui80(MessageBuffer* buffer, const ui80 *data, int count );
	DSError AppendBuffer_ui128(MessageBuffer* buffer, const ui128 *data, int count );
	
	DSError ReadBuffer1_ui8(MessageBuffer* buffer, ui8 *data );
	DSError ReadBuffer1_ui16(MessageBuffer* buffer, ui16 *data );
	DSError ReadBuffer1_ui32(MessageBuffer* buffer, ui32 *data );
	DSError ReadBuffer1_ui64(MessageBuffer* buffer, ui64 *data );
	DSError ReadBuffer1_ui80(MessageBuffer* buffer, ui80 *data );
	DSError ReadBuffer1_ui128(MessageBuffer* buffer, ui128 *data );
	DSError ReadBuffer_uix( MessageBuffer* buffer, void *data, 
						size_t length, int count );
	DSError ReadBuffer_ui8(MessageBuffer* buffer, ui8 *data, int count );
	DSError ReadBuffer_ui16(MessageBuffer* buffer, ui16 *data, int count );
	DSError ReadBuffer_ui32(MessageBuffer* buffer, ui32 *data, int count );
	DSError ReadBuffer_ui64(MessageBuffer* buffer, ui64 *data, int count );
	DSError ReadBuffer_ui80(MessageBuffer* buffer, ui80 *data, int count );
	DSError ReadBuffer_ui128(MessageBuffer* buffer, ui128 *data, int count );
	
	
__end_extern_c
#endif /* _MSGBUF_H_ */
