/******************************************************************************************/
/*

DESCRIPTION

	TRK command handlers.  Default implementation.
	
	To override implementation, create the overriding functions and add it to
	the link list (project window) before this file.
	
COPYRIGHT

	Copyright 1999-2003, 2006, 2007, 2009, 2010 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

/* Common global message descriptors initialized here */
#define TRK_MSGCMD_INIT

#include "trk_prefix.h"

/*
 *	portable includes
 */
#include "usr_put.h"		// must be included before msgcmd.h
#include "nubassrt.h"
#include "nubevent.h"
#include "msg.h"
#include "msghndlr.h"
#include "break.h"
#include "step.h"
#include "msgparse.h"
#include "ethconfig.h"

/*
 *	portable includes for target-specific implementation
 */
#include "targimpl.h"
#include "reset.h"

/*
 *	transport
 */
#include "transport.h"
#include "transport_tbl.h"


/*
** Set this to 1 to see detailed information about normal command
** message traffic between host and target.  (error msgs output via
** __puts() when one of the DEBUGIO_ macros is defined will always be
** emitted regardless of whether DEBUG_MSG_HANDLER is (1) or (0) .)
*/

#define DEBUG_MSG_HANDLER	(1)


#define ACK_TRY_COUNT 3			/* max # tries to try to resend ack */


#ifndef IS_SMP
extern int gCurTransport;
/*
** Each time an ACK is sent in response to a received message, that
** ACK message is saved.  If a duplicate message is received, the
** saved ACK is simply resent rather than re-processing the duplicate.
** Duplicates are detected by matching the saved ACK sequence ID with
** the incoming message sequence ID.
*/
MessageBuffer *gPreviousAck = NULL;
/*
** This tracks the connected status.  Other parts of the system
** can test this value to modify behavior based on whether or not
** the host is currently connected.  This could be used to suppress
** or queue up messages such as host I/O or notifications when not
** connected.
*/
bool gTrkConnected = false;
/*
** This holds the auto download option from the host. 
** For target entities which are able to auto load the targeted elf
** file (e.g HV) on reset, the host debugger may choose to let the
** target perform the download from the internal memory for speed 
** purpose.
*/
bool gTrkAutoDownload = false;
#else
#define gPreviousAck PRIVATE_DATA(PREVIOUS_ACKNOWLEDGE)
#define gTrkConnected PRIVATE_DATA(TRK_CONNECTED)
#define gTrkAutoDownload PRIVATE_DATA(TRK_AUTO_DOWNLOAD)
#endif


#if __MWERKS__
#pragma mark  Standard ACK/NAK
#endif /* __MWERKS__ */
/****************************************************************************/
/*
 *	MessageIntoReply
 *
 *	Modify the receive buffer to become the reply, using the ack sequence ID
 */
/****************************************************************************/
void MessageIntoReply(MessageBuffer* b,
						ui8 ackCmd /* MessageCommandID*/,
						DSReplyError errSentInAck)
{
	ASSERT(b);
	ASSERT(ackCmd == kDSReplyACK || ackCmd == kDSReplyNAK);
	
	b->fFields.fData[ TRK_MSG_COMMAND_INDEX ] = ackCmd;
	/* don't change the sequence ID */
	b->fFields.fData[ TRK_MSG_REPLY_ERROR_INDEX ] = errSentInAck;
	TrkConstructMessage( b, &gTrkMsgReplyHeader );
}


/****************************************************************************/
/*
 *	SendACK
 *
 *	Send an ACK to the sender with retries.
 */
/****************************************************************************/
DSError SendACK(MessageBuffer* b)
{
	DSError			err;
	int				ackTries;
	/*
	 *	Send the acknowledgement
	 */
	err = kNoError;
	ackTries = ACK_TRY_COUNT;		/* try a few times */

#if TRK_MSG_SEQUENCE_IDS
	/*
	** If using sequence IDs, then save the most recent ACK.  If TRK
	** receives a duplicate of the message to which it just replied,
	** it must re-send the stored ACK without re-processing the message.
	*/

	if (b->fFields.fData[ TRK_MSG_COMMAND_INDEX ] == kDSReplyACK)
	{
		gPreviousAck = b;
	}
#endif

	do {
		/* Access top layer write function */
		err = MSG_LAYER_WRITE( TRANSPORT_LAYER( gCurTransport ), b );

		ackTries--;
	} while (err != kNoError && ackTries > 0);
	
	return err;
}


/****************************************************************************/
/*
 *	StandardACK
 *
 *	Send an ACK to the sender.  Reuse the receive buffer for sending.
 */
/****************************************************************************/
DSError StandardACK(MessageBuffer* b,
						ui8 ackCmd /* MessageCommandID*/,
						DSReplyError errSentInAck)
{
	ASSERT(b);
	ASSERT(ackCmd == kDSReplyACK || ackCmd == kDSReplyNAK);
	
	/*
	 *	convert the received message into a reply
	 */
	MessageIntoReply(b, ackCmd, errSentInAck);
	
	return SendACK(b);
}


#if __MWERKS__
#pragma mark -
#pragma mark  Non-Commands
#endif /* __MWERKS__ */
/****************************************************************************/

DSError DoError(MessageBuffer* b)
{
#if __MWERKS__
	#pragma unused (b)
#endif /* __MWERKS__ */
	#if DEBUG_MSG_HANDLER
		__puts("DoError\r\n");
	#endif
	/* incorrect command */
	return kNoError;
}

DSError DoUnsupported(MessageBuffer* b)
{
	#if DEBUG_MSG_HANDLER
		__puts("DoUnsupported\r\n");
	#endif
	return StandardACK(b, kDSReplyACK, kDSReplyUnsupportedCommandError);
}

#if __MWERKS__
#pragma mark -
#pragma mark  Meta Commands
#endif /* __MWERKS__ */
/****************************************************************************/
/*
 *	meta command set
 */
/****************************************************************************/

/****************************************************************************/
/*
 *	DoPing
 */
/****************************************************************************/
DSError DoPing(MessageBuffer* b)
{
	#if DEBUG_MSG_HANDLER
		__puts("DoPing\r\n");
	#endif

	/*
	** Nothing but ACK.
	*/

	return StandardACK(b, kDSReplyACK, kDSReplyNoError);
}

/****************************************************************************/
/*
 *	DoConnect
 */
/****************************************************************************/
DSError DoConnect(MessageBuffer* b)
{
	DSError err = kNoError;
	int index;
	
	#if DEBUG_MSG_HANDLER
		__puts("DoConnect\r\n");
	#endif

	/*
	** When establishing a new debug session, clean up anything
	** left over from the previous session (in case the previous
	** session was not properly disconnected).
	*/

	CleanupTarget();
	gTrkConnected = true;
	
	/*
	** Close all communication channels except the current one.
	** This should allow the target program to use all of the
	** ports which TRK is not using.
	*/
	for(index = 0; 
	    index < gTrkTransportNumCommPorts && err == kNoError; 
	    index++)
	{
		if (index != gCurTransport)
			TransportTerminateIndex( index );
	}
	
	TargetConnect();
	return StandardACK(b, kDSReplyACK, kDSReplyNoError);
}

/****************************************************************************/
/*
 *	DoDisconnect
 */
/****************************************************************************/
DSError DoDisconnect(MessageBuffer* b)
{
	DSError 	err = kNoError;
	int index;
	
	/* check received message size */
	if (b->fLength < DS_MIN_MSG_DISCONNECT_LENGTH) {
		return StandardACK(b, kDSReplyACK, kDSReplyPacketSizeError);
	}
	
	err = StandardACK(b, kDSReplyACK, kDSReplyNoError);
	
	if (err == kNoError) {
		
		#if DEBUG_MSG_HANDLER
			__puts("DoDisconnect\r\n");
		#endif

		/*
		** Remove target state associated with the current session.
		*/
		
		CleanupTarget();
		gTrkConnected = false;
		TargetDisconnect();
		/*
		** retsore all communication channels except the current one.
		*/
		for(index = 0; 
		    index < gTrkTransportNumCommPorts && err == kNoError; 
	    	index++)
		{
			if (index != gCurTransport)
				TransportInitializeIndex( index );
		}
	}
	return err;
}


/****************************************************************************/
/*
 *	DoReset
 */
/****************************************************************************/
DSError DoReset(MessageBuffer* b)
{
#if __MWERKS__
	#pragma unused (b)
#endif /* __MWERKS__ */

	#if DEBUG_MSG_HANDLER
		__puts("DoReset\r\n");
	#endif

	/*
	** Send out an ACK before the reset, since the reset call will not return.
	*/
	
	StandardACK( b, kDSReplyACK, kDSReplyNoError );
	
	/*
	** This call will not return.  It will cause the DS to restart from
	** scratch.
	*/
	
	__reset();
	
	return kNoError;
}


/****************************************************************************/
/*
 *	DoVersions - Return kernel and protocol versions
 */
/****************************************************************************/
DSError DoVersions(MessageBuffer* b)
{
	DSError					err = kNoError;
	DSVersions				versions;

	#if DEBUG_MSG_HANDLER
		__puts("DoVersions\r\n");
	#endif

	/* check received message size */
	if (b->fLength != DS_MIN_MSG_VERSIONS_LENGTH) {
		return StandardACK(b, kDSReplyACK, kDSReplyPacketSizeError);
	}
		

	/*
	** Reuse buffer to write the reply.
	** Append command (ui8), error (ui8) - standard reply header
	*/

	MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );	
		
	/*
	** Append the four version bytes.
	*/
		
	if (err == kNoError)
		err = TargetVersions( &versions );
	if (err == kNoError)
		err = MessageAdd_ui8( b, versions.kernelMajor );
	if (err == kNoError)
		err = MessageAdd_ui8( b, versions.kernelMinor );
	if (err == kNoError)
		err = MessageAdd_ui8( b, versions.protocolMajor );
	if (err == kNoError)
		err = MessageAdd_ui8( b, versions.protocolMinor );
	
		
	if (err != kNoError) {
		return StandardACK(b, kDSReplyACK, kDSReplyCWDSError);
	}

	#if DEBUG_MSG_HANDLER
		__puts("DoVersions: sending version data\r\n");
	#endif
	err = SendACK(b);
	
	if (err != kNoError) {
		__puts("error occurred\r\n");
		__puthex8(err);
		__puts("\r\n");
	}

	return err;
}

/****************************************************************************/
/*
 *	DoSupportMask - Send bit mask indicating which commands are supported.
 */
/****************************************************************************/
DSError DoSupportMask(MessageBuffer* b)
{
	DSError					err = kNoError;
	DSSupportMask			mask;
	

	#if DEBUG_MSG_HANDLER
		__puts("DoSupportMask\r\n");
	#endif

	/* check received message size */
	if (b->fLength != DS_MIN_MSG_SUPPORTMASK_LENGTH) {
		return StandardACK(b, kDSReplyACK, kDSReplyPacketSizeError);
	}
	

	/*
	** Reuse buffer to write the reply.
	** Append command (ui8), error (ui8) - standard reply header
	*/

	MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );	
		
	/*
	** Append support mask.  The support mask is a byte stream, so it is
	** OK to use MessageAddBlock.
	*/

	if (err == kNoError)
		err = TargetSupportMask( &mask );
	
	if (err == kNoError)
		err = MessageAddBlock(b, &mask, sizeof(mask));
		
	/*
	** Append the protocol level.
	*/
	
	if (err == kNoError)
		err = MessageAdd_ui8( b, DS_PROTOCOL );

	if (err != kNoError) {
		return StandardACK(b, kDSReplyACK, kDSReplyCWDSError);
	}

	#if DEBUG_MSG_HANDLER
		__puts("DoSupportMask: sending support mask\r\n");
	#endif

	err = SendACK(b);
	
	if (err != kNoError) {
		__puts("error occurred\r\n");
		__puthex8(err);
		__puts("\r\n");
	}

	return err;
}

/****************************************************************************/
/*
 *	DoCPUType - Return CPU type and attributes.
 */
/****************************************************************************/
DSError DoCPUType(MessageBuffer* b)
{
	DSError					err = kNoError;
	DSCPUType				cpuType;

	#if DEBUG_MSG_HANDLER
		__puts("DoCPUType\r\n");
	#endif

	/* check received message size */
	if (b->fLength != DS_MIN_MSG_CPUTYPE_LENGTH) {
		return StandardACK(b, kDSReplyACK, kDSReplyPacketSizeError);
	}
		

	/*
	** Reuse buffer to write the reply.
	** Append command (ui8), error (ui8) - standard reply header
	*/

	MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );	
		
	/*
	** Append the four version bytes.
	*/
		
	if (err == kNoError)
		err = TargetCPUType( &cpuType );
	if (err == kNoError)
		err = MessageAdd_ui8( b, cpuType.cpuMajor );
	if (err == kNoError)
		err = MessageAdd_ui8( b, cpuType.cpuMinor );
	if (err == kNoError)
		err = MessageAdd_ui8( b, cpuType.bigEndian );
	if (err == kNoError)
		err = MessageAdd_ui8( b, cpuType.defaultTypeSize );
	if (err == kNoError)
		err = MessageAdd_ui8( b, cpuType.fpTypeSize );
	if (err == kNoError)
		err = MessageAdd_ui8( b, cpuType.extended1TypeSize );
	if (err == kNoError)
		err = MessageAdd_ui8( b, cpuType.extended2TypeSize );
	
		
	if (err != kNoError) {
		return StandardACK(b, kDSReplyACK, kDSReplyCWDSError);
	}

	#if DEBUG_MSG_HANDLER
		__puts("DoCPUType: sending version data\r\n");
	#endif

	err = SendACK(b);
	
	if (err != kNoError) {
		__puts("error occurred\r\n");
		__puthex8(err);
		__puts("\r\n");
	}

	return err;
}


/****************************************************************************/
/*
 *	DoConfigTransport
 *
 *	Configure one of the transport channels (currently only ethernet channels
 *	can be configured).
 *
 */
/****************************************************************************/
DSError DoConfigTransport( MessageBuffer *b )
{
	DSError err = kNoError;
	TrkEthConfig config;
	int index;

	/*
	** Parse the next segment of the message.
	*/

	err = TrkParseMessageSegment( b, &gTrkMsgConfigTransport, true );
	

	if (err == kNoError)
	{
		switch (b->fFields.fData[ TRK_MSG_CT_TYPE_INDEX ])
		{
			case TRK_MSG_TRANSPORT_TYPE_ETHERNET:
				/*
				** Only the ethernet case is supported for now.  Get the config values and set
				** them in the global config.
				*/
				err =  TrkParseMessageSegment( b, &gTrkMsgConfigTransportEthernet, true );
				
				if (err == kNoError)
				{
					config.options = b->fFields.fData[ TRK_MSG_CTE_ETHOPT_INDEX ];
					config.address = b->fFields.fData[ TRK_MSG_CTE_ADDRESS_INDEX ];
					config.port = b->fFields.fData[ TRK_MSG_CTE_PORT_INDEX ];
					config.gateway = b->fFields.fData[ TRK_MSG_CTE_GATEWAY_INDEX ];
					config.netmask = b->fFields.fData[ TRK_MSG_CTE_NETMASK_INDEX ];
					TrkSetEthConfig( &config );

					/*
					** Close and re-open the ethernet channel.
					*/
					for(index = 0; 
					    index < gTrkTransportNumCommPorts && err == kNoError; 
					    index++)
					{
						if (TRANSPORT_PORT( index ) == kEthernet)
						{
							TransportTerminateIndex( index );
							TransportInitializeIndex( index );
							break;
						}
					}
		
				}
				break;
				
			default:
				err = kUnsupportedError;
		}
	
	}
	
	/*
	** Reply with status.
	*/
	return( StandardACK(b, kDSReplyACK, TrkStandardReplyError( err )) );
}


/****************************************************************************/
/*
 *	DoConfigAutoDownload
 *
 *	Configures the target to perform an auto download
 *  	of the elf file from the internal memory
 */
/****************************************************************************/
DSError DoConfigAutoDownload(MessageBuffer* b)
{
	DSError	err = kNoError;
	ui8 	msg_options;
	
	#if DEBUG_MSG_HANDLER
		__puts("DoConfigAutoDownload\r\n");
		__puts(" options: ");
		__puthex2(msg_options);
		__puts("\r\n");		
	#endif
	
	SetBufferPosition( b, TRK_MSG_HEADER_LENGTH, false );
	
	err = MessageGet_ui8( b, &msg_options );
	if (err == kNoError)
	{
	    /*
	    ** Reuse buffer to write the reply.
	    ** Append command (ui8), error (ui8) - standard reply header
	    */
	    MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );
	    gTrkAutoDownload = (msg_options == DS_MSG_AUTO_DOWNLOAD_ENABLE) ? true : false;
	}
	else
	{
		__puts("error occurred\r\n");
		__puthex8(err);
		__puts("\r\n");
	}	
	
	#if DEBUG_MSG_HANDLER
		__puts("DoConfigAutoDownload: sending ack\r\n");
	#endif
	
	err = SendACK(b);	
	if (err != kNoError) {
		__puts("error occurred\r\n");
		__puthex8(err);
		__puts("\r\n");
	}

	return err;
}


#if __MWERKS__
#pragma mark -
#pragma mark  State Commands
#endif /* __MWERKS__ */

/****************************************************************************/
/*
 *	target "state" command set
 */
/****************************************************************************/
 
/****************************************************************************/
/*
 *	DoReadMemory
 *
 *	uses > DS_MAXREADWRITELENGTH of stack space
 */
/****************************************************************************/
DSError DoReadMemory(MessageBuffer* b)
{
	DSError					err = kNoError;
	DSReplyError			replyErr = kDSReplyNoError;
	ui8						msg_options;
	ui16					msg_length;
	ui32					msg_startHigh;
	ui32					msg_startLow;
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	ui32					msg_process;
	ui32					msg_thread;
#endif
#if DS_PROTOCOL == DS_PROTOCOL_HYPER
	ui8						msg_extraOptions;
#endif
	
	/* check received message size */
	if (b->fLength != DS_MIN_MSG_READMEMORY_LENGTH) {
		return StandardACK(b, kDSReplyACK, kDSReplyPacketSizeError);
	}
	
	/*
	** Get ReadMemory parameters from the message.
	*/
	
	SetBufferPosition( b, TRK_MSG_HEADER_LENGTH, false );
	if (err == kNoError)
		err = MessageGet_ui8( b, &msg_options );
	if (err == kNoError)
		err = MessageGet_ui16( b, &msg_length );
	if ((err == kNoError) && (msg_options & DS_MSG_MEMORY_EXTENDED))
		err = MessageGet_ui32( b, &msg_startHigh );
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_startLow );
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_process );
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_thread );
#endif
#if DS_PROTOCOL == DS_PROTOCOL_HYPER
	if (err == kNoError)
		err = MessageGet_ui8( b, &msg_extraOptions );
#endif


	#if DEBUG_MSG_HANDLER
		__puts("DoReadMemory\r\n");
		__puts(" options: ");
		__puthex2(msg_options);
		__puts("\r\n length: ");
		__puthex4(msg_length);
		if (msg_options & DS_MSG_MEMORY_EXTENDED)
		{
		    __puts("\r\n addrHigh: ");
		    __puthex8(msg_startHigh);
		}
		__puts("\r\n addrLow: ");
		__puthex8(msg_startLow);
		#if DS_PROTOCOL == DS_PROTOCOL_RTOS
			__puts("\r\n process: ");
			__puthex8(msg_process);
			__puts("\r\n thread: ");
			__puthex8(msg_thread);
		#endif
		#if DS_PROTOCOL == DS_PROTOCOL_HYPER
			__puts(" \r\n extraOptions: ");
		    __puthex2(msg_extraOptions);
		#endif
		__puts("\r\n");
	#endif

	/*
	 *	do some preliminary checks
	 */

	if (msg_length > DS_MAXREADWRITELENGTH) {
		__puts( "Length too large: " );
		__puthex4( msg_length );
		__puts( " > " );
		__puthex4( DS_MAXREADWRITELENGTH );
		__puts( "\r\n" );
		return StandardACK(b, kDSReplyACK, kDSReplyParameterError);
	}
	
	/*
	** Reuse buffer to write the reply.
	** Append command (ui8), error (ui8) - standard reply header
	*/

	MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );


	if (err == kNoError) {
		ui8 buffer[DS_MAXREADWRITELENGTH];
		
		size_t length = msg_length;
		
		MemoryAccessOptions accessOptions;

#if DS_PROTOCOL == DS_PROTOCOL_HYPER
		if (msg_extraOptions & DS_MSG_MEMORY_SPACE_VIRT)
		{
		    if (msg_options & DS_MSG_MEMORY_SPACE_DATA)
			accessOptions = kVirtualDataSpace;
		    else
			accessOptions = kVirtualInstructionSpace;
		}
		else
		    accessOptions = kPhysicalSpace;
		    
#else
		accessOptions = (msg_options & DS_MSG_MEMORY_USERVIEW) ? kUserMemory : kDebuggerMemory;
#endif
		
		// for addresses > 32b
		if (msg_options & DS_MSG_MEMORY_EXTENDED)
		{	
			/*
			** Do the memory read.
			*/
#if DS_PROTOCOL == DS_PROTOCOL_HYPER
			err = XTargetAccessExtendedMemory(buffer, (void*) msg_startHigh, (void*) msg_startLow, 
							  &length, accessOptions, true);
#else
			err = DoUnsupported(b);
#endif

		
		}
		else
		{
			/*
			** Do the memory read.
			*/
			err = XTargetAccessMemory(buffer, (void*) msg_startLow, &length,
								 accessOptions, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
								 , msg_process, msg_thread
#endif
								 );		
		}

		/*
		** Append the length and data to the message.
		*/
		
		msg_length = (ui16)length;
		if (err == kNoError)
			err = MessageAdd_ui16( b, msg_length );
	
		if (err == kNoError) {
			err = MessageAddBlock(b, (const void*) buffer, length);
		}
	}
	
	if (err != kNoError)
	{
	
		switch (err)
		{
			case kCWDSException:
				replyErr = kDSReplyCWDSException;
				break;
				
			case kInvalidMemory:
				replyErr = kDSReplyInvalidMemoryRange;
				break;

			case kInvalidProcessId:
				replyErr = kDSReplyInvalidProcessId;
				break;
								
			case kInvalidThreadId:
				replyErr = kDSReplyInvalidThreadId;
				break;
								
			case kOsError:
				replyErr = kDSReplyOsError;
				break;
				
			default:
				replyErr = kDSReplyCWDSError;
				break;
		}

		__puts("DoReadMemory error: ");
		__puthex8(err);
		__puts( " (" );
		__puthex2( replyErr );
		__puts(")\r\n");

		return StandardACK(b, kDSReplyACK, replyErr);
	}

	#if DEBUG_MSG_HANDLER
		__puts("DoReadMemory: sending memory read data\r\n");
	#endif
	err = SendACK(b);
	
	return err;
}


/****************************************************************************/
/*
 *	DoWriteMemory
 *
 *	uses > DS_MAXREADWRITELENGTH of stack space
 */
/****************************************************************************/
DSError DoWriteMemory(MessageBuffer* b)
{
	DSError					err = kNoError;
	DSReplyError			replyErr = kDSReplyNoError;
	ui8						msg_options;
	ui16					msg_length;
	ui32					msg_startHigh;
	ui32					msg_startLow;
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	ui32					msg_process;
	ui32					msg_thread;
#endif
#if DS_PROTOCOL == DS_PROTOCOL_HYPER
	ui8						msg_extraOptions;
#endif
	

	/* check received message size */
	if (b->fLength <= DS_MIN_MSG_WRITEMEMORY_LENGTH) {
		return StandardACK(b, kDSReplyACK, kDSReplyPacketSizeError);
	}
	
	/*
	** Get WriteMemory parameters from the message.
	*/
	
	SetBufferPosition( b, TRK_MSG_HEADER_LENGTH, false );
	if (err == kNoError)
		err = MessageGet_ui8( b, &msg_options );
	if (err == kNoError)
		err = MessageGet_ui16( b, &msg_length );
	if ((err == kNoError) && (msg_options & DS_MSG_MEMORY_EXTENDED))
		err = MessageGet_ui32( b, &msg_startHigh );
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_startLow );	
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_process );
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_thread );
#endif
#if DS_PROTOCOL == DS_PROTOCOL_HYPER
	if (err == kNoError)
		err = MessageGet_ui8( b, &msg_extraOptions );
#endif

	#if DEBUG_MSG_HANDLER
		__puts("DoWriteMemory\r\n");
		__puts(" options: ");
		__puthex2(msg_options);
		__puts("\r\n length: ");
		__puthex4(msg_length);
		if (msg_options & DS_MSG_MEMORY_EXTENDED)
		{
		    __puts("\r\n addrHigh: ");
		    __puthex8(msg_startHigh);
		}
		__puts("\r\n addrLow: ");
		__puthex8(msg_startLow);
		#if DS_PROTOCOL == DS_PROTOCOL_RTOS
			__puts("\r\n process: ");
			__puthex8(msg_process);
			__puts("\r\n thread: ");
			__puthex8(msg_thread);
		#endif
		#if DS_PROTOCOL == DS_PROTOCOL_HYPER
			__puts(" \r\n extraOptions: ");
		    __puthex2(msg_extraOptions);
		#endif
		__puts("\r\n");
	#endif
	/*
	 *	do some preliminary checks
	 */

	if ((b->fLength != (DS_MIN_MSG_WRITEMEMORY_LENGTH + msg_length)) ||
		(msg_length > DS_MAXREADWRITELENGTH)) {
		
		__puts("Size error. fLength = ");
		__puthex8( b->fLength );
		__puts("\r\n");
		return StandardACK(b, kDSReplyACK, kDSReplyParameterError);
	}
	
	/*
	 *	write memory; write into the message buffer
	 */
	if (err == kNoError)
	{
		ui8 buffer[DS_MAXREADWRITELENGTH];
		
		size_t length = msg_length;
		
		MemoryAccessOptions accessOptions;
		
		err = ReadBuffer(b, buffer, length);		

#if DS_PROTOCOL == DS_PROTOCOL_HYPER
		if (msg_extraOptions & DS_MSG_MEMORY_SPACE_VIRT)
		{
		    if (msg_options & DS_MSG_MEMORY_SPACE_DATA)
			accessOptions = kVirtualDataSpace;
		    else
			accessOptions = kVirtualInstructionSpace;
		}
		else
		    accessOptions = kPhysicalSpace;
#else
		accessOptions = (msg_options & DS_MSG_MEMORY_USERVIEW) ? kUserMemory : kDebuggerMemory;
#endif

		
		// for addresses > 32b
		if (msg_options & DS_MSG_MEMORY_EXTENDED)
		{
			/*
			** Do the memory write.
			*/
			
			if (err == kNoError)
#if DS_PROTOCOL == DS_PROTOCOL_HYPER
			    err = XTargetAccessExtendedMemory(buffer, (void*) msg_startHigh, (void*) msg_startLow, 
							      &length, accessOptions, false);
#else
			    err = DoUnsupported(b);
#endif
		}
		else
		{
			/*
			** Do the memory write.
			*/
			
			if (err == kNoError)
			    err = XTargetAccessMemory(buffer, (void*) msg_startLow, &length,
								 accessOptions, false
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
								 , msg_process, msg_thread
#endif
								 );
		}
		
		msg_length = (ui16)length;
	}
	
	
	/*
	** Reuse buffer to write the reply.
	** Append command (ui8), error (ui8) - standard reply header
	*/

	if (err == kNoError)
		MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );

	/*
	** Append the data length.
	*/

	if (err == kNoError)
		err = MessageAdd_ui16( b, msg_length );
	
	
	if (err != kNoError)
	{
	
		switch (err)
		{
			case kCWDSException:
				replyErr = kDSReplyCWDSException;
				break;
				
			case kInvalidMemory:
				replyErr = kDSReplyInvalidMemoryRange;
				break;
				
			case kInvalidProcessId:
				replyErr = kDSReplyInvalidProcessId;
				break;
								
			case kInvalidThreadId:
				replyErr = kDSReplyInvalidThreadId;
				break;
								
			case kOsError:
				replyErr = kDSReplyOsError;
				break;
				
			default:
				replyErr = kDSReplyCWDSError;
				break;
		}

		__puts("DoWriteMemory error: ");
		__puthex8(err);
		__puts( " (" );
		__puthex2( replyErr );
		__puts(")\r\n");

		return StandardACK(b, kDSReplyACK, replyErr);
	}

	__puts("DoWriteMemory: sending ack\r\n");
	err = SendACK(b);
	
	return err;
}


/****************************************************************************/
/*
 *	DoReadRegisters
 */
/****************************************************************************/
DSError DoReadRegisters(MessageBuffer* b)
{
	DSError						err = kNoError;
	DSReplyError				replyErr = kDSReplyNoError;
	DSMessageRegisterOptions	options;
	ui8							msg_options;
	ui16						msg_firstRegister;
	ui16						msg_lastRegister;
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	ui32						msg_process;
	ui32						msg_thread;
#endif

	size_t						registerDataLength;

	/* check received message size */
	if (b->fLength != DS_MIN_MSG_READREGISTERS_LENGTH) {
		return StandardACK(b, kDSReplyACK, kDSReplyPacketSizeError);
	}
	
	/*
	** Get command parameters from the message.
	*/
	
	SetBufferPosition( b, TRK_MSG_HEADER_LENGTH, false );
	if (err == kNoError)
		err = MessageGet_ui8( b, &msg_options );
	if (err == kNoError)
		err = MessageGet_ui16( b, &msg_firstRegister );
	if (err == kNoError)
		err = MessageGet_ui16( b, &msg_lastRegister );
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_process );
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_thread );
#endif

	#if DEBUG_MSG_HANDLER
		__puts("\r\nDoReadRegisters\r\n");
		__puts(" options: ");
		__puthex2(msg_options);
		__puts("\r\n firstRegister: ");
		__puthex4(msg_firstRegister);
		__puts("\r\n lastRegister: ");
		__puthex4(msg_lastRegister);
		#if DS_PROTOCOL == DS_PROTOCOL_RTOS
			__puts("\r\n process: ");
			__puthex8(msg_process);
			__puts("\r\n thread: ");
			__puthex8(msg_thread);
		#endif
		__puts("\r\n");
	#endif

	if (!XTargetStopped(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
						msg_process, msg_thread
#endif
						))
	{
		__puts( "Target is not stopped\r\n" );
		return StandardACK( b, kDSReplyACK, kDSReplyNotStopped );
	}

	/*
	** Check parameters.
	*/

	if (msg_firstRegister > msg_lastRegister)
	{
		return StandardACK( b, kDSReplyACK, kDSReplyInvalidRegisterRange );
	}


	/*
	** Reuse buffer to write the reply.
	** Append command (ui8), error (ui8) - standard reply header
	*/

	if (err == kNoError)
		MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );	

	
	/*
	 *	read registers; write into the message buffer
	 */
	
	options = (DSMessageRegisterOptions)
		((msg_options >> DS_MSG_REGISTERS_TYPE_SHIFT ) & DS_MSG_REGISTERS_TYPE_MASK);
	switch (options) {
		case kDSRegistersDefault:
			err = XTargetAccessDefault(msg_firstRegister, msg_lastRegister,
									b, &registerDataLength, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									,  msg_process, msg_thread
#endif
									);
			break;

		case kDSRegistersFP:
			err = XTargetAccessFP(msg_firstRegister, msg_lastRegister,
								b, &registerDataLength, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
								,  msg_process, msg_thread
#endif
								);
			break;
			
		case kDSRegistersExtended1:
			err = XTargetAccessExtended1(msg_firstRegister, msg_lastRegister,
									b, &registerDataLength, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									,  msg_process, msg_thread
#endif
									);
			break;

		case kDSRegistersExtended2:
			err = XTargetAccessExtended2(msg_firstRegister, msg_lastRegister,
									b, &registerDataLength, true
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									,  msg_process, msg_thread
#endif
									);
			break;

		case kDSRegistersTLB0:
#if DS_PROTOCOL == DS_PROTOCOL_HYPER		
			err = XTargetAccessTLBs(msg_firstRegister, msg_lastRegister,
									b, true, 0
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									,  msg_process, msg_thread
#endif
									);
#else
			err = DoUnsupported(b);
#endif
			break;

		case kDSRegistersTLB1:
#if DS_PROTOCOL == DS_PROTOCOL_HYPER
			err = XTargetAccessTLBs(msg_firstRegister, msg_lastRegister,
									b, true, 1
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									,  msg_process, msg_thread
#endif
									);
#else
			err = DoUnsupported(b);
#endif
			break;
			
		default:
			err = kUnsupportedError;
			break;
	}
	
	if (err != kNoError) {
	
		switch (err)
		{
			case kUnsupportedError:
				replyErr = kDSReplyUnsupportedOptionError;
				break;
			case kInvalidRegister:
				replyErr = kDSReplyInvalidRegisterRange;
				break;
			case kCWDSException:
				replyErr = kDSReplyCWDSException;
				break;
			case kInvalidProcessId:
				replyErr = kDSReplyInvalidProcessId;
				break;
			case kInvalidThreadId:
				replyErr = kDSReplyInvalidThreadId;
				break;
			case kOsError:
				replyErr = kDSReplyOsError;
				break;
			default:
				replyErr = kDSReplyCWDSError;
				break;
		}

		__puts( "\r\nDoReadRegister error: " );
		__puthex8( err );
		__puts( " (" );
		__puthex2( replyErr );
		__puts(")\r\n");
			
		return StandardACK(b, kDSReplyACK, replyErr );
	}

	__puts("\r\nDoReadRegisters: sending memory read data\r\n");
	err = SendACK(b);
	
	return err;
}


/****************************************************************************/
/*
 *	DoWriteRegisters
 */
/****************************************************************************/
DSError DoWriteRegisters(MessageBuffer* b)
{
	DSError						err = kNoError;
	DSReplyError				replyErr = kDSReplyNoError;
	DSMessageRegisterOptions	options;

	ui8							msg_options;
	ui16						msg_firstRegister;
	ui16						msg_lastRegister;
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	ui32						msg_process;
	ui32						msg_thread;
#endif

	size_t						registerDataLength;

	/* check received message size */
	if (b->fLength <= DS_MIN_MSG_WRITEREGISTERS_LENGTH) {
		return StandardACK(b, kDSReplyACK, kDSReplyPacketSizeError);
	}
	
	/*
	** Get command parameters from the message.
	*/
	

	SetBufferPosition( b, TRK_MSG_HEADER_LENGTH, false );
	if (err == kNoError)
		err = MessageGet_ui8( b, &msg_options );
	if (err == kNoError)
		err = MessageGet_ui16( b, &msg_firstRegister );
	if (err == kNoError)
		err = MessageGet_ui16( b, &msg_lastRegister );
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_process );
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_thread );
#endif

	#if DEBUG_MSG_HANDLER
		__puts("\r\nDoWriteRegisters");
		__puts("\r\n options: ");
		__puthex2(msg_options);
		__puts("\r\n firstRegister: ");
		__puthex4(msg_firstRegister);
		__puts("\r\n lastRegister: ");
		__puthex4(msg_lastRegister);
		#if DS_PROTOCOL == DS_PROTOCOL_RTOS
			__puts("\r\n process: ");
			__puthex8(msg_process);
			__puts("\r\n thread: ");
			__puthex8(msg_thread);
		#endif
		__puts("\r\n");
	#endif

	if (!XTargetStopped(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
						msg_process, msg_thread
#endif
						))
	{
		__puts( "Target is not stopped\r\n" );
		return StandardACK( b, kDSReplyACK, kDSReplyNotStopped );
	}

	/*
	** Check parameters.
	*/

	if (msg_firstRegister > msg_lastRegister)
	{
		return StandardACK( b, kDSReplyACK, kDSReplyInvalidRegisterRange );
	}
	
	options = (DSMessageRegisterOptions) msg_options;
	
	switch (options) {
		case kDSRegistersDefault:
			err = XTargetAccessDefault(msg_firstRegister, msg_lastRegister,
									b, &registerDataLength, false
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									,  msg_process, msg_thread
#endif
								);
			break;

		case kDSRegistersFP:
			err = XTargetAccessFP(msg_firstRegister, msg_lastRegister,
								b, &registerDataLength, false
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
								,  msg_process, msg_thread
#endif
								);
			break;
			
		case kDSRegistersExtended1:
			err = XTargetAccessExtended1(msg_firstRegister, msg_lastRegister,
									b, &registerDataLength, false
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									,  msg_process, msg_thread
#endif
									);
			break;

		case kDSRegistersExtended2:
#if DS_PROTOCOL == DS_PROTOCOL_HYPER
			err = XTargetAccessExtended2(msg_firstRegister, msg_lastRegister,
									b, &registerDataLength, false			
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
									,  msg_process, msg_thread
#endif
									);
#else
			err = DoUnsupported(b);
#endif		
			break;
			
		default:
			err = kUnsupportedError;
			break;
	}
	

	/*
	** Reuse buffer to write the reply.
	** Append command (ui8), error (ui8) - standard reply header
	*/

	if (err == kNoError)
		MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );	

	
	if (err != kNoError)
	{
		switch (err)
		{
			case kUnsupportedError:
				replyErr = kDSReplyUnsupportedOptionError;
				break;
			case kInvalidRegister:
				replyErr = kDSReplyInvalidRegisterRange;
				break;
			case kMessageBufferUnderflow:
				/*
				** Underflow happens because data in packet was insufficient.
				*/
				replyErr = kDSReplyPacketSizeError;
				break;
			case kCWDSException:
				replyErr = kDSReplyCWDSException;
				break;
			case kInvalidProcessId:
				replyErr = kDSReplyInvalidProcessId;
				break;
			case kInvalidThreadId:
				replyErr = kDSReplyInvalidThreadId;
				break;
			case kOsError:
				replyErr = kDSReplyOsError;
				break;
			default:
				replyErr = kDSReplyCWDSError;
				break;
		}
			
		__puts("DoWriteRegisters error: ");
		__puthex8(err);
		__puts( " (" );
		__puthex2( replyErr );
		__puts(")\r\n");

		return StandardACK(b, kDSReplyACK, replyErr);
	}


	#if DEBUG_MSG_HANDLER
		__puts("DoWriteRegisters: sending ack\r\n");
	#endif
	err = SendACK(b);
	
	return err;
}


/****************************************************************************/
/*
 *	DoFlushCache
 */
/****************************************************************************/
DSError DoFlushCache(MessageBuffer* b)
{
	DSError						err = kNoError;
	DSReplyError				replyErr = kDSReplyNoError;

	ui8							msg_options;
	ui32						msg_start;
	ui32						msg_end;

	/* check received message size */
	if (b->fLength != DS_MIN_MSG_FLUSHCACHE_LENGTH) {
		return StandardACK(b, kDSReplyACK, kDSReplyPacketSizeError);
	}
	
	/*
	** Get command parameters from the message.
	*/
	
	SetBufferPosition( b, TRK_MSG_HEADER_LENGTH, false );
	if (err == kNoError)
		err = MessageGet_ui8( b, &msg_options );
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_start );
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_end );

	#if DEBUG_MSG_HANDLER
		__puts("\r\nDoWriteRegisters");
		__puts("\r\n options: ");
		__puthex2(msg_options);
		__puts("\r\n start: ");
		__puthex4(msg_start);
		__puts("\r\n end: ");
		__puthex4(msg_end);
		__puts("\r\n");
	#endif

	/*
	** Check parameters.
	*/

	if (msg_start > msg_end)
	{
		return StandardACK( b, kDSReplyACK, kDSReplyInvalidMemoryRange );
	}

	if (err == kNoError)
		err = TargetFlushCache( msg_options, (void *)msg_start, 
								(void *)msg_end );

	/*
	** Reuse buffer to write the reply.
	** Append command (ui8), error (ui8) - standard reply header
	*/

	if (err == kNoError)
		MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );	

	
	if (err != kNoError)
	{
		switch (err)
		{
			case kUnsupportedError:
				replyErr = kDSReplyUnsupportedOptionError;
				break;
			default:
				replyErr = kDSReplyCWDSError;
				break;
		}
			
		__puts("DoFlushCache error: ");
		__puthex8(err);
		__puts( " (" );
		__puthex2( replyErr );
		__puts(")\r\n");

		return StandardACK(b, kDSReplyACK, replyErr);
	}


	#if DEBUG_MSG_HANDLER
		__puts("DoFlushCache: sending ack\r\n");
	#endif
	err = SendACK(b);
	
	return err;
}

/****************************************************************************/
/*
 *	DoSetBreak
 *
 *	For 32-bit address spaces, this command implementation should be
 *	sufficent for setting breakpoints.  To customize this function, make
 *	a copy and use the OverrideDispatch function to dispatch commands to
 *	that function instead.
 *
 */
/****************************************************************************/
#if TRK_SUPPORT_BREAKPOINTS
DSError DoSetBreak(MessageBuffer* b)
{
	DSError err = kNoError;
	BreakpointId bpid;
	Breakpoint *bp;
	bool hardware;
	BreakpointCond conditions;
	

	/*
	** Parse the next segment of the message
	*/

	err = TrkParseMessageSegment( b, &gTrkMsgSetBreak, true );
	
	/*
	** Extract the breakpoint type and conditions from the options
	** field.  Verify that at least one breakpoint condition was specified.
	*/
	
	hardware = (b->fFields.fData[ TRK_MSG_SETBREAK_OPTIONS_INDEX ] &
		~DS_MSG_BREAK_CONDITION_MASK) == kDSBreakHardware;
	conditions = b->fFields.fData[ TRK_MSG_SETBREAK_OPTIONS_INDEX ] &
		DS_MSG_BREAK_CONDITION_MASK;

	if (conditions == 0)
		err = kParameterError;

	/*
	** Get a free breakpoint object, and set the breakpoint based on
	** the message parameters.
	*/

	if (err == kNoError)
		err = GetFreeBreakpoint( &bpid, &bp );

	if (err == kNoError)
	{
		err = SetBreakpoint( bp,
							b->fFields.fData[ TRK_MSG_SETBREAK_ADDRESS_INDEX ],
							b->fFields.fData[ TRK_MSG_SETBREAK_LENGTH_INDEX ],
							hardware,
							conditions,
							b->fFields.fData[ TRK_MSG_SETBREAK_COUNT_INDEX ],
							b->fFields.fData[ TRK_MSG_SETBREAK_TOPTIONS_INDEX ],
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
							b->fFields.fData[ TRK_MSG_SETBREAK_PID_INDEX ],
							b->fFields.fData[ TRK_MSG_SETBREAK_TID_INDEX ]
#else
							0, 0
#endif
							);
							
		/*
		** If everything went well, construct the reply with the new
		** breakpoint ID.  Otherwise, release the breakpoint object.
		*/

		if (err == kNoError)
		{
			MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );	
			b->fFields.fData[ TRK_MSG_SETBREAKREPLY_BID_INDEX ] = bpid;
			err = TrkConstructMessage( b, &gTrkMsgSetBreakReply );
		}
		else
		{
			ReleaseBreakpoint( bpid );
		}
	}
	
	/*
	** Send the reply.
	*/
	if (err == kNoError)
		return( SendACK( b ) );
	
	/*
	** It is an error to reach this point.
	*/
	return( StandardACK(b, kDSReplyACK, TrkStandardReplyError( err )) );
}
#endif /* #if TRK_SUPPORT_BREAKPOINTS */


/****************************************************************************/
/*
 *	DoClearBreak
 *
 *	Clear breakpoints.
 *
 */
/****************************************************************************/
#if TRK_SUPPORT_BREAKPOINTS
DSError DoClearBreak(MessageBuffer* b)
{
	DSError err = kNoError;
	BreakpointId bpid;
	Breakpoint *bp;
	unsigned long bpaddr=0;

	/*
	** Parse the next segment of the message
	*/
	err = TrkParseMessageSegment( b, &gTrkMsgClearBreak, true );

	/*
	** Get the breakpoint object specified by the message, clear it,
	** and release the breakpoint object.
	*/

	if (err == kNoError)
	{
		bpid = b->fFields.fData[ TRK_MSG_CLEARBREAK_BID_INDEX ];
		
		/*
		** For backward compatibility check that Message contains more
		** than one field, i.e breakpoint address is present or not
		*/
		if(b->fLength > 6)
			bpaddr = b->fFields.fData[ TRK_MSG_CLEARBREAK_BADDR_INDEX ];
		
		bp = GetBreakpoint( bpid );
		if (bp == NULL)
			err = kParameterError;
	}
	
	/*
	** Check whether the breakpoint id contained in the message 
	** is in use and same as in current breakpoint table.
	*/
	if (err == kNoError && bp->fInUse && 
						(bpaddr == 0 ||	bp->fStart == bpaddr))
	{
		//Clear breakpoint
		err = ClearBreakpoint( bp );
		ReleaseBreakpoint( bpid );
	}
	

	/*
	** Send the reply.
	*/
	return( StandardACK(b, kDSReplyACK, TrkStandardReplyError( err )) );
}
#endif /* #if TRK_SUPPORT_BREAKPOINTS */


/****************************************************************************/
/*
 *	DoModifyBreakThread
 *
 *	Change the thread associated with a breakpoint.
 *
 */
/****************************************************************************/
#if DS_PROTOCOL == DS_PROTOCOL_RTOS && TRK_SUPPORT_BREAKPOINTS
DSError DoModifyBreakThread(MessageBuffer* b)
{
	DSError err = kNoError;
	BreakpointId bpid;
	Breakpoint *bp;
	

	/*
	** Parse the next segment of the message
	*/

	err = TrkParseMessageSegment( b, &gTrkMsgModifyBreakThread, true );

	/*
	** Get the breakpoint object specified by the message and associate
	** it with the specified thread.
	*/

	if (err == kNoError)
	{
		bpid = b->fFields.fData[ TRK_MSG_MBT_BID_INDEX ];
		bp = GetBreakpoint( bpid );
		if (bp == NULL)
			err = kParameterError;
	}
	
	if (err == kNoError)
		err = OsTargetModifyBreakpointThread( bp, b->fFields.fData[ TRK_MSG_MBT_TID_INDEX ] );
	

	/*
	** Send the reply.
	*/
	return( StandardACK(b, kDSReplyACK, TrkStandardReplyError( err )) );
}
#endif /* #if DS_PROTOCOL == DS_PROTOCOL_RTOS && TRK_SUPPORT_BREAKPOINTS */


/****************************************************************************/
/*
 *	DoCheckTargetState
 *
 *	Checks the state of the target.
  */
/****************************************************************************/
DSError DoCheckTargetState(MessageBuffer* b)
{
	DSError	err = kNoError;
	ui32 state;

#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	ui32						msg_process;
	ui32						msg_thread;
	
	err = MessageGet_ui32( b, &msg_process );
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_thread );
#endif

	#if DEBUG_MSG_HANDLER
		__puts("DoCheckTargetState\r\n");
	#endif
	
	/*
	** Reuse buffer to write the reply.
	** Append command (ui8), error (ui8) - standard reply header
	*/

	MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );

	if (XTargetStopped(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
						msg_process, msg_thread
#endif
						))
		state = 0;
	else
		state = 1;
	
	// add the state in the message : 0 means stopped and 1 means running	
	err = MessageAdd_ui8( b, state);
	
	if (err != kNoError) {
		return StandardACK(b, kDSReplyACK, TrkStandardReplyError(err));
	}
	
	#if DEBUG_MSG_HANDLER
		__puts("DoCheckTargetState: sending state data\r\n");
	#endif
	
	err = SendACK(b);
	
	if (err != kNoError) {
		__puts("error occurred\r\n");
		__puthex8(err);
		__puts("\r\n");
	}

	return err;
}

#if __MWERKS__
#pragma mark -
#pragma mark  Execution Commands
#endif /* __MWERKS__ */
/****************************************************************************/
/*
 *	target "execution" command set
 */
/****************************************************************************/
 
/****************************************************************************/
/*
 *	DoContinue
 */
/****************************************************************************/
DSError DoContinue(MessageBuffer* b)
{
	DSError 				err;
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	ui32					msg_process;
	ui32					msg_thread;
#endif

	err = kNoError;

#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	SetBufferPosition( b, TRK_MSG_HEADER_LENGTH, false );
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_process );
	if (err == kNoError)
		err = MessageGet_ui32( b, &msg_thread );
#endif

	#if DEBUG_MSG_HANDLER
		__puts("DoContinue");
		#if DS_PROTOCOL == DS_PROTOCOL_RTOS
			__puts("\r\n process: ");
			__puthex8(msg_process);
			__puts("\r\n thread: ");
			__puthex8(msg_thread);
		#endif
		__puts( "\r\n" );
	#endif
	
	if (!XTargetStopped(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
						msg_process, msg_thread
#endif
						))
	{
		__puts( "Target is not stopped\r\n" );
		return StandardACK( b, kDSReplyACK, kDSReplyNotStopped );
	}

	err = StandardACK(b, kDSReplyACK, kDSReplyNoError);

	if (err == kNoError)  {
		__console_dump();
		err = XTargetContinue(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
							msg_process, msg_thread
#endif
							);
		__console_dump();
	}
	
	return err;
}


/****************************************************************************/
/*
 *	DoStep
 */
/****************************************************************************/
	
DSError DoStep(MessageBuffer* b)
{
	ui8 msg_options;
	ui8	msg_count;
	ui32 msg_rangeStart;
	ui32 msg_rangeEnd;
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	ui32 msg_process;
	ui32 msg_thread;
#endif
	DSError err = kNoError;
	
	/* check received message size */
	if (b->fLength < DS_MIN_MSG_STEP_LENGTH) {
		return StandardACK(b, kDSReplyACK, kDSReplyPacketSizeError);
	}
	
	/*
	** Get Step parameters from the message.
	*/
	
	SetBufferPosition( b, TRK_MSG_HEADER_LENGTH, false );
	if (err == kNoError)
		err = MessageGet_ui8( b, &msg_options );
		
	#if DEBUG_MSG_HANDLER
		__puts("DoStep");
		__puts("\r\n options: ");
		__puthex2(msg_options);
	#endif

	/*
	** Read parameters based on options.
	*/

	switch (msg_options)
	{
		case kDSStepIntoCount:
		case kDSStepOverCount:

			if (err == kNoError)
				err = MessageGet_ui8( b, &msg_count );

			#if DEBUG_MSG_HANDLER
				__puts("\r\n count: ");
				__puthex2(msg_count);
			#endif
			
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
			if (err == kNoError)
				err = MessageGet_ui32( b, &msg_process );
			if (err == kNoError)
				err = MessageGet_ui32( b, &msg_thread );
		
			#if DEBUG_MSG_HANDLER
				__puts("\r\n process: ");
				__puthex8(msg_process);
				__puts("\r\n thread: ");
				__puthex8(msg_thread);
			#endif
			
			if (err == kNoError)
			{
				
				err = OsTargetValidStep( msg_process, msg_thread );
				switch( err )
				{
					case kInvalidThreadId:
						return StandardACK( b, kDSReplyACK, kDSReplyInvalidThreadId );
						break;
					case kInvalidProcessId:
						return StandardACK( b, kDSReplyACK, kDSReplyInvalidProcessId );
						break;
					case kOsError:
						return StandardACK( b, kDSReplyACK, kDSReplyOsError );
						break;
				}
			}
#endif

			/*
			** Must step over at least one instruction.
			*/

			if ((unsigned)msg_count < 1)
				return StandardACK( b, kDSReplyACK, kDSReplyParameterError );
				
			break;
			
		case kDSStepIntoRange:
		case kDSStepOverRange:
		
			/* check received message size */
			if (b->fLength != DS_MIN_MSG_STEP_RANGE_LENGTH) {
				return StandardACK(b, kDSReplyACK, kDSReplyPacketSizeError);
			}
			if (err == kNoError)
				err = MessageGet_ui32( b, &msg_rangeStart );
			if (err == kNoError)
				err = MessageGet_ui32( b, &msg_rangeEnd );
				
			#if DEBUG_MSG_HANDLER
				__puts( "\r\n rangeStart: " );
				__puthex8( msg_rangeStart );
				__puts( "\r\n rangeEnd: " );
				__puthex8( msg_rangeEnd );
			#endif

#if DS_PROTOCOL == DS_PROTOCOL_RTOS
			if (err == kNoError)
				err = MessageGet_ui32( b, &msg_process );
			if (err == kNoError)
				err = MessageGet_ui32( b, &msg_thread );
		
			#if DEBUG_MSG_HANDLER
				__puts("\r\n process: ");
				__puthex8(msg_process);
				__puts("\r\n thread: ");
				__puthex8(msg_thread);
			#endif
			
			if (err == kNoError)
			{
				err = OsTargetValidStep( msg_process, msg_thread );
				switch( err )
				{
					case kInvalidThreadId:
						return StandardACK( b, kDSReplyACK, kDSReplyInvalidThreadId );
						break;
					case kInvalidProcessId:
						return StandardACK( b, kDSReplyACK, kDSReplyInvalidProcessId );
						break;
					case kOsError:
						return StandardACK( b, kDSReplyACK, kDSReplyOsError );
						break;
				}
			}
#endif

			/*
			** If the PC is not within the specified range already,
			** then the step does not make sense.
			**
			** Note: The conversion to (long) before converting to (ui32)
			** is a workaround for a MIPS compiler bug.  Only the
			** (ui32) conversion should be needed.
			*/
			
			/*
			** SCM: disable this check.  This allows stepping through
			** single-instruction loops.  By not including the current
			** PC in the range, a loop to itself will stop on each
			** iteration.
			*/

			break;
			
		default:

			__puts( "\r\n  Option Invalid\r\n" );		
			return StandardACK(b, kDSReplyACK, kDSReplyUnsupportedOptionError);
			break;
	}

	#if DEBUG_MSG_HANDLER
		__puts( "\r\n" );		
	#endif

	/*
	 *	do some preliminary checks
	 */

	if (!XTargetStopped(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
						msg_process, msg_thread
#endif
						))
	{
		__puts( "Target is not stopped\r\n" );
		return StandardACK( b, kDSReplyACK, kDSReplyNotStopped );
	}


	/*
	** Send the ACK before doing the single step, since the single step will
	** return control to the target program.
	*/

	err = StandardACK(b, kDSReplyACK, kDSReplyNoError);

	if (err == kNoError)
	{
		switch (msg_options)
		{
			case kDSStepIntoCount:
			case kDSStepOverCount:

				err = XTargetSingleStep( (unsigned)msg_count,
										msg_options == kDSStepOverCount
#if   DS_PROTOCOL < DS_PROTOCOL_HYPER /*  DS_PROTOCOL_BOARD */
										, 0, 0
#elif DS_PROTOCOL > DS_PROTOCOL_HYPER /* DS_PROTOCOL_RTOS */
										,  msg_process, msg_thread
#endif
										);
				break;

			case kDSStepIntoRange:
			case kDSStepOverRange:

				err = XTargetStepOutOfRange( msg_rangeStart,
											msg_rangeEnd,
											msg_options == kDSStepOverRange
#if   DS_PROTOCOL < DS_PROTOCOL_HYPER /*  DS_PROTOCOL_BOARD */
										, 0, 0
#elif DS_PROTOCOL > DS_PROTOCOL_HYPER /* DS_PROTOCOL_RTOS */
										,  msg_process, msg_thread
#endif
											);
				break;
		}
	}

	return err;
}


/****************************************************************************/
/*
 *	DoStop
 */
/****************************************************************************/
DSError DoStop(MessageBuffer* b)
{
	DSError err;
	DSReplyError replyErr;
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	ui8		msg_options;
	ui32	msg_process;
	ui32	msg_thread;
#endif

	err = kNoError;

#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	SetBufferPosition( b, TRK_MSG_HEADER_LENGTH, false );
	if (err == kNoError)
		err = MessageGet_ui8( b, &msg_options );
#endif

	#if DEBUG_MSG_HANDLER
		__puts("DoStop");
	#endif
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
	if (err == kNoError &&
		(msg_options == kDSStopProcess || msg_options == kDSStopThread))
		err = MessageGet_ui32( b, &msg_process );

	if (err == kNoError &&
		msg_options == kDSStopThread)
		err = MessageGet_ui32( b, &msg_thread );

	#if DEBUG_MSG_HANDLER
		__puts("\r\n options: ");
		__puthex2(msg_options);
		__puts("\r\n process: ");
		__puthex8(msg_process);
		__puts("\r\n thread: ");
		__puthex8(msg_thread);
	#endif
#endif
	#if DEBUG_MSG_HANDLER
		__puts("\r\n");
	#endif
	
	err = XTargetStop(
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
					msg_options, msg_process, msg_thread
#endif
					);

	switch (err)
	{
		case kNoError:
			replyErr = kDSReplyNoError;
			break;
		case kInvalidProcessId:
			replyErr = kDSReplyInvalidProcessId;
			break;
		case kInvalidThreadId:
			replyErr = kDSReplyInvalidThreadId;
			break;
		case kOsError:
			replyErr = kDSReplyOsError;
			break;
		default:
			replyErr = kDSReplyError;
			break;
	}
	err = StandardACK(b, kDSReplyACK, replyErr);

	return err;
}


#if DS_PROTOCOL == DS_PROTOCOL_RTOS


#if __MWERKS__
#pragma mark -
#pragma mark  Host->Target File Commands
#endif /* __MWERKS__ */
/****************************************************************************/
/*
 *	host->target file access command set
 */
/****************************************************************************/
 

/****************************************************************************/
/*
 *	DoOsWriteFile
 *
 *	Write data to the specified file.
 *
 */
/****************************************************************************/
DSError DoOsWriteFile( MessageBuffer *b )
{
	DSError err = kNoError;
	unsigned int dataPos;
	size_t msgLength;
	DSIOResult ioResult;

	/*
	** Parse the next segment of the message.
	*/

	err = TrkParseMessageSegment( b, &gTrkMsgOsWriteFile, true );
	msgLength = b->fFields.fData[ TRK_MSG_OWF_LENGTH_INDEX ];

	/*
	** Verify that the data field size matches the length parm.
	*/
	
	dataPos = b->fPosition;
	if (b->fLength != (msgLength + dataPos))
		err = kMessageSizeError;
	
	if (err == kNoError)
	{
		/*
		** Write the data to the file.
		*/
	
		ioResult = TargetWriteFile( b->fFields.fData[ TRK_MSG_OWF_HANDLE_INDEX ],
				&(b->fData[ dataPos ]), &msgLength );
				
		/*
		** Build the reply based on the results of the file access.
		*/
	
		MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );	
		b->fFields.fData[ TRK_MSG_OWFR_IORESULT_INDEX ] = ioResult;
		b->fFields.fData[ TRK_MSG_OWFR_LENGTH_INDEX ] = msgLength;

		err = TrkConstructMessage( b, &gTrkMsgOsWriteFileReply );
	}
	
	/*
	** Send the reply.
	*/
	if (err == kNoError)
		return( SendACK( b ) );
	
	/*
	** It is an error to reach this point.
	*/
	return( StandardACK(b, kDSReplyACK, TrkStandardReplyError( err )) );
}

/****************************************************************************/
/*
 *	DoOsReadFile
 *
 *	Read data from the specified file.
 *
 */
/****************************************************************************/
DSError DoOsReadFile( MessageBuffer *b )
{
	DSError err = kNoError;
	int dataPos;
	int temp;
	size_t msgLength;
	DSIOResult ioResult;

	/*
	** Parse the next segment of the message.
	*/

	err = TrkParseMessageSegment( b, &gTrkMsgOsReadFile, true );
	msgLength = b->fFields.fData[ TRK_MSG_ORF_LENGTH_INDEX ];

	/*
	** Get the position of the data field in the reply message by 
	** getting the reply descriptor total byte size.  The data field
	** is not included in that total, so the size should correspond
	** to the start of the data field.
	*/
	
	if (err == kNoError)
		err = TrkGetMessageTotal( &gTrkMsgOsReadFileReply, &dataPos, &temp );
		
	if (err == kNoError)
	{
		/*
		** Read the data from the file.
		*/
	
		ioResult = TargetReadFile( b->fFields.fData[ TRK_MSG_ORF_HANDLE_INDEX ],
				&(b->fData[ dataPos ]), &msgLength );
				
		/*
		** Build the reply based on the results of the file access.
		*/
	
		MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );	
		b->fFields.fData[ TRK_MSG_ORFR_IORESULT_INDEX ] = ioResult;
		b->fFields.fData[ TRK_MSG_ORFR_LENGTH_INDEX ] = msgLength;

		err = TrkConstructMessage( b, &gTrkMsgOsReadFileReply );
		
		/*
		** Make sure data field size is included in outgoing message.
		*/
		
		if (err == kNoError)
			err = SetBufferPosition( b, dataPos + msgLength, false );
	}
	
	/*
	** Send the reply.
	*/
	if (err == kNoError)
		return( SendACK( b ) );
	
	/*
	** It is an error to reach this point.
	*/
	return( StandardACK(b, kDSReplyACK, TrkStandardReplyError( err )) );
}

/****************************************************************************/
/*
 *	DoOsOpenFile
 *
 *	Open the specified file.
 *
 */
/****************************************************************************/
DSError DoOsOpenFile( MessageBuffer *b )
{
	DSError err = kNoError;
	unsigned int dataPos;
	size_t msgLength;
	DSIOResult ioResult;
	DSFileOpenModes modes;
	ui32 handle;
	ui32 timestamp;

	/*
	** Parse the next segment of the message.
	*/

	err = TrkParseMessageSegment( b, &gTrkMsgOsOpenFile, true );
	msgLength = b->fFields.fData[ TRK_MSG_OOF_LENGTH_INDEX ];

	/*
	** Verify that the name field size matches the length parm.
	*/
	
	dataPos = b->fPosition;
	if (b->fLength != (msgLength + dataPos))
		err = kMessageSizeError;
		
	if (err == kNoError)
	{
		/*
		** Add a string terminator to the name.  The buffer is 1 byte larger
		** than the max message size so that this is guaranteed to be safe.
		** Don't use a MessageAdd* function for this, because it won't allow
		** access to that extra byte.
		*/
		
		b->fData[ b->fLength ] = '\0';
	
		/*
		** If the append or create modes are specified, make sure the write
		** mode is also specified.
		*/
		
		modes = b->fFields.fData[ TRK_MSG_OOF_MODES_INDEX ];
		if ((modes & kDSFileOpenAppend) || (modes & kDSFileOpenCreate))
			modes |= kDSFileOpenWrite;
			
		/*
		** Open the file.
		*/
	
		ioResult = TargetOpenFile( modes, &(b->fData[ dataPos ]),
				&handle, &timestamp );
				
		/*
		** Build the reply based on the results of the file access.
		*/
	
		MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );	
		b->fFields.fData[ TRK_MSG_OOFR_IORESULT_INDEX ] = ioResult;
		b->fFields.fData[ TRK_MSG_OOFR_HANDLE_INDEX ] = handle;
		b->fFields.fData[ TRK_MSG_OOFR_TIMESTAMP_INDEX ] = timestamp;

		err = TrkConstructMessage( b, &gTrkMsgOsOpenFileReply );
	}
	
	/*
	** Send the reply.
	*/
	if (err == kNoError)
		return( SendACK( b ) );
	
	/*
	** It is an error to reach this point.
	*/
	return( StandardACK(b, kDSReplyACK, TrkStandardReplyError( err )) );
}

/****************************************************************************/
/*
 *	DoOsCloseFile
 *
 *	Close the specified file.
 *
 */
/****************************************************************************/
DSError DoOsCloseFile( MessageBuffer *b )
{
	DSError err = kNoError;
	DSIOResult ioResult;

	/*
	** Parse the next segment of the message.
	*/

	err = TrkParseMessageSegment( b, &gTrkMsgOsCloseFile, true );

	if (err == kNoError)
	{
		/*
		** Close the file.
		*/
	
		ioResult = TargetCloseFile( b->fFields.fData[ TRK_MSG_OCF_HANDLE_INDEX ],
				b->fFields.fData[ TRK_MSG_OCF_TIMESTAMP_INDEX ] );
				
		/*
		** Build the reply based on the results of the file access.
		*/
	
		MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );	
		b->fFields.fData[ TRK_MSG_OCFR_IORESULT_INDEX ] = ioResult;

		err = TrkConstructMessage( b, &gTrkMsgOsCloseFileReply );
	}
	
	/*
	** Send the reply.
	*/
	if (err == kNoError)
		return( SendACK( b ) );
	
	/*
	** It is an error to reach this point.
	*/
	return( StandardACK(b, kDSReplyACK, TrkStandardReplyError( err )) );
}

/****************************************************************************/
/*
 *	DoOsPositionFile
 *
 *	Set the IO position for the specified file.
 *
 */
/****************************************************************************/
DSError DoOsPositionFile( MessageBuffer *b )
{
	DSError err = kNoError;
	DSIOResult ioResult;

	/*
	** Parse the next segment of the message.
	*/

	err = TrkParseMessageSegment( b, &gTrkMsgOsPositionFile, true );
	
	if (err == kNoError)
	{
		/*
		** Set the file position
		*/
	
		ioResult = TargetPositionFile( b->fFields.fData[ TRK_MSG_OPF_MODES_INDEX ],
				b->fFields.fData[ TRK_MSG_OPF_HANDLE_INDEX ],
				b->fFields.fData[ TRK_MSG_OPF_OFFSET_INDEX ] );
				
		/*
		** Build the reply based on the results of the file access.
		*/
	
		MessageIntoReply( b, kDSReplyACK, kDSReplyNoError );	
		b->fFields.fData[ TRK_MSG_OPFR_IORESULT_INDEX ] = ioResult;

		err = TrkConstructMessage( b, &gTrkMsgOsPositionFileReply );
	}
	
	/*
	** Send the reply.
	*/
	if (err == kNoError)
		return( SendACK( b ) );
	
	/*
	** It is an error to reach this point.
	*/
	return( StandardACK(b, kDSReplyACK, TrkStandardReplyError( err )) );
}

#endif /* #if DS_PROTOCOL == DS_PROTOCOL_RTOS */


#if __MWERKS__
#pragma mark -
#pragma mark  Host->Target Event Notification Commands
#endif /* __MWERKS__ */
/****************************************************************************/
/*
 *	host->target event notification command set
 */
/****************************************************************************/
 

/****************************************************************************/
/*
 *	DoNotifyFileInput
 *
 *	New input is available on a target program input file.  Handle it.
 *
 */
/****************************************************************************/
DSError DoNotifyFileInput( MessageBuffer *b )
{
	DSError err = kNoError;

	/*
	** Parse the next segment of the message.
	*/

	err = TrkParseMessageSegment( b, &gTrkMsgNotifyFileInput, true );
	

	if (err == kNoError)
	{
	
		/*
		** Handle the notification.
		*/
		
		err = XTargetNotifyFileInput(
				b->fFields.fData[ TRK_MSG_NFI_HANDLE_INDEX ]
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
				,b->fFields.fData[ TRK_MSG_NFI_PID_INDEX ]
				,b->fFields.fData[ TRK_MSG_NFI_TID_INDEX ]
#endif
			);
	}
	
	/*
	** Reply with status.
	*/
	return( StandardACK(b, kDSReplyACK, TrkStandardReplyError( err )) );
}

/****************************************************************************/
/*
 *	DoBlockFileIo
 *
 *	Block/unblock file IO requests from the target.
 *
 */
/****************************************************************************/
DSError DoBlockFileIo( MessageBuffer *b )
{
	DSError err = kNoError;

	/*
	** Parse the next segment of the message.
	*/

	err = TrkParseMessageSegment( b, &gTrkMsgBlockFileIo, true );
	

	if (err == kNoError)
		err = XTargetBlockFileIo( b->fFields.fData[ TRK_MSG_BFIO_BLOCK_INDEX ] );
	
	/*
	** Reply with status.
	*/
	return( StandardACK(b, kDSReplyACK, TrkStandardReplyError( err )) );
}
