/******************************************************************************************/
/*
	
DESCRIPTION

	TRK message parsing/construction utilities.
	
COPYRIGHT

	Copyright 2000-2003, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"

#include "usr_put.h"	/* must be included before msgcmd.h */

/*
 *	portable includes
 */
#include "nubassrt.h"
#include "nubevent.h"
#include "msgcmd.h"
#include "msg.h"
#include "msgparse.h"
#include "targimpl.h"
#include "break.h"
#include "string_TRK.h"


/*
** Set this to 1 to see detailed information about normal parsing of
** messages.  (error msgs output via __puts() when one of the DEBUGIO_
** macros is defined will always be emitted regardless of whether
** DEBUG_MSG_PARSE is (1) or (0) .)
*/

#define DEBUG_MSG_PARSE	(1)



/****************************************************************************/
/*
 *	TrkParseMessageSegment
 *
 *	Parses a message segment, reading field values from the message buffer
 *	and storing them in the field value array.  After this function is
 *	called, the buffer will be positioned at the end of the parsed segment.
 *
 *	Parms:
 *		b		- Buffer containing message to be parsed.
 *		desc	- Message segment descriptor.
 *		setPos	- Boolean.  If true, will set the position of the buffer
 *				  to the appropriate point before reading the field values.
 *				  Otherwise, assumes that the position is already set
 *				  correctly.  It is recommended that this always be true
 *				  unless execution time is a major concern or the current
 *				  segment is not at a fixed location in the message.
 */
/****************************************************************************/
DSError TrkParseMessageSegment(MessageBuffer* b, const TrkMsgDesc* desc,
								bool setPos )
{
	DSError 	err;
	int 		i;
	int			startPos;
	int			startIndex;
	int			totalLength;
	ui8			val8;
	ui16		val16;
	ui32		val32;
	
	err = kNoError;
	
	/*
	** Determine the buffer position and field index for the start of the
	** current message segement.  This is determined by summing the sizes
	** and field counts of its headers.
	*/

	err = TrkGetMessageTotal( desc->header, &startPos, &startIndex );	

	/*
	** If the setPos flag is true, set the buffer to the appropriate
	** position first.
	*/
		
	if (err == kNoError && setPos)
		SetBufferPosition( b, startPos, false );
		
	/*
	** Iterate through the fields and read each value based on 
	** its size.
	*/
	
#if DEBUG_MSG_PARSE
	__puts( desc->name );
	__puts( "\r\n" );
#endif
	
	totalLength = 0;
	for (i = 0; (i < desc->count) && (err == kNoError); i++)
	{
		switch( desc->farray[ i ].length )
		{
			case 0:
				/*
				** 0 length is used to skip optional fields.
				*/
				b->fFields.fData[ i + startIndex ] = (MessageFieldType)0;
				break;
			case -1:
				/*
				** -1 length is used for zero-terminated strings.
				** The string address (the address of the current
				** point in the buffer) is stored in the data field,
				** and the buffer position is advanced past the zero-
				** terminator.
				*/
				b->fFields.fData[ i + startIndex ] = (ui32)&(b->fData[ b->fPosition ]);
				do
				{
					err = MessageGet_ui8( b, &val8 );
				} while (err == kNoError && val8 != '\0');
				break;
			case sizeof( ui8 ):
				err = MessageGet_ui8( b, &val8 );
				b->fFields.fData[ i + startIndex ] = (MessageFieldType)val8;
				break;
			case sizeof( ui16 ):
				err = MessageGet_ui16( b, &val16 );
				b->fFields.fData[ i + startIndex ] = (MessageFieldType)val16;
				break;
			case sizeof( ui32 ):
				err = MessageGet_ui32( b, &val32 );
				b->fFields.fData[ i + startIndex ] = (MessageFieldType)val32;
				break;
			default:
				__puts( " unsupported field size in message descriptor\r\n" );
				err = kInternalError;
		}
		
		if (desc->farray[ i ].length != -1)
			totalLength += desc->farray[ i ].length;
		
#if DEBUG_MSG_PARSE
		__puts( "  " );
		__puts( desc->farray[ i ].name );
		__puts( "= " );
		if (desc->farray[ i ].length != -1)
			__puthex8( b->fFields.fData[ i + startIndex ] );
		else
			__puts( (const char *)b->fFields.fData[ i + startIndex ] );
		__puts( "\r\n" );
#endif
	}
	

	/*
	** Do a sanity check on the message/field lengths.
	*/
	if (err == kNoError && totalLength != desc->length)
	{
		__puts( "Message segment length does not match sum of field lengths.\r\n" );
		err = kInternalError;
	}

	return( err );

}


/****************************************************************************/
/*
 *	TrkConstructMessage
 *
 *	Constructs a message, writing field values to the message buffer
 *	based on values stored in the field value array.
 *
 *	Parms:
 *		b		- Buffer to contain constructed message.
 *		desc	- Message segment descriptor.
 */
/****************************************************************************/
DSError TrkConstructMessage(MessageBuffer* b, const TrkMsgDesc* desc )
{
	DSError 	err;
	
	err = kNoError;
	
	/*
	** Construct headers recursively.  If there is no header, this is the
	** first segment of the message so initialize the buffer.
	*/
	if (desc->header != NULL)
		err = TrkConstructMessage( b, desc->header );
	else
		ResetBuffer(b, kDontZero);		/* rewind buffer, don't clear */
		
	if (err == kNoError)
		err = TrkConstructMessageSegment( b, desc, false, true, NULL );
		
	return( err );
}


/****************************************************************************/
/*
 *	TrkConstructMessageSegment
 *
 *	Constructs a message segment, writing field values to the message buffer
 *	based on values stored in the field value array.
 *
 *	Parms:
 *		b		- Buffer to contain constructed message.
 *		desc	- Message segment descriptor.
 *		setPos	- Boolean.  If true, will set the position of the buffer
 *				  to the appropriate point before writing the field values.
 *				  Otherwise, assumes that the position is already set
 *				  correctly.  It is recommended that this always be true
 *				  unless execution time is a major concern or the current
 *				  segment is not at a fixed location in the message.
 *		relIndex- Boolean. If true, will determine the appropriate starting
 *				  relative index into the field value array for this 
 *				  segment.  Otherwise, assumes that the values for this
 *				  segment begin at index 0.
 *		fields	- The field value array.  If NULL, then the field value 
 *				  array from the message buffer will be used.
 */
/****************************************************************************/
DSError TrkConstructMessageSegment(MessageBuffer* b, const TrkMsgDesc* desc,
							bool setPos, bool relIndex, MessageFields *fields )
{
	DSError 	err;
	int 		i;
	int			startPos;
	int			startIndex;
	int			totalLength;
	int			remaining;
	int			slen = 0;
	
	err = kNoError;
	
	if (fields == NULL)
		fields = &(b->fFields);
		
	/*
	** Determine the field index and buffer position for the start of the current
	** message segment. This is determined by summing the sizes and field counts
	** of its headers.
	*/
	
	if (err == kNoError && (setPos || relIndex))
		err = TrkGetMessageTotal( desc->header, &startPos, &startIndex );

	/*
	** If the setPos flag is true, set the buffer to the appropriate
	** position first.
	*/
		
	if (err == kNoError && setPos)
		SetBufferPosition( b, startPos, false );

	/*
	** If relIndex is false, start at 0.
	*/
	
	if (!relIndex)
		startIndex = 0;
			
	/*
	** No segment should be larger than the maximum r/w length.
	** Zero-terminated strings will be truncated to fit.  However,
	** the zero-terminator can't be removed, so reserve space for
	** it.  For simplicity, reserve 1 byte for each field.
	*/

	remaining = DS_MAXREADWRITELENGTH - desc->count;

	/*
	** Iterate through the fields and append each value based on 
	** its size.
	*/
	
#if DEBUG_MSG_PARSE
	__puts( desc->name );
	__puts( "\r\n" );
#endif
	
	totalLength = 0;
	for (i = 0; (i < desc->count) && (err == kNoError); i++)
	{
		switch( desc->farray[ i ].length )
		{
			case 0:
				/*
				** 0 length is used to skip optional fields
				*/
				break;
			case -1:
				/*
				** -1 length is used for zero-terminated strings.
				** If it is larger than the remaining length for
				** the segment, truncate it to fit.  Space has
				** already been reserved for the zero-terminator.
				*/
				if (fields->fData[ i + startIndex ] == 0)
				{
					slen = 0;
				}
				else
				{
					slen = TRK_strlen( (char *)fields->fData[ i + startIndex ] );
					if (slen > remaining)
						slen = remaining;
					err = MessageAddBlock_ui8( b,
							(const ui8 *)fields->fData[ i + startIndex ],
							slen );
				}
				if (err == kNoError)
					err = MessageAdd_ui8( b, '\0' );
				break;
			case sizeof( ui8 ):
				err = MessageAdd_ui8( b,
						(const ui8)fields->fData[ i + startIndex ] );
				break;
			case sizeof( ui16 ):
				err = MessageAdd_ui16( b,
						(const ui16)fields->fData[ i + startIndex ] );
				break;
			case sizeof( ui32 ):
				err = MessageAdd_ui32( b,
						(const ui32)fields->fData[ i + startIndex ] );
				break;
			default:
				__puts( " unsupported field size in message descriptor\r\n" );
				err = kInternalError;
		}
		
		/*
		** totalLength tracks only fixed-length fields, so 
		** zero-terminated strings are not accounted for.
		** remaining does account for them.
		*/
		if (desc->farray[ i ].length != -1)
		{
			totalLength += desc->farray[ i ].length;
			remaining -= desc->farray[ i ].length;
		}
		else
		{
			remaining -= slen;
		}
		
#if DEBUG_MSG_PARSE
		__puts( "  " );
		__puts( desc->farray[ i ].name );
		__puts( "= " );
		if (desc->farray[ i ].length != -1 ||
			fields->fData[ i + startIndex ] == 0)
			__puthex8( fields->fData[ i + startIndex ] );
		else
			__puts( (const char *)fields->fData[ i + startIndex ] );
		__puts( "\r\n" );
#endif
	}
	

	/*
	** Do a sanity check on the message/field lengths.
	*/
	if (err == kNoError && totalLength != desc->length)
	{
		__puts( "Message segment length does not match sum of field lengths.\r\n" );
		err = kInternalError;
	}

	return( err );
}

/****************************************************************************/
/*
 *	TrkStandardReplyError
 *
 *	Maps CodeWarriorTRK internal error codes (DSError) to reply message error
 *	codes (DSReplyError).  The mapping is generic and may not be suitable
 *	for all cases.  
 *
 *	Parms:
 *		err		- The internal error code to be mapped.
 */
/****************************************************************************/
DSReplyError TrkStandardReplyError( DSError err )
{
	DSReplyError replyErr;
	
	switch (err)
	{
		case kNoError:
			replyErr = kDSReplyNoError;
			break;
			
		case kCWDSException:
			replyErr = kDSReplyCWDSException;
			break;
			
		case kInvalidMemory:
			replyErr = kDSReplyInvalidMemoryRange;
			break;

		case kInvalidProcessId:
			replyErr = kDSReplyInvalidProcessId;
			break;
							
		case kInvalidThreadId:
			replyErr = kDSReplyInvalidThreadId;
			break;
							
		case kOsError:
			replyErr = kDSReplyOsError;
			break;
			
		case kMessageOverflow:
		case kMessageBufferUnderflow:
		case kMessageSizeError:
			replyErr = kDSReplyPacketSizeError;
			break;
			
		case kParameterError:
			replyErr = kDSReplyParameterError;
			break;
		
		case kUnsupportedError:
			replyErr = kDSReplyUnsupportedOptionError;
			break;

		case kInvalidRegister:
			replyErr = kDSReplyInvalidRegisterRange;
			break;

		case kBreakpointsFull:
			replyErr = kDSReplyBreakpointsFull;
			break;
			
		case kBreakpointConflict:
			replyErr = kDSReplyBreakpointConflict;
			break;

		default:
			replyErr = kDSReplyCWDSError;
			break;
	}

	return( replyErr );
}

/****************************************************************************/
/*
 *	TrkGetMessageTotal
 *
 *	Gets the total (including headers) byte length and field count for the
 *	given message segment.
 *
 *	Parms:
 *		desc	- Message segment descriptor.
 *		length	- Return storage for total byte length.
 *		count	- Return storage for total field count.
 */
/****************************************************************************/
DSError TrkGetMessageTotal( const TrkMsgDesc* desc,
								int *length, int *count )
{
	DSError 	err;
	const TrkMsgDesc	*idesc;
	
	err = kNoError;
	
	/*
	** Determine the buffer position and field index for the end of the
	** given message.  This is determined by summing the sizes and field
	** counts of it and its headers.
	*/
	
	*length = 0;
	*count = 0;
	for (idesc = desc; idesc != NULL; idesc = idesc->header)
	{
		*length += idesc->length;
		*count += idesc->count;
	}

	return( err );
}
