/******************************************************************************************/
/*

DESCRIPTION

	Event notification from TRK to host debugger.
	
	Send messages (notification) from TRK to the host debugger.
	
COPYRIGHT

	Copyright 1999-2003, 2006, 2007, 2009, 2010 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"

#ifndef _NO_SYSTEM_INCLUDES
#include <string.h>
#endif

#include "usr_put.h"		// must be included before msgcmd.h
#include "nubassrt.h"
#include "notify.h"
#include "msgbuf.h"
#include "msg.h"
#include "msgcmd.h"
#include "usr_put.h"
#include "targimpl.h"
#include "support.h"
#include "msgparse.h"

/*
** Set this to 1 to see detailed information about normal processing
** of Notify events.  (error msgs output via __puts() when one of the
** DEBUGIO_ macros is defined will always be emitted regardless of
** whether DEBUG_NOTIFY is (1) or (0) .)
*/

#define DEBUG_NOTIFY	(1)

#define NOTIFY_TRY_COUNT 3


/******************************************************************************/
/*
 *	DoNotifyStopped - used to send stop
 */
/******************************************************************************/

DSError DoNotifyStopped( MessageCommandID command
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
						, ui32 process_id, ui32 thread_id
#endif
						)
{
	DSError			err;
	
	MessageBufferId	bufferID;
	MessageBuffer*	buffer;

#if DEBUG_NOTIFY	
	__puts( "DoNotifyStopped: " );
	__puthex2( command );
#endif
	
	/*
	 *	Get a send buffer, create a message
	 */
	
	err = GetFreeBuffer(&bufferID, &buffer);
	
	if (err == kNoError) {
		/*
		** Add the command & sequence ID to the message.  Requests and
		** notifications (all outgoing messages) share the same ID sequence.
		*/

		buffer->fFields.fData[ TRK_MSG_COMMAND_INDEX ] = command;
		buffer->fFields.fData[ TRK_MSG_SID_INDEX ] = GetNextSendSeqId();
		err = TrkConstructMessage( buffer, &gTrkMsgHeader );
		
		/*
		 * Add target-defined info.  Usually the PC, instruction at
		 * PC, and exception word.
		 */
		 
		if (err == kNoError)
		{
			err = XTargetAddStopInfo( buffer
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
										, process_id, thread_id
#endif
										);
		}
			
		FAIL(err == kNoError);			/* shouldn't be possible */

		/*
		 *	Send the notification
		 */
		err = DoNotify( buffer );
		ReleaseBuffer(bufferID);		
	}
	
	return( err );
}

/******************************************************************************/
/*
 *	DoNotifyCoreReleased - used to notify the host that a core has been 
 *			       released from spinning
 */
/******************************************************************************/

DSError DoNotifyCoreReleased(ui32 coreID)
{
	DSError			err;
	
	MessageBufferId	bufferID;
	MessageBuffer*	buffer;

#if DEBUG_NOTIFY	
	__puts( "DoNotifyCoreReleased: " );
#endif
	
	/*
	 *	Get a send buffer, create a message
	 */
	
	err = GetFreeBuffer(&bufferID, &buffer);
	
	if (err == kNoError) {
		/*
		** Add the command & sequence ID to the message.  Requests and
		** notifications (all outgoing messages) share the same ID sequence.
		*/

		buffer->fFields.fData[ TRK_MSG_COMMAND_INDEX ] = kDSNotifyCoreReleased;
		buffer->fFields.fData[ TRK_MSG_SID_INDEX ] = GetNextSendSeqId();
		err = TrkConstructMessage( buffer, &gTrkMsgHeader );
		
		/*
		 * Add target-defined info.  Usually the PC, instruction at
		 * PC, and exception word.
		 */
		 
		if (err == kNoError)
		{
			// Append the coreID
			err = MessageAdd_ui32( buffer, coreID );
		}
			
		FAIL(err == kNoError);			/* shouldn't be possible */

		/*
		 *	Send the notification
		 */
		err = DoNotify( buffer );
		ReleaseBuffer(bufferID);		
	}
	
	return( err );
}

/******************************************************************************/
/*
 *	DoNotify - used to send generic notifications
 */
/******************************************************************************/

DSError DoNotify( MessageBuffer *buffer )
{
	DSError err;
	int notifyTries;
	MessageBufferId replyID;
	
	/*
	 *	Send the notification
	 */
	err = kNoError;
	notifyTries = NOTIFY_TRY_COUNT;		/* try a few times */

	err = RequestSend( buffer, &replyID,
					   DS_MIN_REPLY_LENGTH,
					   NOTIFY_TRY_COUNT, true );
	if (err == kNoError)
		ReleaseBuffer( replyID );
	else
	{
#if DEBUG_NOTIFY	
		if (err == kWaitACKError)
			__puts( "DoNotify: Invalid ACK received or timeout\r\n" );
		else
			__puts( "DoNotify: Send failed\r\n" );
#endif
	}
			
	return( err );
}
