/******************************************************************************************/
/*
	
DESCRIPTION

	Event queue.
	
COPYRIGHT

	Copyright 1999-2002, 2004, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"

/*
	system includes
*/
#ifndef _NO_SYSTEM_INCLUDES
#include <string.h>
#include <stdlib.h>
#endif

/*
	CWDS includes
*/
#include "usr_put.h"		// must be included before msgcmd.h
#include "nubassrt.h"
#include "nubevent.h"
#include "mutex_TRK.h"
#include "mem_TRK.h"
#ifdef IS_SMP
#include "smp.h"
#endif


/****************************************************************************/
/*
 *	Local Types and Variables
 */
/****************************************************************************/

/*
** The minimum number of events in the system is 2.  However, if
** multiple threads can generate events simultaneously, then the
** number should be (1 + <max simultaneous thread events>).
**
** If a particular target requires more than 2, it should define
** TRK_EXTRA_EVENT_COUNT in target.h.
*/

#ifndef TRK_EXTRA_EVENT_COUNT
	#define TRK_EXTRA_EVENT_COUNT 0
#endif

#define INITIAL_NUBQUEUELENGTH	(2 + TRK_EXTRA_EVENT_COUNT)

const NubEventId kFirstEventID = 0x100;

typedef struct EventQueue {
	DSMutex			fMutex;
	unsigned int	fCount;						/* "active" entries in the queue */
	unsigned int	fCountAllocated;			/* the number of NubEvent items 'fEventList' points to */
	unsigned int	fFirst;						/* current item in list */
	NubEvent*		fEventList;
	NubEventId		fEventID;					/* next ID to be used */
} EventQueue;

#ifndef IS_SMP
	/*static*/ EventQueue gEventQueue;
#else
	#define gEventQueue (*perCpuEventQueue)
#endif

#if !defined(DS_MALLOC_MAPPING) || (DS_MALLOC_MAPPING == DS_MALLOC_NOT_AVAILABLE)
	#define DS_DYANMIC_EVENT_QUEUE 0
#else
	#define DS_DYANMIC_EVENT_QUEUE 1
#endif

/****************************************************************************/
/*
 *	InitializeEventQueue
 */
/****************************************************************************/
DSError InitializeEventQueue(void)
{
	DSError error = kNoError;
#ifdef IS_SMP
	EventQueue* perCpuEventQueue = TRK_malloc(sizeof(EventQueue));
	PRIVATE_DATA(EVENT_QUEUE) = perCpuEventQueue;
#endif
	InitializeMutex(&gEventQueue.fMutex);
	AcquireMutex(&gEventQueue.fMutex);
	gEventQueue.fCount = 0;
	gEventQueue.fFirst = 0;
	gEventQueue.fEventID = kFirstEventID;
	gEventQueue.fCountAllocated	= INITIAL_NUBQUEUELENGTH;

	#if DS_DYANMIC_EVENT_QUEUE
		gEventQueue.fEventList = TRK_malloc(sizeof(NubEvent) * gEventQueue.fCountAllocated);
		if (!gEventQueue.fEventList)
			error = kOutOfMemory;
	#else
	{
		static NubEvent eventQueue[INITIAL_NUBQUEUELENGTH];
		gEventQueue.fEventList = eventQueue;
	}
	#endif
		
	ReleaseMutex(&gEventQueue.fMutex);
	
	return error;
}

/****************************************************************************/
/*
 *	TerminateEventQueue
 */
/****************************************************************************/
DSError TerminateEventQueue(void)
{
#ifdef IS_SMP
	EventQueue* perCpuEventQueue = PRIVATE_DATA(EVENT_QUEUE);
#endif
	InitializeMutex(&gEventQueue.fMutex);
	AcquireMutex(&gEventQueue.fMutex);
	gEventQueue.fCount = 0;
	gEventQueue.fCountAllocated	= 0;
	gEventQueue.fFirst = 0;
	gEventQueue.fEventID = kFirstEventID;

	#if DS_DYANMIC_EVENT_QUEUE
		if (gEventQueue.fEventList)
			TRK_free(gEventQueue.fEventList);
	#endif

	gEventQueue.fEventList = NULL;
	ReleaseMutex(&gEventQueue.fMutex);

#ifdef IS_SMP
	TRK_free(perCpuEventQueue);
	PRIVATE_DATA(EVENT_QUEUE) = NULL;
#endif
	return kNoError;
}


/****************************************************************************/
/*
 *	CopyEvent
 */
/****************************************************************************/
void CopyEvent(NubEvent* dstEvent, const NubEvent* srcEvent)
{
	ASSERT((dstEvent != NULL) && (srcEvent != NULL));
	TRK_memcpy(dstEvent, srcEvent, sizeof(*dstEvent));
}

/****************************************************************************/
/*
 *	GetNextEvent
 */
/****************************************************************************/
bool GetNextEvent(NubEvent* resultEvent)
{
	bool result = false;
#ifdef IS_SMP
	EventQueue* perCpuEventQueue = PRIVATE_DATA(EVENT_QUEUE);
#endif

	AcquireMutex(&gEventQueue.fMutex);

	ASSERT(gEventQueue.fCount <= gEventQueue.fCountAllocated);
	ASSERT(gEventQueue.fFirst < gEventQueue.fCountAllocated);

	if (gEventQueue.fCount > 0) {
		CopyEvent(resultEvent, &gEventQueue.fEventList[gEventQueue.fFirst]);
		
		gEventQueue.fCount--;

		gEventQueue.fFirst++;
		if (gEventQueue.fFirst == gEventQueue.fCountAllocated)
			gEventQueue.fFirst = 0;

		result = true;
	}

	ASSERT(gEventQueue.fCount < gEventQueue.fCountAllocated);
	ASSERT(gEventQueue.fFirst < gEventQueue.fCountAllocated);

	ReleaseMutex(&gEventQueue.fMutex);

	return result;
}

/****************************************************************************/
/*
 *	PostEvent
 */
/****************************************************************************/
DSError PostEvent(const NubEvent* copiedEvent)
{
	DSError result = kNoError;
	int	nextEntry;
#ifdef IS_SMP
	EventQueue* perCpuEventQueue = PRIVATE_DATA(EVENT_QUEUE);
#endif

	AcquireMutex(&gEventQueue.fMutex);
	ASSERT(gEventQueue.fCount <= gEventQueue.fCountAllocated);
	ASSERT(gEventQueue.fFirst < gEventQueue.fCountAllocated);

	/*
	 * The event queue is implemented as a circular queue. It's capacity
	 * grows as needed, limited only by system memory. 	 
	 */
	if (gEventQueue.fCount == gEventQueue.fCountAllocated)
	{
		/* Time to expand */

		unsigned int firstSegCount = 0;
		NubEvent * oldList = gEventQueue.fEventList;
		unsigned int newAllocatedCount = (gEventQueue.fCountAllocated + INITIAL_NUBQUEUELENGTH);

		/* Allocate the new queue buffer; the increase amount is the intial entry count */
		gEventQueue.fEventList = TRK_malloc(sizeof(NubEvent) * newAllocatedCount);
		if (!gEventQueue.fEventList)
			return kEventQueueFull;

		/* 
		 * The queue is circular; the area starting with the first queue item up to and including
		 * the last item at the end of the buffer we'll consider the "first segment"; the part of
		 * the queue that starts at the beginning of the buffer and ends with the last queue item
		 * we'll consider the "second segment". Remeber...we're here because we have a full queue.
		 * Also note that it's possible to only have a first segment; this is the case if the
		 * first item happens to be positioned at the beginning of the current queue buffer.
		 */
		firstSegCount = gEventQueue.fCountAllocated - gEventQueue.fFirst;		

		/* Copy the first segment into the new queue buffer */		
		TRK_memcpy( &gEventQueue.fEventList[0],
					&oldList[gEventQueue.fFirst], 
					sizeof(NubEvent) * firstSegCount);

		/* Append the second segment */
		if (gEventQueue.fFirst != 0)
		{
			TRK_memcpy( &gEventQueue.fEventList[firstSegCount],
						&oldList[0], 
						sizeof(NubEvent) * (gEventQueue.fCountAllocated - firstSegCount));
		}

		/* Update variables; note that fCount remains the same; it'll get bumped below */
		gEventQueue.fFirst = 0;
		gEventQueue.fCountAllocated	= newAllocatedCount;
		
		TRK_free(oldList);
	}
	
	/* copy event to next slot in event queue */
	
	nextEntry = (gEventQueue.fFirst + gEventQueue.fCount) % gEventQueue.fCountAllocated;
	CopyEvent(&gEventQueue.fEventList[nextEntry], copiedEvent);
	
	/* set the ID for the event in the queue */
	
	gEventQueue.fEventList[nextEntry].fId = gEventQueue.fEventID;
	gEventQueue.fEventID++;
	if (gEventQueue.fEventID < kFirstEventID)
		gEventQueue.fEventID = kFirstEventID;
	
	/* increment count of entries in the queue */
		
	gEventQueue.fCount++;
	
	ASSERT(gEventQueue.fCount > 0 && gEventQueue.fCount <= gEventQueue.fCountAllocated);

	ReleaseMutex(&gEventQueue.fMutex);
	return result;
}

/****************************************************************************/
/*
 *	ConstructEvent
 */
/****************************************************************************/
void ConstructEvent(NubEvent* newEvent, NubEventType type)
{
	ASSERT(newEvent != NULL);
	newEvent->fType = type;
	newEvent->fId = kInvalidEventId;
	newEvent->fMessageBufferId = kInvalidMessageBufferId;
}

/****************************************************************************/
/*
 *	DestructEvent
 */
/****************************************************************************/
void DestructEvent(NubEvent* oldEvent)
{
	if (oldEvent->fMessageBufferId != kInvalidMessageBufferId)
		ReleaseBuffer(oldEvent->fMessageBufferId);
}
