/******************************************************************************************/
/*

DESCRIPTION

	Object array management routines.

COPYRIGHT

	Copyright 2000, 2002, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"

/*
 *	Portable includes
 */
#include "usr_put.h"		// must be included before msgcmd.h
#include "dstypes.h"
#include "dserror.h"
#include "nubassrt.h"
#include "obarray.h"


/****************************************************************************/
/*
	Local Types and Variables
*/
/****************************************************************************/

/****************************************************************************/
/*
	static function declarations
*/
/****************************************************************************/
static void SetArrayObjectUsed(ObjectArray *obArray, ObjectId i, bool inUse);


/****************************************************************************/
/*
	SetObjectUsed
*/
/****************************************************************************/

static void SetArrayObjectUsed(ObjectArray *obArray, ObjectId i, bool inUse)
{
	ObjectHeader *ob;
	ObjectId j;
	
	ob = GetArrayObject( obArray, i );

	ASSERT(ob);
	
	ob->fInUse = inUse;
	
	/*
	** Update max used field.  If setting this one to true, 
	** then set new maxUsed if it is beyond the max.  If 
	** setting this one to false and it is the current max,
	** then find previous used.
	*/
	AcquireMutex( &obArray->fMutex );
	if (inUse)
	{
		if (obArray->fMaxUsed < i || obArray->fMaxUsed == kInvalidObjectId)
			obArray->fMaxUsed = i;
	}
	else
	{
		if (obArray->fMaxUsed == i)
		{
			/*
			** Assume no others are used, set the max below all valid ID's
			*/
			obArray->fMaxUsed = kInvalidObjectId;
			
			/*
			** Work backwards from the previous max.  If a used bp is
			** found, make it the new max and stop.
			*/
			for (j = i - 1; j >= 0; --j)
			{
				ob = GetArrayObject( obArray, j );
				if (ob->fInUse)
				{
					obArray->fMaxUsed = j;
					break;
				}
			}
		}
	}
	ReleaseMutex( &obArray->fMutex );
}

/****************************************************************************/
/*
	InitializeObjectArray
*/
/****************************************************************************/

DSError InitializeObjectArray( ObjectArray *obArray,
								unsigned int count,
								size_t objectSize, 
								InitializeObjectType initFunc, 
								void *objects )
{
	DSError err = kNoError;
	ObjectHeader *ob;
	int i;
	
	InitializeMutex(&obArray->fMutex);
	obArray->fMaxUsed = kInvalidObjectId;
	obArray->fCount = count;
	obArray->fObjectSize = objectSize;
	obArray->fInitializeObject = initFunc;
	obArray->fObjects = objects;
	
	for (i = 0; i < obArray->fCount && err == kNoError; i++)
	{
		ob = GetArrayObject( obArray, i );
		err = InitializeObjectHeader( ob );
		if (err == kNoError)
			err = obArray->fInitializeObject( ob );
	}
	return( err );
}

/****************************************************************************/
/*
	InitializeObjectHeader
*/
/****************************************************************************/

DSError InitializeObjectHeader( ObjectHeader *ob )
{
	ASSERT( ob );
	
	InitializeMutex(&(ob->fMutex));
	AcquireMutex(&(ob->fMutex));
	ob->fInUse = false;
	ReleaseMutex(&(ob->fMutex));

	return kNoError;
}

/****************************************************************************/
/*
	GetFreeArrayObject
*/
/****************************************************************************/

DSError _GetFreeArrayObject( ObjectArray *obArray, 
							ObjectId* resultObId,
							ObjectHeader** resultOb)
{
	DSError result = kObjectArrayFull;		/* assume failure */
	ObjectId i;
	ObjectHeader* ob;
	
	*resultOb = NULL;
	
	/*
		linear search through the objects to find one that is not in use
	*/
	
	for (i = 0; i < obArray->fCount; i++) {
		ob = GetArrayObject( obArray, i );
		AcquireMutex(&ob->fMutex);
		
		if (ob->fInUse == false) {
		
			/* initialize object */
			SetArrayObjectUsed( obArray, i, true );
			obArray->fInitializeObject( ob );
			
			*resultOb = ob;
			*resultObId = i;
			result = kNoError;
			
			i = obArray->fCount;		/* terminate iteration */
		}
		
		ReleaseMutex(&ob->fMutex);
	}	

	return result;
}

/****************************************************************************/
/*
	GetArrayObject
	
	Get a pointer to a single object structure
*/
/****************************************************************************/

ObjectHeader *GetArrayObject(ObjectArray *obArray, ObjectId obId)
{
	ObjectHeader* result = NULL;
	
	if (obId >= 0 && obId < obArray->fCount) {
		result = (ObjectHeader *)(((ui8 *)obArray->fObjects) +
				(obArray->fObjectSize * obId));
	}
	
	return result;
}

/****************************************************************************/
/*
	GetArrayObjectId
	
	Get an object ID from the object structure pointer.
*/
/****************************************************************************/

ObjectId GetArrayObjectId(ObjectArray *obArray, ObjectHeader *ob)
{
	ObjectId result;
	
	result = ((int)ob - (int)(obArray->fObjects)) / (obArray->fObjectSize);
	
	if (result >= 0 && result < obArray->fCount &&
		(ob == (ObjectHeader *)(((ui8 *)obArray->fObjects) +
				(obArray->fObjectSize * result))))
		return( result );
	
	return( kInvalidObjectId );
}

/****************************************************************************/
/*
	ReleaseArrayObject
	
	Mark the object as unused
*/
/****************************************************************************/

void _ReleaseArrayObject(ObjectArray *obArray, ObjectId obId)
{
	ObjectHeader* ob;
	
	if (obId == kInvalidObjectId)
		return;
	
	if (obId >= 0 && obId < obArray->fCount) {


		ob = GetArrayObject( obArray, obId );

		AcquireMutex(&ob->fMutex);
		SetArrayObjectUsed(obArray, obId, false);
		ReleaseMutex(&ob->fMutex);		

	}
}

/****************************************************************************/
/*
	FindArrayObject
	
	Find an object in the array that matches according to the provided
	key and match function.
*/
/****************************************************************************/
DSError FindArrayObject( ObjectArray *obArray,
							ObjectId* resultObId,
							ObjectHeader** resultOb,
							ObjectHeader *keyOb,
							MatchObjectType *matchFunc )
{
	DSError err = kObjectNotFound;
	ObjectId i;
	ObjectHeader *ob;

	for (i = IterateObjectArrayFirst( obArray, &ob );
	     i != kInvalidObjectId && err == kObjectNotFound;
	     i = IterateObjectArrayNext( obArray, i, &ob ))
	{
		
		/*
		** If we find a breakpoint that is in use and matches the key,
		** return it.
		*/
		
		if (ob->fInUse && matchFunc( keyOb, ob ))
			{
				*resultObId = i;
				*resultOb = ob;
				err = kNoError;
			}
	}
	
	return( err );
}

/****************************************************************************/
/*
	IterateObjectArrayFirst
	
	Returns the the "first" object ID to use for iteration over the array.
	Starts at the end with the max-used ID.
*/
/****************************************************************************/
ObjectId IterateObjectArrayFirst( ObjectArray *obArray, ObjectHeader **resultOb )
{
	/*
	** Return the max-used ID and object.  We assume that this object is
	** actually in use.
	*/
	if (obArray->fMaxUsed != kInvalidObjectId)
	{
		*resultOb = GetArrayObject( obArray, obArray->fMaxUsed );
		ASSERT( (*resultOb)->fInUse );
	}
	else
	{
		*resultOb = NULL;
	}
		
	return( obArray->fMaxUsed );
}

/****************************************************************************/
/*
	IterateObjectArrayNext
	
	Returns the the "next" object ID to use for iteration over the array.
*/
/****************************************************************************/
ObjectId IterateObjectArrayNext( ObjectArray *obArray,
									ObjectId prev, ObjectHeader **resultOb )
{
	ObjectId next;
	ObjectHeader *ob;

	/*
	** Work backwards through the IDs to find an object whose in-use flag
	** is set.  When one is found, return it immediately.
	*/
	
	if (prev != kInvalidObjectId)
	{
		for (next = prev - 1; next >= 0; next--)
		{
			ob = GetArrayObject( obArray, next );
			if (ob->fInUse)
			{
				*resultOb = ob;
				return( next );
			}
		}
	}

	/*
	** No more left.
	*/
		
	*resultOb = NULL;
	return( kInvalidObjectId );
}
