/******************************************************************************************/
/*
	
DESCRIPTION

	General object array management for CodeWarriorTRK.
	
	Objects are allocated out of a pool (array) of objects, preallocated
	at initialization.  If dynamic allocation is allowed, then the array
	may be expanded as needed.
	
	Usage
	
		Call InitializeBreakpoints once to initialize the breakpoints. When a
		new breakpoint is desired, call "GetFreeBreakpoint" and a breakpoint ID
		plus a pointer to the Breakpoint struct is returned.  Call "ResetBreakpoint"
		before using the breakpoint.  When using the Breakpoint, acquire and
		release the mutex before and after using or modifying it.  Once the
		breakpoint is no longer being used, call ReleaseBreakpoint.

	Initialization
	
		DSError InitializeBreakpoints(void);
		
			Call once at the beginning of the program.
	
	Getting and Releasing
	
		DSError GetFreeBreakpoint(BreakpointId* returnedBreakpointId,
									Breakpoint** breakpoint);
		Breakpoint* GetBreakpoint(BreakpointId breakpointId);
		
		
			GetFreeBreakpoint will acquire a breakpoint if it is available.
			Returned values:
			kNoError	the returnedBreakpointId and breakpoint pointers refer
						to the allocated breakpoint
						
			kBreakpointsFull
						no more breakpoints are available.
			
			GetBreakpoint gets a pre-allocated Breakpoint pointer from the
			breakpoint's ID.  If the breakpoint was not allocated or the ID is out
			of range, GetBreakpoint retuns NULL.
			
			
		void ReleaseBreakpoint(BreakpointId id);
			
			When a breakpoint is no longer needed, it can be returned to
			the storage pool.
		
			
	Operations on Breakpoints
		SetBreakpoint()
		ClearBreakpoint()
		EnableBreakpoint()
		DisableBreakpoint()
		FindBreakpoint()
		

COPYRIGHT

	Copyright 2000, 2002, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#ifndef _OBARRAY_H_
#define _OBARRAY_H_

/****************************************************************************/
/*
 *	Includes
 */
/****************************************************************************/
#ifndef _NO_SYSTEM_INCLUDES
#include <stddef.h>
#endif
#include "dstypes.h"
#include "dserror.h"
#include "mutex_TRK.h"
#include "target.h"
#include "msgcmd.h"
#include "usr_put_config.h"

/*
** Set this to 1 to get debug messages showing the get/release calls
** for the array objects.  This helps to track down leaks.
*/
#define DEBUG_OBARRAY_USAGE		(0)

#if DEBUG_OBARRAY_USAGE && \
	(defined( DEBUGIO_SERIAL ) || defined( DEBUGIO_RAM ) || defined( DEBUGIO_CUSTOM ))
	#define GetFreeArrayObject( a, id, b )	\
										( __puts( "Calling GetFreeArrayObject from " ), \
										  __puts( __FILE__ ), \
										  __putchar( ':' ), \
										  __puthex8( __LINE__ ), \
										  __puts( "\r\n" ), \
										  _GetFreeArrayObject( a, id, b ) )

	#define ReleaseArrayObject( a, id )	\
										( __puts( "Calling ReleaseArrayObject from " ), \
										  __puts( __FILE__ ), \
										  __putchar( ':' ), \
										  __puthex8( __LINE__ ), \
										  __puts( "\r\n" ), \
										  _ReleaseArrayObject( a, id ) )
#else /* #if DEBUG_OBARRAY_USAGE */
	#define GetFreeArrayObject( a, id, b )	_GetFreeArrayObject( a, id, b )
	#define ReleaseArrayObject( a, id )		_ReleaseArrayObject( a, id )
#endif /* #if DEBUG_OBARRAY_USAGE */

/****************************************************************************/
/*
 *	Errors
 */
/****************************************************************************/
enum {								/* DSErroror; see dserror.h */
	kObjectArrayFull = 0x0a00,		/* no more available obarray space */
	kObjectNotFound					/* could not find specified object in obarray */
};

/****************************************************************************/
/*
 *	Types
 */
/****************************************************************************/


/*
 *	ObjectId
 */
typedef int ObjectId;

enum {
	kInvalidObjectId = -1
};

/*
 ** struct ObjectHeader.  All objects in an object array must start with
 ** this header.
 */

typedef struct ObjectHeader {
	DSMutex			fMutex;					/* synchronization */
	bool			fInUse;					/* true if objects is not free */
} ObjectHeader;

/*
** Prototype for an object initialization function.  This is called by
**	the general-purpose object array routines, so the passed in type 
**	assumes the header only.  Note that the header itself should NOT be
**	initialized by this routine, only the additional object-specific fields.
*/

typedef DSError InitializeObjectType( ObjectHeader *ob );

/*
** struct ObjectArray.  The actual C array of objects is pointed to
** by the fObjects member.
*/

typedef struct ObjectArray {
	DSMutex					fMutex;
	ObjectId 				fMaxUsed;
	unsigned				fCount;
	size_t					fObjectSize;
	InitializeObjectType	*fInitializeObject;  // should not initialize the header fields
	void*					fObjects;
} ObjectArray;

/*
** Prototype for a function which can find a match between two objects.
** It returns true if they match and false otherwise.  Such functions
** are used by the find routine.
*/

typedef bool MatchObjectType( ObjectHeader *ob1, ObjectHeader *ob2 );

/****************************************************************************/
/*
 *	Public Function Prototypes
 */
/****************************************************************************/
__extern_c

	/*
	 *	Object array management
	 */
	
	DSError InitializeObjectArray( ObjectArray *obArray,
							unsigned int count,
							size_t objectSize, 
							InitializeObjectType initFunc, 
							void *objects );
	DSError InitializeObjectHeader( ObjectHeader *ob );
	DSError _GetFreeArrayObject( ObjectArray *obArray, 
							ObjectId* resultObId,
							ObjectHeader** resultOb);
	ObjectHeader* GetArrayObject( ObjectArray *obArray, ObjectId obId);
	ObjectId GetArrayObjectId(ObjectArray *obArray, ObjectHeader *ob);
	void _ReleaseArrayObject( ObjectArray *obArray, ObjectId obId);
	DSError FindArrayObject( ObjectArray *obArray,
							ObjectId* resultObId,
							ObjectHeader** resultOb,
							ObjectHeader *keyOb,
							MatchObjectType *matchFunc );
	ObjectId IterateObjectArrayFirst( ObjectArray *obArray,
							ObjectHeader **resultOb );
	ObjectId IterateObjectArrayNext( ObjectArray *obArray,
							ObjectId prev, ObjectHeader **resultOb );
	
__end_extern_c

#endif /* _OBARRAY_H_ */
