/******************************************************************************************/
/*

DESCRIPTION

	Implements the required macros in order to move the global variables defined
	in the agent to a per virtual core private way. Contains a structure of the
	default layout of the global variables used in Portable part

COPYRIGHT

	Copyright 2000-2010 Freescale Semiconductor, Inc.
	
*/
/*****************************************************************************************/

#ifndef __SMP_H
#define __SMP_H


#include "trk_prefix.h"
#include "dstypes.h"
#include "target.h"

#ifdef IS_SMP

/**
 * A target agent will have to implement the PRIVATE_CONTEXT macro to point
 * to the start of a memory region where the TrkPrivateContext structure
 * type will begin.
 *
 * It will also have to implement the two below function in order to initialise
 * there own version of the private context data as well as their private
 * transport layers and stacks:
 * void InitializePrivateData( void );
 * void InitializeTransportTable( void );
 */

/* Structure to store all private portable global data */
typedef struct
{
	/* Portable private TRK data common to all agents */
	/* Event queue */
	void* m_gEventQueue; /* current event queue */
	void* m_gMsgBufs;    /* auxiliary data for message buffers */

	/* Transport table and layers */
	void* m_gTransportTable;                 /* holds the current transport table */
	int   m_gCurTransport;                   /* holds the current transport id */
	ui8   m_gTrkTransportNumCommPorts;       /* the number of communication ports */
	ui8   m_gTrkTransportPrimaryIndex;       /* the primary comunication port index */
	ui8   m_gTrkTransportNumCommPortsActive; /* the maximum number of active ports */
	bool  m_gTrkConnected;                   /* the trk agent connection state */
	bool  m_gTrkAutoDownload;                /* if a target entity should auto download the targeted elf (e.g HV) */

	/* Framing state machine */
	void* m_gFramingState;   /* current state entry */
	void* m_gPreviousAck;    /* previous acknowledge message sent */
	ui8   m_gNextSendSeqId;
	ui8   m_gNextRcvSeqId;
	bool  m_gRequestCached;
	volatile ui8 m_gInputPending; /* if its interrupt driven (and it is) holds data available on input */
} TrkPrivateContext;


/* utility macros to access certain private members in TrkPrivateContext data
 * structure. Should be used as below:
 * PRIVATE_DATA( EVENT_QUEUE )
 */
#define EVENT_QUEUE m_gEventQueue
#define MESSAGE_BUFFERS m_gMsgBufs

#define TRANSPORT_TABLE_ENTRY m_gTransportTable
#define TRANSPORT_CURRENT m_gCurTransport
#define TRANSPORT_NUM_PORTS m_gTrkTransportNumCommPorts
#define TRANSPORT_PRIMARY_INDEX m_gTrkTransportPrimaryIndex
#define TRANSPORT_NUM_PORTS_ACTIVE m_gTrkTransportNumCommPortsActive
#define TRK_CONNECTED m_gTrkConnected
#define TRK_AUTO_DOWNLOAD m_gTrkAutoDownload

#define COM_FRAME_STATE m_gFramingState
#define PREVIOUS_ACKNOWLEDGE m_gPreviousAck
#define SENT_SEQUENCE m_gNextSendSeqId
#define RECEIVED_SEQUENCE m_gNextRcvSeqId
#define REQUEST_CACHED m_gRequestCached
#define INPUT_PENDING m_gInputPending

/* macro that will be used by trk agent to reference a private global data */
#define PRIVATE_DATA(struct_member) \
	((TrkPrivateContext*)PRIVATE_CONTEXT)->struct_member

/* Initialize any private data.
 * Should be called prior to calling Target Agent Initialization
 */
void InitializePrivateData( void );

/* Initialize the private transport table containing the stacks of communication layers
 * Should be called after InitializePrivateData
 */
void InitializeTransportTable( void );

#endif /* IS_SMP */

#endif
