/******************************************************************************************/
/*

DESCRIPTION

	Support commands from CWDS to host debugger.
	
	Send messages (support) from CWDS to the host debugger.
	
COPYRIGHT

	Copyright 1999-2004, 2006, 2007, 2009 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#include "trk_prefix.h"

#ifndef _NO_SYSTEM_INCLUDES
#include <string.h>
#include <stddef.h>
#endif

#include "usr_put.h"		// must be included before msgcmd.h
#include "nubassrt.h"
#include "notify.h"
#include "msgbuf.h"
#include "msg.h"
#include "msgcmd.h"
#include "usr_put.h"
#include "support.h"
#include "target.h"
#include "input.h"
#include "msgparse.h"

/* transport layer include */
#include "transport.h"
#include "transport_tbl.h"


/*
** Set this to 1 to see detailed information about  normal processing
** in these support functions (error msgs output via __puts() when
** one of the DEBUGIO_ macros is defined will always be emitted
** regardless of whether DEBUG_SUPPORT_CALLS is (1) or (0) .)
*/

#define DEBUG_SUPPORT_CALLS	(1)

#define SUPPORT_TRY_COUNT 3

/*
** Global storage for the next sequence ID for outgoing request/notify
** messages.
*/

#ifndef IS_SMP
ui8 gNextSendSeqId = 0;
#else
#define gNextSendSeqId PRIVATE_DATA(SENT_SEQUENCE)
#endif

/*
** Global indicating that an incoming request was cached while waiting for
** a message reply. The TRK uses a fixed, small number of buffers. Only
** one is reserved for this purpose (derring an incoming request while
** waiting for an ACK).
*/
#ifndef IS_SMP
bool gRequestCached = false;
#endif


/******************************************************************************/
/*
 *	SuppAccessFile
 *
 *	Handles support requests from the target program to read/write
 *	data to/from a file on the host.  Converts the request into one
 *	or more messages and sends them to the host.
 */
/******************************************************************************/
DSError SuppAccessFile( ui32 file_handle, ui8 *data, size_t *count, 
						DSIOResult *io_result, bool need_reply,
						bool read
#if DS_PROTOCOL == DS_PROTOCOL_RTOS
						, ui32 process_id, ui32 thread_id
#endif
						)
{
	DSError			err;
	
	MessageBufferId replyID;
	MessageBuffer 	*replyBuffer = NULL;
	MessageBufferId	bufferID;
	MessageBuffer	*buffer;
	size_t			length;
	size_t			done;
	ui8				msg_io_result;
	ui16			msg_length;
	bool			abort;
	
	ASSERT( count != NULL && io_result != NULL );
	
	if (data == NULL || *count <= 0) // Warning : C2897E: odd unsigned comparison with 0: '<='
		return( kParameterError );

	/*
	** The message size is limited, so the access may need to
	** be broken up into multiple read/write messages.  Loop until
	** everything has been read/written or an error occurs.
	**
	** 'length' is the amount read/written in a particular iteration,
	** and 'done' tracks the total amount read/written so far.
	*/

	for (abort = false, done = 0, err = kNoError, *io_result = kDSIONoError;
		 !abort && done < *count && err == kNoError && *io_result == kDSIONoError;
		 done += length)
	{

		/*
		** Set the length to max if there is more than that
		** left to read/write.  Otherwise, set it to the number
		** of remaining bytes.
		*/

		if ((*count - done) > DS_MAXREADWRITELENGTH)
			length = DS_MAXREADWRITELENGTH;
		else
			length = *count - done;
			
		/*
		** Get a buffer and build the message.
		*/

		err = GetFreeBuffer( &bufferID, &buffer );
		
		/*
		** The header is built using the automatic message construction
		** routine, but the remainder is still built the old way.
		** Eventually, the entire message should be handled by the
		** automatic method.
		*/
		buffer->fFields.fData[ TRK_MSG_COMMAND_INDEX ] =
			read ? kDSReadFile : kDSWriteFile;
		buffer->fFields.fData[ TRK_MSG_SID_INDEX ] =
			GetNextSendSeqId();
		if (err == kNoError)
			err = TrkConstructMessage( buffer, &gTrkMsgHeader );
			
		if (err == kNoError)
			err = MessageAdd_ui32( buffer, file_handle );
		if (err == kNoError)
			err = MessageAdd_ui16( buffer, length );

#if DS_PROTOCOL == DS_PROTOCOL_RTOS
		if (err == kNoError)
			err = MessageAdd_ui32( buffer, process_id );
		if (err == kNoError)
			err = MessageAdd_ui32( buffer, thread_id );
#endif

		/*
		** If writing, copy the write data to the message.
		*/
		
		if (!read)
		{
			if (err == kNoError)
				err = MessageAddBlock_ui8( buffer, data + done, length );
		}

		#if DEBUG_SUPPORT_CALLS
			__puts( "SuppAccessFile: sending request\r\n" );
		#endif
		
		if (err == kNoError)
		{

			/*
			** If a reply is desired, send it as a request,
			** which will look for a reply and retry the
			** message up to the specified count if no reply
			** or an improper reply is given.  The reply 
			** is returned.
			** If no reply is desired, just use MessageSend()
			** to send it out once.
			*/

			if (need_reply)
			{
				msg_length = 0;
				msg_io_result = kDSIONoError;
			
				/*
				** In earlier revisions, console reads would never
				** time out.  However, it is not a good idea to
				** hold up the communications channel in this way.
				** Now, the host debugger should return a response
				** immediately, even if it is empty.  If the host 
				** debugger replies with no input, then the next
				** time it has some input, it should send a
				** NotifyFileInput message to TRK. Therefore, 
				** all console requests should now time out if
				** they don't get a response.
				*/
				err = RequestSend( buffer, &replyID, 
								   (read) ? DS_MIN_REPLY_READFILE_LENGTH :
								   			DS_MIN_REPLY_WRITEFILE_LENGTH,
								   SUPPORT_TRY_COUNT, 
								   true );
				
				if (err == kNoError)
				{
					replyBuffer = GetBuffer( replyID );
					/*
					** Skip the ACK & error-code in the reply.
					** The RequestSend() routine did not return
					** an error, so it is safe to assume they
					** are OK.
					*/
					SetBufferPosition( replyBuffer,
									   DS_MIN_REPLY_LENGTH,
									   false );
									   
	
				}
				
				/*
				** Get the io_result code and the read/written
				** length from the reply.
				*/

				if (err == kNoError)
					err = MessageGet_ui8( replyBuffer, &msg_io_result );
				if (err == kNoError)
					err = MessageGet_ui16( replyBuffer, &msg_length );

				/*
				** If reading, get the returned data from the 
				** reply. If the returned length is bigger than
				** what CodeWarriorTRK asked for, don't copy the 
				** data to the reply buffer since it will cause
				** buffer overflow.
				*/
									
				if (read && err == kNoError)
				{
						
						if (replyBuffer->fLength != 
							(DS_MIN_REPLY_READFILE_LENGTH + msg_length))
						{
							msg_length = replyBuffer->fLength -
										 DS_MIN_REPLY_READFILE_LENGTH;
							if (msg_io_result == kDSIONoError)
								msg_io_result = kDSIOError;
						}
						
						if( !(msg_length > length) )
							err = MessageGetBlock_ui8( replyBuffer,
													   data + done,
													   (size_t) msg_length );
				}
								
				/*
				** If the returned length is not what 
				** CodeWarriorTRK asked for, then CodeWarriorTRK will
				** stop the current file access and set
				** msg_io_result to kDSIOError depending 
				** on the file mode.
				**
				** In reading mode
				** ===============
				**
				** The returned length will be smaller than
				** what CodeWarriorTRK asked for when the target
				** program request an entire buffer of input,
				** but the console return only a single
				** line.
				**
				** For example: The target program requests
				** an array of characters, but the console 
				** returns empty string.
				**
				** The returned length will be larger than
				** what CodeWarriorTRK asked for when the target 
				** program request single line of input, 
				** but the console returns the entire buffer
				** of input.
				**
				** For example: The target program requests
				** a single character input, but the console
				** returns more than one character input.
				**
				** In writting mode
				** ================
				**
				** When the returned length is smaller or
				** larger than what CodeWarriorTRK asked for, CodeWarriorTRK
				** will set msg_io_result to kDSIOError in writting
				** mode.
				*/
				if (msg_length != length)
				{
					if (!(read && (msg_length < length)))
						if (msg_io_result == kDSIONoError)
							msg_io_result = kDSIOError;

					length = msg_length;						
					abort = true;
				}

				*io_result = (DSIOResult)msg_io_result;
				
				ReleaseBuffer( replyID );
			}
			else
			{
				/* write via current transport interface */
				err = MSG_LAYER_WRITE( TRANSPORT_LAYER( gCurTransport ), buffer );
			}
		}
		
		ReleaseBuffer( bufferID );
	}

	/*
	** Set the count to the actual amount read/written and return.
	*/
	
	*count = done;
	
	return( err );
}

/******************************************************************************/
/*
 *	RequestSend
 *
 *	Send a request message to the host and wait for a reply.  If
 *	no reply is returned within a certain timeout period (target-
 *	defined), then re-send the request.  If an improper reply is
 *	received (command is not an ACK, or an error code is returned,
 *	or the received reply is too small), then that reply is 
 *	thrown away and the request is re-sent.  The retries parameter
 *	indicates the maximum number of times to re-send the message,
 *	not including the original send.
 */
/******************************************************************************/
DSError RequestSend( MessageBuffer *buffer, MessageBufferId *replyID,
					 size_t min_size, int max_retries, bool check_timeout )
{
	DSError err = kNoError;
	MessageBuffer *replyBuffer;
	TimeoutType timer;
	int tries = max_retries + 1;
	bool bad_reply = true;

	ResetTimeout(&timer);

	/*
	** Loop, sending the message and waiting for a reply.
	** If the reply is bad or there is a timeout, the loop
	** repeats until the number of retries is exhausted.
	** On loop exit, if a good reply was received, the
	** *replyID will be a valid buffer ID, otherwise it
	** will be kInvalidMessageBufferID.  
	** An internal error will abort the loop early.
	*/

	for (tries = max_retries + 1, *replyID = kInvalidMessageBufferId;
		 tries && *replyID == kInvalidMessageBufferId && err == kNoError;
		 tries--)
	{
		/* write via current transport interface */
		err = MSG_LAYER_WRITE( TRANSPORT_LAYER( gCurTransport ), buffer );
	    
		if (err == kNoError)
		{
			if (check_timeout)
				ResetTimeout( &timer );
		
			/*
			** When non-acknowledgement messages arrive, continue
			** waiting.
			*/
			
			do
			{
				/*
				** Wait for a packet to arrive.
				*/

				do {
					MSG_LAYER_READ( TRANSPORT_LAYER( gCurTransport ), replyID );
				} while ((*replyID == kInvalidMessageBufferId) &&
						 !(check_timeout && Timeout( &timer )));

				/*
				** If a reply was received, verify that it is valid.
				*/

				if (*replyID != kInvalidMessageBufferId)
				{
					#if DEBUG_SUPPORT_CALLS
						__puts("\r\nRequestSend: packet received\r\n");
					#endif
					
					bad_reply = false;
				
					replyBuffer = GetBuffer( *replyID );
					err = TrkParseMessageSegment( replyBuffer, &gTrkMsgHeader, true );

					if (err == kNoError)
					{
						/*
						** If the reply is not an acknowledgement, assume it 
						** is an asynchronous request from the debugger and 
						** defer it by putting it on the msg queue. However, 
						** the TRK uses a fixed, small number of buffers, and 
						** only one is reserved for deferring requests this
						** way. And, in fact, that's really all that's needed
						** because the debugger should never send more than
						** one request at a time. The one exception where we
						** might end up needing more than one cache buffer
						** is if, in the handling of a deferred request, we
						** start scanning the communication channel after
						** sending the ACK but before returning to the main loop.
						** This would be the case if during that deffered request
						** handling we sent the debugger a request; that would
						** require us to scan for a reply, causing us to possibly
						** pick up a new command, which cannot, at this point, 
						** be deffered (because the reserved msg buffer is already 
						** in use). Because the number of msg buffers are, again, 
						** fixed, this is something we must make sure to avoid.
						** Rather than send the new request immediately, the TRK
						** should post a support event (which does not require
						** a msg buffer) so that the request is sent to the 
						** debugger upon returning to the main loop.
						*/
					
						if (replyBuffer->fFields.fData[ TRK_MSG_COMMAND_INDEX ] < kDSReplyACK)
						{
							if (!gRequestCached)
							{
								#if DEBUG_SUPPORT_CALLS
									__puts( "Caching incoming request\r\n" );
								#endif
								PutMsgOnEventQueue( *replyID );
								gRequestCached = true;
							}
							else
							{
								/* Either the debugger screwed up, or we did */
								ASSERT(0);
								__puts( "Ignoring incoming request (too many)\r\n" );
								ReleaseBuffer( *replyID );
							}
							*replyID = kInvalidMessageBufferId;
							gRequestCached = true;
							continue;
						}
					}
				
					if (err == kNoError)
						err = TrkParseMessageSegment( replyBuffer, &gTrkMsgReplyHeader, false );
					
#if TRK_MSG_SEQUENCE_IDS
					if (err == kNoError)
					{
						/*
						** If the reply is an ACK and the sequence ID does not match the
						** sent sequence ID, then assume it is a duplicate reply for an
						** earlier re-sent message.  Ignore it and continue waiting.
						*/
						if (replyBuffer->fFields.fData[ TRK_MSG_COMMAND_INDEX ] == kDSReplyACK &&
							replyBuffer->fFields.fData[ TRK_MSG_SID_INDEX ] != buffer->fFields.fData[ TRK_MSG_SID_INDEX ])
						{
							ReleaseBuffer( *replyID );
							*replyID = kInvalidMessageBufferId;
							continue;
						}
					}
#endif /* TRK_MSG_SEQUENCE_IDS */

				}
				
				break;
			} while (1);
				
			if (*replyID != kInvalidMessageBufferId)
			{
				/*
				** The reply must be at least the minimum size.
				*/
				
				if (replyBuffer->fLength < min_size)
				{
					bad_reply = true;
					__puts( "\r\nRequestSend: reply too small\r\n" );
				}
				
				/*
				** Get the message command and error values and
				** verify that this is an ACK message with no
				** errors.
				*/
					
				if (err == kNoError && !bad_reply)
				{
					if ((MessageCommandID)replyBuffer->fFields.fData[ TRK_MSG_COMMAND_INDEX ] != kDSReplyACK ||
						(DSReplyError)replyBuffer->fFields.fData[ TRK_MSG_REPLY_ERROR_INDEX ] != kDSReplyNoError)
					{
						bad_reply = true;
						__puts( "\r\nRequestSend: Non-ACK or ACK-error received.\r\n" );
					}
				}
				

				/*
				** If the reply wasn't good, get rid of it and
				** reset the *replyID.
				*/

				if (err != kNoError || bad_reply)
				{
					__puts( "\r\nRequestSend: bad reply\r\n" );
				
					ReleaseBuffer( *replyID );
					*replyID = kInvalidMessageBufferId;
				}
#if DEBUG_SUPPORT_CALLS
				else
				{
					__puts( "\r\nRequestSend: reply OK\r\n" );
				}
#endif
			}
			else
			{
				__puts( "RequestSend: timed out waiting for reply\r\n" );
			}
			
		}
	}

	if (*replyID == kInvalidMessageBufferId)
		err = kWaitACKError;

	return err;
}
/******************************************************************************/
/*
 *	GetNextSendSeqId
 *
 *	Get the next sequence ID for an outgoing request/notify message.
 */
/******************************************************************************/
ui8 GetNextSendSeqId()
{
	ui8 result;
	
	result = gNextSendSeqId;
	gNextSendSeqId = (gNextSendSeqId == 0xff) ? 1 : (gNextSendSeqId + 1);
	return( result );
}

/******************************************************************************/
/*
 *	ResetRequestSeqId
 *
 *	Reset the ID sequence for outgoing request/notify messages.
 */
/******************************************************************************/
void ResetSendSeqId( ui8 val )
{
	gNextSendSeqId = val;
}
