/******************************************************************************************/
/*

DESCRIPTION

	API for target-specific implementation.  Implementing these basic
	functions will allow a primitive version of CWDS (CWDS L1) to
	become operational.
	
	Functions may only work for 32-bit (or smaller) processors.
	
COPYRIGHT

	Copyright 1999-2003, 2006-2010 Freescale Semiconductor, Inc.
	
*/
/******************************************************************************************/

#ifndef __TARGIMPL_H__
#define __TARGIMPL_H__

#ifndef _NO_SYSTEM_INCLUDES
#include <stddef.h>
#endif

#include "dstypes.h"
#include "dserror.h"
#include "nubevent.h"
#include "target.h"
#include "step.h"

typedef enum ValidMemoryOptions {
	kValidMemoryReadable,
	kValidMemoryWriteable
} ValidMemoryOptions;

typedef enum MemoryAccessOptions {
	kUserMemory,					/* modify user memory (invisible bp) */
	kDebuggerMemory,				/* modify real memory (visible bp) */
	kPhysicalSpace,					/* access on physical memory space */
	kVirtualInstructionSpace,			/* access instructions on virtual memory space */
	kVirtualDataSpace				/* access data on virtual memory space */
} MemoryAccessOptions;

enum {								/* DSError; see dserror.h */
	kInvalidMemory = 0x0700,
	kInvalidRegister,
	kCWDSException,
	kUnsupportedError,
	kInvalidProcessId,
	kInvalidThreadId,
	kOsError
};

typedef struct DSVersions {
	ui8			kernelMajor;				/* (1) */
	ui8			kernelMinor;				/* (1) */
	ui8			protocolMajor;				/* (1) */
	ui8			protocolMinor;				/* (1) */
} DSVersions;

typedef struct DSCPUType {
	ui8			cpuMajor;					/* (1) */
	ui8			cpuMinor;					/* (1) */
	ui8			bigEndian;					/* (1) */
	ui8			defaultTypeSize;			/* (1) */
	ui8			fpTypeSize;					/* (1) */
	ui8			extended1TypeSize;			/* (1) */
	ui8			extended2TypeSize;			/* (1) */
} DSCPUType;
__extern_c

	DSError	ValidMemory32(const void* addr, size_t length,
							ValidMemoryOptions readWriteable);
	
	DSError TargetAccessMemory(void* Data,
							void* virtualAddress,
							size_t* memorySize,
							MemoryAccessOptions accessOptions,
							bool read );
	DSError HyperTargetAccessMemory(void* Data,
							void* virtualAddress,
							size_t* memorySize,
							MemoryAccessOptions accessOptions,
							bool read );
	DSError OsTargetAccessMemory(void* Data,
							void* virtualAddress,
							size_t* memorySize,
							MemoryAccessOptions accessOptions,
							bool read,
							ui32 processId,
							ui32 threadId );
							
	DSError TargetAccessExtendedMemory(void* Data,
							void* addressHigh,
							void* addressLow,
							size_t* memorySize,
							MemoryAccessOptions accessOptions,
							bool read );
							
	DSError HyperTargetAccessExtendedMemory(void* Data,
							void* addressHigh,
							void* addressLow,
							size_t* memorySize,
							MemoryAccessOptions accessOptions,
							bool read );
	DSError OsTargetAccessExtendedMemory(void* Data,
							void* addressHigh,
							void* addressLow,
							size_t* memorySize,
							MemoryAccessOptions accessOptions,
							bool read,
							ui32 processId,
							ui32 threadId );

	DSError	TargetAccessDefault(unsigned int firstRegister, unsigned int lastRegister,
							MessageBuffer *b, size_t* registerStorageSize,
							bool read );
	DSError	HyperTargetAccessDefault(unsigned int firstRegister, unsigned int lastRegister,
								MessageBuffer *b, size_t* registerStorageSize,
								bool read );
	DSError	OsTargetAccessDefault(unsigned int firstRegister, unsigned int lastRegister,
							MessageBuffer *b, size_t* registerStorageSize,
							bool read,
							ui32 processID,
							ui32 threadID );

	DSError	TargetAccessDefaultSingle(unsigned int reg, DefaultType *value,
							bool read );
	DSError	HyperTargetAccessDefaultSingle(unsigned int reg, DefaultType *value,
								bool read );
	DSError	OsTargetAccessDefaultSingle(unsigned int reg, DefaultType *value,
							bool read,
							ui32 processId,
							ui32 threadId );


	DSError	TargetAccessFP(unsigned int firstRegister, unsigned int lastRegister,
							MessageBuffer *b, size_t* registerStorageSize,
							bool read );
	DSError	HyperTargetAccessFP(unsigned int firstRegister, unsigned int lastRegister,
								MessageBuffer *b, size_t* registerStorageSize,
								bool read );
	DSError	OsTargetAccessFP(unsigned int firstRegister, unsigned int lastRegister,
							MessageBuffer *b, size_t* registerStorageSize,
							bool read,
							ui32 processId,
							ui32 threadId );

	DSError	TargetAccessFPSingle(unsigned int reg, FPType *value,
							bool read );
	DSError	HyperTargetAccessFPSingle(unsigned int reg, FPType *value,
								bool read );
	DSError	OsTargetAccessFPSingle(unsigned int reg, FPType *value,
							bool read,
							ui32 processId,
							ui32 threadId );

	DSError	TargetAccessExtended1(unsigned int firstRegister, unsigned int lastRegister,
							MessageBuffer *b, size_t* registerStorageSize,
							bool read );
	DSError	HyperTargetAccessExtended1(unsigned int firstRegister, unsigned int lastRegister,
								MessageBuffer *b, size_t* registerStorageSize,
								bool read );
	DSError	OsTargetAccessExtended1(unsigned int firstRegister, unsigned int lastRegister,
							MessageBuffer *b, size_t* registerStorageSize,
							bool read,
							ui32 processId,
							ui32 threadId );

	DSError	TargetAccessExtended1Single(unsigned int reg, Extended1Type *value,
							bool read );
	DSError	HyperTargetAccessExtended1Single(unsigned int reg, Extended1Type *value,
								bool read );
	DSError	OsTargetAccessExtended1Single(unsigned int reg, Extended1Type *value,
							bool read,
							ui32 processId,
							ui32 threadId );

	DSError	TargetAccessExtended2(unsigned int firstRegister, unsigned int lastRegister,
							MessageBuffer *b, size_t* registerStorageSize,
							bool read );
	DSError	HyperTargetAccessExtended2(unsigned int firstRegister, unsigned int lastRegister,
								MessageBuffer *b, size_t* registerStorageSize,
								bool read );
	DSError	OsTargetAccessExtended2(unsigned int firstRegister, unsigned int lastRegister,
							MessageBuffer *b, size_t* registerStorageSize,
							bool read,
							ui32 processId,
							ui32 threadId );
							
	DSError	TargetAccessTLBs(unsigned int firstRegister, unsigned int lastRegister,
							MessageBuffer *b,
							bool read,
							unsigned int tlb );
	DSError	HyperTargetAccessTLBs(unsigned int firstRegister, unsigned int lastRegister,
								MessageBuffer *b, 
								bool read, 
								unsigned int tlb);
	DSError	OsTargetAccessTLBs(unsigned int firstRegister, unsigned int lastRegister,
							MessageBuffer *b,
							bool read,
							unsigned int tlb,
							ui32 processId,
							ui32 threadId );

#ifdef TRK_EXTENDED2_SIZE
	DSError	TargetAccessExtended2Single(unsigned int reg, Extended2Type *value,
							bool read );
	DSError	HyperTargetAccessExtended2Single(unsigned int reg, Extended2Type *value,
								bool read );
	DSError	OsTargetAccessExtended2Single(unsigned int reg, Extended2Type *value,
							bool read,
							ui32 processId,
							ui32 threadId );
#endif

	DSError TargetVersions( DSVersions *versions );
	DSError TargetSupportMask( DSSupportMask *mask );
	DSError TargetCPUType( DSCPUType *cpuType );
	DSError TargetCheckException(void);

    DSError InitializeTarget( void );
    DSError CleanupTarget( void );

	DSError	TargetContinue( void );
	DSError	HyperTargetContinue( void );
	DSError	OsTargetContinue( ui32 processId, ui32 threadId );

	DSError	TargetInterrupt(NubEvent* event);

	DSError TargetAddStopInfo( MessageBuffer* buffer );
	DSError HyperTargetAddStopInfo( MessageBuffer* buffer );
	DSError OsTargetAddStopInfo( MessageBuffer* buffer,
							ui32 processId,
							ui32 threadId );
	DSError TargetAddExceptionInfo( MessageBuffer *buffer );
	DSError HyperTargetAddExceptionInfo( MessageBuffer *buffer );
	DSError OsTargetAddExceptionInfo( MessageBuffer *buffer );

	DSError TargetSingleStep( unsigned count, bool stepOver,
							ui32 processId,
							ui32 threadId );
	DSError HyperTargetSingleStep( unsigned count, bool stepOver );

	DSError TargetStepOutOfRange( ui32 start, ui32 end,
							bool stepOver,
							ui32 processId,
							ui32 threadId );
	DSError HyperTargetStepOutOfRange( ui32 start, ui32 end, bool stepOver );

	PCType TargetGetPC( void );
	PCType HyperTargetGetPC(void);
	PCType OsTargetGetPC( ui32 processId, ui32 threadId );

	DSError TargetSupportRequest( NubEvent *event );
	DSError TargetFlushCache( ui8 options,
							  void *start, void *end );
	bool TargetStopped( void );
	bool HyperTargetStopped( void );
	bool OsTargetStopped( ui32 processId, ui32 threadId );
	void TargetSetStopped( bool stop_state );
	DSError TargetStop( void );
	DSError HyperTargetStop( void );
	DSError OsTargetStop( ui8 options, ui32 processId, ui32 threadId );

	ui8 TargetCPUMinorType( void );

	void *TargetTranslate( void *address );
	
		
	DSError OsTargetValidStep( ui32 processId, ui32 threadId );
	
	DSError TargetAccessInstruction(InstructionType* readData,
								void* virtualAddress,
								size_t maxSize,
								MemoryAccessOptions accessOptions,
								bool read);
	DSError HyperTargetAccessInstruction(InstructionType* readData,
								void* virtualAddress,
								size_t maxSize,
								MemoryAccessOptions accessOptions,
								bool read);
	DSError OsTargetAccessInstruction(InstructionType* readData,
								void* virtualAddress,
								size_t maxSize,
								MemoryAccessOptions accessOptions,
								bool read,
								ui32 processId, ui32 threadId );

	StepStatii *TargetGetStepStatii( void );
	StepStatii *HyperTargetGetStepStatii( void );
	StepStatii *OsTargetGetStepStatii( ui32 processId, ui32 threadId );

	DSError TargetSetupTraceNoBreakpoints( StepStatus *status );
	DSError HyperTargetSetupTraceNoBreakpoints( StepStatus *status );
	DSError OsTargetSetupTraceNoBreakpoints( StepStatus *status,
										ui32 processId, ui32 threadId );
	DSError TargetCleanupTraceNoBreakpoints( StepStatus *status );
	DSError HyperTargetCleanupTraceNoBreakpoints( StepStatus *status );
	DSError OsTargetCleanupTraceNoBreakpoints( StepStatus *status,
											ui32 processId, ui32 threadId );

	DSError TargetSetupTrace( StepStatus *status,
								ui32 processId, ui32 threadId );

	DSError TargetCleanupTrace( StepStatus *status,
								ui32 processId, ui32 threadId );

	bool AtTraceBreakPoint( StepStatus *status );
	bool HyperAtTraceBreakPoint( StepStatus *status );
	bool OsAtTraceBreakPoint( StepStatus *status, ui32 processId, ui32 threadId );

	DSIOResult TargetWriteFile( ui32 handle, ui8 *data, size_t *length );
	DSIOResult TargetReadFile( ui32 handle, ui8 *data, size_t *length );
	DSIOResult TargetOpenFile( DSFileOpenModes modes, unsigned char *name,
			ui32 *handle, ui32 *timestamp );
	DSIOResult TargetCloseFile( ui32 handle, ui32 timestamp );
	DSIOResult TargetPositionFile( DSFilePositionModes mode, ui32 handle, ui32 offset );

	DSError TargetNotifyFileInput( ui32 handle );
	DSError HyperTargetNotifyFileInput( ui32 handle );
	DSError OsTargetNotifyFileInput( ui32 handle, ui32 processId, ui32 threadId );

	DSError TargetBlockFileIo( bool block );
	DSError HyperTargetBlockFileIo( bool block );
	DSError OsTargetBlockFileIo( bool block );

	DSError TargetSetBreakpoint( Breakpoint *b );
	DSError HyperTargetSetBreakpoint( Breakpoint *b );
	DSError OsTargetSetBreakpoint( Breakpoint *b );

	DSError TargetClearBreakpoint( Breakpoint *b );
	DSError HyperTargetClearBreakpoint( Breakpoint *b );
	DSError OsTargetClearBreakpoint( Breakpoint *b );

	DSError TargetHideBreakpoint( Breakpoint *b, void *data, size_t maxLen  );
	DSError HyperTargetHideBreakpoint( Breakpoint *b, void *data, size_t maxLen  );
	DSError OsTargetHideBreakpoint( Breakpoint *b, void *data, size_t maxLen  );

	bool TargetMatchExecBreakpoint( Breakpoint *b, unsigned int start,
									unsigned int len );
	bool HyperTargetMatchExecBreakpoint( Breakpoint *b, unsigned int start,
									unsigned int len );
	bool OsTargetMatchExecBreakpoint( Breakpoint *b, unsigned int start,
									unsigned int len,
									ui32 processId, ui32 threadId );
	DSError OsTargetModifyBreakpointThread( Breakpoint *b, ui32 threadId );
	DSError	TargetValidateBreakpoint( Breakpoint *b );

	DSError TargetConnect( void );
	DSError TargetDisconnect( void );

#if TRK_TRANSPORT_INT_DRIVEN
	void TargetSetInputPendingPtr( volatile unsigned char *inputPendingPtr );
#endif /* TRK_TRANSPORT_INT_DRIVEN */

__end_extern_c

/*
** Define aliases for the functions which have both OS and non-OS
** variants.  Each alias references the variant which is appropriate
** for the current protocol level.
*/

#if DS_PROTOCOL < DS_PROTOCOL_HYPER /* DS_PROTOCOL_MIN, DS_PROTOCOL_BOARD */
	#define XTargetAccessMemory		TargetAccessMemory
	#define XTargetAccessExtendedMemory	TargetAccessExtendedMemory
	#define XTargetAccessDefault	TargetAccessDefault
	#define XTargetAccessFP			TargetAccessFP
	#define XTargetAccessExtended1	TargetAccessExtended1
	#define XTargetAccessExtended2	TargetAccessExtended2
	#define XTargetAccessTLBs	TargetAccessTLBs
	#define XTargetContinue			TargetContinue
	#define XTargetGetPC			TargetGetPC
	#define XTargetStopped			TargetStopped
	#define XTargetStop				TargetStop
	#define XTargetAddStopInfo		TargetAddStopInfo
	#define XTargetAddExceptionInfo	TargetAddExceptionInfo
	#define XTargetSingleStep       TargetSingleStep
	#define XTargetStepOutOfRange   TargetStepOutOfRange
	#define XTargetSetBreakpoint	TargetSetBreakpoint
	#define XTargetClearBreakpoint	TargetClearBreakpoint
	#define XTargetHideBreakpoint	TargetHideBreakpoint
	#define XTargetGetStepStatii	TargetGetStepStatii
	#define XTargetSetupTraceNoBreakpoints	TargetSetupTraceNoBreakpoints
	#define XTargetCleanupTraceNoBreakpoints	TargetCleanupTraceNoBreakpoints
	#define XAtTraceBreakPoint		AtTraceBreakPoint
	#define XTargetAccessDefaultSingle	TargetAccessDefaultSingle
	#define XTargetAccessFPSingle	TargetAccessFPSingle
	#define XTargetAccessExtended1Single	TargetAccessExtended1Single
	#define XTargetAccessExtended2Single	TargetAccessExtended2Single
	#define XTargetAccessInstruction	TargetAccessInstruction
	#define XTargetNotifyFileInput	TargetNotifyFileInput
	#define XTargetMatchExecBreakpoint	TargetMatchExecBreakpoint
	#define XTargetBlockFileIo		TargetBlockFileIo

#elif DS_PROTOCOL < DS_PROTOCOL_RTOS /* DS_PROTOCOL_HYPER */
	#define XTargetAccessMemory     HyperTargetAccessMemory
	#define XTargetAccessExtendedMemory	HyperTargetAccessExtendedMemory
	#define XTargetAccessDefault    HyperTargetAccessDefault
	#define XTargetAccessFP			HyperTargetAccessFP
	#define XTargetAccessExtended1  HyperTargetAccessExtended1
	#define XTargetAccessExtended2  HyperTargetAccessExtended2
	#define XTargetAccessTLBs	HyperTargetAccessTLBs
	#define XTargetContinue         HyperTargetContinue
	#define XTargetGetPC            HyperTargetGetPC
	#define XTargetStopped          HyperTargetStopped
	#define XTargetStop	            HyperTargetStop
	#define XTargetAddStopInfo      HyperTargetAddStopInfo
	#define XTargetAddExceptionInfo HyperTargetAddExceptionInfo
	#define XTargetSingleStep       HyperTargetSingleStep
	#define XTargetStepOutOfRange   HyperTargetStepOutOfRange
	#define XTargetSetBreakpoint    HyperTargetSetBreakpoint
	#define XTargetClearBreakpoint  HyperTargetClearBreakpoint
	#define XTargetHideBreakpoint   HyperTargetHideBreakpoint
	#define XTargetGetStepStatii    HyperTargetGetStepStatii
	#define XTargetSetupTraceNoBreakpoints   HyperTargetSetupTraceNoBreakpoints
	#define XTargetCleanupTraceNoBreakpoints HyperTargetCleanupTraceNoBreakpoints
	#define XAtTraceBreakPoint      HyperAtTraceBreakPoint
	#define XTargetAccessDefaultSingle   HyperTargetAccessDefaultSingle
	#define XTargetAccessFPSingle        HyperTargetAccessFPSingle
	#define XTargetAccessExtended1Single HyperTargetAccessExtended1Single
	#define XTargetAccessExtended2Single HyperTargetAccessExtended2Single
	#define XTargetAccessInstruction     HyperTargetAccessInstruction
	#define XTargetNotifyFileInput       HyperTargetNotifyFileInput
	#define XTargetMatchExecBreakpoint   HyperTargetMatchExecBreakpoint
	#define XTargetBlockFileIo           HyperTargetBlockFileIo

#else
	#define XTargetAccessMemory     OsTargetAccessMemory
	#define XTargetAccessExtendedMemory	OsTargetAccessExtendedMemory
	#define XTargetAccessDefault    OsTargetAccessDefault
	#define XTargetAccessFP         OsTargetAccessFP
	#define XTargetAccessExtended1  OsTargetAccessExtended1
	#define XTargetAccessExtended2  OsTargetAccessExtended2
	#define XTargetAccessTLBs	OsTargetAccessTLBs
	#define XTargetContinue         OsTargetContinue
	#define XTargetGetPC            OsTargetGetPC
	#define XTargetStopped          OsTargetStopped
	#define XTargetStop             OsTargetStop
	#define XTargetAddStopInfo      OsTargetAddStopInfo
	#define XTargetAddExceptionInfo OsTargetAddExceptionInfo
	#define XTargetSingleStep       TargetSingleStep
	#define XTargetStepOutOfRange   TargetStepOutOfRange
	#define XTargetSetBreakpoint    OsTargetSetBreakpoint
	#define XTargetClearBreakpoint  OsTargetClearBreakpoint
	#define XTargetHideBreakpoint   OsTargetHideBreakpoint
	#define XTargetGetStepStatii    OsTargetGetStepStatii
	#define XTargetSetupTraceNoBreakpoints   OsTargetSetupTraceNoBreakpoints
	#define XTargetCleanupTraceNoBreakpoints OsTargetCleanupTraceNoBreakpoints
	#define XAtTraceBreakPoint           OsAtTraceBreakPoint
	#define XTargetAccessDefaultSingle   OsTargetAccessDefaultSingle
	#define XTargetAccessFPSingle        OsTargetAccessFPSingle
	#define XTargetAccessExtended1Single OsTargetAccessExtended1Single
	#define XTargetAccessExtended2Single OsTargetAccessExtended2Single
	#define XTargetAccessInstruction     OsTargetAccessInstruction
	#define XTargetNotifyFileInput       OsTargetNotifyFileInput
	#define XTargetMatchExecBreakpoint   OsTargetMatchExecBreakpoint
	#define XTargetBlockFileIo           OsTargetBlockFileIo

#endif /* #if DS_PROTOCOL < DS_PROTOCOL_RTOS */

#endif /* __TARGIMPL_H__ */
